﻿#include "pch.h"
#include "Camera.h"

void Camera::shake() {
	t_since_shake = 0;
}


Camera::Camera() { }

void Camera::update_matrices() {
	
	this->prev_jitt_x = this->curr_jitt_x;
	this->prev_jitt_y = this->curr_jitt_y;
	// #ifdef EDITOR
	if (WEngine->io->mouse_grabbed) {
		float* pitch = &this->pitch_yaw_roll[0];
		float* yaw = &this->pitch_yaw_roll[1];
		*pitch += WEngine->io->delta_mouse_y * 1.0;
		*yaw += WEngine->io->delta_mouse_x * 1.0;
		*pitch = glm::max(*pitch,-0.49f);
		*pitch = glm::min(*pitch,0.49f);
	}
	// #endif

	float pitch = this->pitch_yaw_roll[0];
	float yaw = this->pitch_yaw_roll[1];

	if(this->mode == CameraMode::PILOT) {
		dir = glm::vec3(0,0,1);
		dir = glm::normalize(dir);

		dir = get_rot_mat(0, -pitch * tau * 0.5f ) * dir;
		dir = get_rot_mat(1, yaw * tau * 0.5f ) * dir;
	} else {
		glm::vec3 gdir = glm::vec3(this->look_at.x, this->look_at.y, this->look_at.z);
		gdir = gdir - glm::vec3(this->pos.x, this->pos.y, this->pos.z);
		gdir = normalize(gdir);
		dir = gdir;
	}

	right = this->world_up * -1.0f;
	// right = right.cross(dir);
	right = glm::cross(right, dir);
	right = glm::normalize(right);


	up = dir;
	up = glm::cross(up, right);
	up = glm::normalize(up);

	// this->V_prev.copy_from_other(this->V);
	// this->P_prev.copy_from_other(this->P);
	this->V_prev = this->V;
	this->P_prev = this->P;

	// transformation_matrix.get_transformation_matrix(-pos.x, -pos.y, -pos.z);
	glm::mat4 identity = glm::mat4(
		1,0,0,pos.x,
		0,1,0,pos.y,
		0,0,1,pos.z,
		0,0,0,1
	);
	transformation_matrix = identity;
	// transformation_matrix = glm::translate(identity,pos);

	// ?
	// this->V = transformation_matrix * glm::lookAt(this->pos, this->pos + dir, glm::vec3(0,1,0));
	// this->V = glm::lookAt(this->pos, this->pos + dir, glm::vec3(0,1,0)) * transformation_matrix;
	this->V = glm::lookAt(this->pos, this->pos + dir, glm::vec3(0,1,0));
	// this->V.get_look_at_matrix(right, up, this->dir, this->pos);
	// this->V.mul(this->transformation_matrix);

	// this->V.identity();
	// this->V.mul(transformation_matrix);
	// this->V.mul(look_at_mat);
	float ratio = float(WEngine->RESX) / float(WEngine->RESY);

	this->P = glm::perspective(this->fov*2.0f, ratio, this->_near, this->_far);
	// this->P.get_perspective_proj_mat(
	// 	this->fov, this->_near, this->_far, WEngine->RESX, WEngine->RESY
	// );

	float shake_amt = 50*glm::smoothstep<float>(0.0,0.02, t_since_shake) * (1.-glm::smoothstep<float>(0.2,0.5, t_since_shake));

	this->curr_jitt_x = Camera::radical_inverse[((WEngine->frame) % 16) * 2] * (
		this->taa_jitt_amt
		+ shake_amt
	) / WEngine->RESX;
	this->curr_jitt_y = Camera::radical_inverse[((WEngine->frame + 1) % 16) * 2] * (
		this->taa_jitt_amt
		+ shake_amt
	) / WEngine->RESY;

	// float shake_amt = smoothstep(0.0,0.02, t_since_shake);

	// this->curr_jitt_x *= 1. + shake_amt;
	// this->curr_jitt_y *= 1. + shake_amt;

	this->curr_offs_x = 0;
	this->curr_offs_y = 0;
	this->curr_offs_x += viewspace_offs.x;
	this->curr_offs_y += viewspace_offs.y;


	// this->P.at(3, 0) += this->curr_jitt_x;
	// this->P.at(3, 1) += this->curr_jitt_y;

	this->P[2][0] += this->curr_jitt_x;
	this->P[2][1] += this->curr_jitt_y;
	this->P[2][0] += this->curr_offs_x/ WEngine->RESX;
	this->P[2][1] += this->curr_offs_y/ WEngine->RESY;
	// this->P[0][2] += this->curr_jitt_x;
	// this->P[1][2] += this->curr_jitt_y;
	// this->P[0][2] += this->curr_offs_x/ WEngine->RESX;
	// this->P[1][2] += this->curr_offs_y/ WEngine->RESY;

	// this->P.at(2, 0) += this->curr_jitt_x;
	// this->P.at(2, 1) += this->curr_jitt_y;
	// this->P.at(2, 0) += this->curr_offs_x/ WEngine->RESX;
	// this->P.at(2, 1) += this->curr_offs_y/ WEngine->RESY;

	this->P_inv = glm::inverse(this->P);
	this->V_inv = glm::inverse(this->V);
	// this->P_inv.copy_from_other(this->P);
	// this->V_inv.copy_from_other(this->V);


	if(this->prev_pos != this->pos || this->prev_right != this->right) {
		camera_moved_this_frame = true;
	}
	prev_pos = pos;
	prev_right = right;
	//
	// this->P_inv.inverse();
	// this->V_inv.inverse();
}

void Camera::handle_movement() {
	#ifdef EDITOR
	if(this->noclip) {
		if (WEngine->io->mouse_grabbed) {
			glm::vec2 keyInput = glm::vec2(0, 0);
			if (WEngine->io->get_key(Key::A).down) {
				keyInput[0] -= 1;
			}
			if (WEngine->io->get_key(Key::D).down) {
				keyInput[0] += 1;
			}
			if (WEngine->io->get_key(Key::W).down) {
				keyInput[1] += 1;
			}
			if (WEngine->io->get_key(Key::S).down) {
				keyInput[1] -= 1;
			}

			float walk_speed = WEngine->delta_time * 5.0 * this->speed; // deltatime

			if (WEngine->io->get_key(Key::LShift).down) {
				walk_speed *= 2.0;
			}

			glm::vec3 deltaDir = dir * keyInput[1] * walk_speed;
			glm::vec3 deltaRight = right * keyInput[0] * walk_speed;

			glm::vec3 offsVec = deltaRight + deltaDir;
			// offsVec.normalize();
			this->pos += offsVec;
		}
	}
	if(!this->walking_locked) {
		// if (WEngine->editor->mouse_grabbed) {
		// 	Vec3 keyInput = Vec3({0, 0});
		// 	if (WEngine->editor->get_key(Key::A).down) {
		// 		keyInput.vals[0] -= 1;
		// 	}
		// 	if (WEngine->editor->get_key(Key::D).down) {
		// 		keyInput.vals[0] += 1;
		// 	}
		// 	if (WEngine->editor->get_key(Key::W).down) {
		// 		keyInput.vals[1] += 1;
		// 	}
		// 	if (WEngine->editor->get_key(Key::S).down) {
		// 		keyInput.vals[1] -= 1;
		// 	}
		//
		// 	float walk_speed = WEngine->delta_time * 5.0 * this->speed; // deltatime
		//
		// 	if (WEngine->editor->get_key(Key::LShift).down) {
		// 		walk_speed *= 2.0;
		// 	}
		//
		// 	Vec3 deltaDir = dir * keyInput[1] * walk_speed;
		// 	Vec3 deltaRight = right * keyInput[0] * walk_speed;
		//
		// 	Vec3 offsVec = deltaRight + deltaDir;
		// 	// offsVec.normalize();
		// 	pos += offsVec;
		// }
	}
	#endif
}

void Camera::update() {
	this->t_since_shake += WEngine->delta_time;
	if (this->mode == CameraMode::PILOT) {
		this->update_matrices();
		this->handle_movement();
	} else {
		this->update_matrices();
	}
}
