#pragma once
#include "gl/Texture.h"

// #include "WAlloc.h"
// #include "util/WArray.h"

class LUT;
class DrawerMDI;
class StartMenu;
class TextDrawer;
class Font;
class AudioFile;
class ModelNode;
class ShaderProgram;
class Light;
class Shader;
class Camera;
struct ShaderManager;
class Editor;
class Framebuffer;
class Texture;
class Buffer;
class WGL;

class Model;

class AudioMixer;
class Physics;

#include "engine/Physics.h"



class WEntity;
struct WCollision;
struct DebugDrawer;
struct Drawer2D;
class Drawer3D;
class IO;


#include "WEntity.h"
#include "engine/Drawer3D.h"


enum class CellType {
	WALL = '*',
	FLOOR = ' ',
	ENEMY = 'B',
	DOOR = 'D',
};

class Cell {
public:
	CellType cell_type;
	EnemyType enemy_type;
	Cell() = default;
	Cell(CellType _cell_type) {
		this->cell_type = _cell_type;
	}
};

class Level {
public:
	class BFS {
		public:
		struct AStarNode {
			bool active = false;
			bool finished = false;
			bool start_node = false;
			bool end_node = false;
			AStarNode* prev_node = nullptr;
			glm::ivec2 pos;
			AStarNode() = default;
		};
		Level* level = nullptr;
		std::vector<std::vector<AStarNode>> nodes;
		BFS(){
		
		}
		void resize(Level* _level) {
			this->level = _level;
			nodes.resize(level->height);
			for(auto& vh : nodes) {
				vh.resize(level->width);
			}
		}

		glm::ivec2 found_walk_dir = glm::ivec2(0);
		bool run(glm::ivec2 start_pos, glm::ivec2 end_pos, const std::function<bool(Cell&)>& is_cell_walkable, Drawer3D* drawer_3d) {
					// glm::ivec2 start_pos = entt->props.player_props.pos;
					// glm::ivec2 end_pos = v.player->props.player_props.pos;

					auto get_node = [&](glm::ivec2 p) -> AStarNode& {
						return nodes[p.y][p.x];
					};
					
					for(int h = 0; h < level->height; h++) {
						for(int w = 0; w < level->width; w++) {
							glm::ivec2 p(w,h);
							get_node(glm::ivec2(w,h)) = AStarNode();
							get_node(glm::ivec2(w,h)).pos = p;
						}
					}

					// auto get_is_node_walkable = [&](glm::ivec2 p) {
					// 	// if(v.level.get_cell())
					// };

					auto try_node = [&](AStarNode& prev_node, glm::vec2 p) -> bool {
						AStarNode& n = get_node(p);
						Cell& c = level->get_cell(p);
						bool walkable = is_cell_walkable(c);
						if(walkable && n.active == false) {
							n.active = true;
							n.prev_node = &prev_node;
							
							// drawer_3d->draw_line(
							// 	{glm::vec3(prev_node.pos.x, 0, prev_node.pos.y)*2.0f},
							// 	{glm::vec3(n.pos.x, 0, n.pos.y)*2.0f}
							// 	);
							return true;
						} else {
							return false;
						}
					};
					auto try_walk = [&](AStarNode& n) -> bool {
						if(n.end_node && n.active) {
							return true;
						}
						if(n.active && !n.finished) {
							bool t = try_node(n, n.pos + glm::ivec2(0,1));
							bool b = try_node(n, n.pos + glm::ivec2(0,-1));
							bool r = try_node(n, n.pos + glm::ivec2(1, 0));
							bool l = try_node(n, n.pos + glm::ivec2(-1, 0));
							if(!t && !b && !r && !l) {
								n.finished = true;
							}
						}
						return false;
					};

				
					get_node(start_pos).start_node = true;
					get_node(start_pos).active = true;
					
					get_node(end_pos).end_node = true;


					bool found = false;
					for(int iter = 0; iter < 50; iter++) {
						if(found) {
							break;
						}
						for(int h = 0; h < level->height; h++) {
							if(found) {
								break;
							}
							for(int w = 0; w < level->width; w++) {
								AStarNode& n = get_node(glm::ivec2(w,h));
								bool finished = try_walk(n);
								if(finished) {
									// editor->gui_dbg_print_msg("finished");
									found = true;
									break;
								}
							}
						}
					}
				AStarNode* node_to_find = &get_node(end_pos);
				AStarNode* start_node = &get_node(start_pos);
			
				// bool found_dir_to_walk_towards = false;
				while(true) {
					// if(node_to_find->prev_node->pos == get_node(start_pos).pos) {
					if(node_to_find->prev_node == start_node) {
						
						break;
						// node_to_find = node_to_find->prev_node;
					} else {
						node_to_find = node_to_find->prev_node;
					}
				}
			this->found_walk_dir = node_to_find->pos - get_node(start_pos).pos;
			return found;
		}
	};
	
	int curr_lev = 0;
	int width = 0;
	int height = 0;
	bool finished_game = false;

	BFS bfs;

	glm::ivec2 player_start_pos = glm::ivec2(0);

	std::vector<std::vector<Cell>> level;
	std::vector<glm::ivec2> torches;
	glm::ivec2 exit_door_pos;
	
	Level() {
		parse(curr_lev);
	}

	Cell& get_cell(glm::ivec2 p) {
		return level[p.y][p.x];
	}

	void parse(int lev) {
		std::string filename = "src/assets/levels/level";
		filename += std::to_string(lev);
		filename += ".txt";

		std::ifstream file(filename);

		if (!file.is_open()) {
			// std::cerr << "Error opening file: " << filename << std::endl;
			finished_game = true;
			// exit(EXIT_FAILURE);
		} else {
			torches.resize(0);
			this->level.resize(0);

			std::string line;

			width = 0;
			height = 0;
			// int height = 0;
			// int width = 0;

			int idx_h = 0;
			// Read the file line by line
			while (std::getline(file, line)) {
				std::vector<Cell> row;

				// Process each character in the line
				int idx_w = 0;
				for (char c : line) {
					if (c != '\n') {
						glm::ivec2 p(idx_w, idx_h);
						if(c == 'D') {
							exit_door_pos = p;
						}
						if(c == 'o') {
							player_start_pos = p;
							row.push_back(Cell(CellType(' ')));
						} else if (c == 't') {
							torches.push_back(p);
							row.push_back(Cell(CellType(' ')));
						} else if (c == 'B' || c == 'P' || c == 'H' || c == 'V') {
							EnemyType enemy_type = EnemyType(c);
							Cell cell(CellType('B'));
							cell.enemy_type = enemy_type;
							row.push_back(cell);
						}else {
							row.push_back(Cell(CellType(c)));
						}
					}
					idx_w++;
				}

				level.push_back(row);

				// Update width if needed
				if (row.size() > width) {
					width = row.size();
				}

				idx_h++;
				height++;
			}


			if(file.is_open()) {
				file.close();
			}

			std::cout << "Width: " << width << std::endl;
			std::cout << "Height: " << height << std::endl;
			bfs.resize(this);
		}
	}
};



enum class EngineScene {
	SceneTracer,
	SceneClouds,
	SceneReSTIR,
	SceneRevision2024,
	SceneRevision2025,
	SceneBruh,
	SceneRevision2024B,
	SceneFlameDOF,
	SceneGGJ,
	Scene
};

struct Scene {
	virtual void init() = 0;
	virtual void loop() = 0;
};

#include "scenes/_SceneVars.h"

class Engine {
	private:
	void update_time();
	void update_shared_uniforms();
	void handle_winapi_start_frame();

public:
	std::vector<Framebuffer*> framebuffers;
	// WArray<Texture*>* textures;

	bool fullscreen = false;
	bool borderless = false;

	Camera* camera;

	uint32_t RESX;
	uint32_t RESY;
	
	uint32_t desktop_fullscreen_res_x;
	uint32_t desktop_fullscreen_res_y;

	AudioMixer* audio_mixer;
	Physics physics;

	float ambient_light_intensity = 1.0;

	WGL* wgl;
  Framebuffer* default_fb;
	ShaderManager* shader_manager;
	DebugDrawer* debug_drawer;
	ShaderProgram* prog_clean_dbg;
	ShaderProgram* prog_gif;

	TextDrawer* text_drawer = nullptr;
	Drawer3D* drawer_3d = nullptr;
	Drawer2D* drawer_2d = nullptr;
	DrawerMDI* drawer_mdi = nullptr;

	WAlloc<Model, 10>& alloc_scenes = WAlloc<Model, 10>::get_instance();
	WAlloc<Texture, 500>& alloc_textures = WAlloc<Texture, 500>::get_instance();
	WAlloc<Light, 10000>& alloc_lights = WAlloc<Light, 10000>::get_instance();
	WAlloc<WEntity, 10000>& alloc_game_object = WAlloc<WEntity, 10000>::get_instance();

	std::vector<Buffer*> buffers;

	Model* scene_model = nullptr;


	IO* io = nullptr;
	Editor* editor = nullptr;

	Shader* passthrough_frag_shader;
	Shader* shadowmap_geom_shader;
	Shader* vert_sh_fullscreen_quad;
	Buffer* shared_ssbo;
	Buffer* lights_ssbo;


	bool to_reset_whole_engine = false;

	std::unordered_map<JPH::BodyID, WEntity*> body_to_game_object;
	std::unordered_map<long, WEntity*> entitties;
	std::vector<WEntity*> objects_to_destroy;
	// std::vector<Framebuffer*> framebuffers_to_resize;
	std::vector<Framebuffer*> shadow_map_textures;
	// WArray<GameObject*> collided_game_objects;
	std::vector<WCollision> collisions;
	std::vector<Light> lights;

	std::atomic<bool> scene_reloaded;


	std::mutex collisions_mutex;

	long gam_obj_id_cnt = 0;

	bool initializing = true;

	long timer;
	long timerStart;
	double time = 0.0;
	double delta_time = 0.0;
	double delta_time_internal = 0.0;
	long frame = 0;
	
  HMODULE giflib_instance = nullptr;
  HMODULE libvlc_instance = nullptr;

	Engine();
	void endFrame();

	void initialize();
	void startFrame();
	void start_timer_query(const char* name);


	void init_game(Vars& v);

	void init_tracer(VarsCloudTracer& v);
	void init_bruh(VarsBruh& v);
	void run();

	void init_ggj(VarsGGJ& v);
	// void run_ggj();
	// void eloop_revision_2025(VarsRevision2025& v);
	void eloop_bruh(VarsBruh& v);
	void eloop_revision_2024b(VarsRevision2024B& v);
	// void eloop_revision_2024(VarsRevision2024& v);
	// void eloop_restir(VarsReSTIR& v);
	void eloop_ggj(VarsGGJ& v);
	void eloop_flame_dof(VarsFlameDOF& v);
	void eloop(Vars& v);
	void eloop_cloud_tracer(VarsCloudTracer& v);


	void reset();
	void destroy_entity(WEntity* entity);

};
