﻿#include "pch.h"
#include "Text.h"

#include <codecvt>

#include "engine/Engine.h"
#include "gl/Shader.h"

#include <Shlwapi.h>
#include <shellapi.h>


Font::Font(const std::string& font_path) {
		const std::string root_fonts_dir = ".\\src\\assets\\msdf\\";
		// const std::string font_name = "ferritecore.otf";
		const std::string font_name = font_path;
		const std::string msdf_atlas_gen_path = root_fonts_dir + "msdf-atlas-gen.exe";
		const std::string font_path_full = root_fonts_dir + font_name;
		const std::string charset_path = root_fonts_dir + "charset.txt";
		const std::string out_image_path = font_path_full.substr(0, font_path_full.find_last_of('.')) + ".png";
		const std::string out_json_path = font_path_full.substr(0, font_path_full.find_last_of('.')) + ".json";

		this->load_from_msdf(font_path);


		this->shader_prog = new ShaderProgram({ new Shader("text.vert"), new Shader("text.frag") });


		this->shader_prog->setUniform("font_tex", this->atlas_tex);
	}

	void Font::load_from_msdf(const std::string& font_path) {
		const std::string root_fonts_dir = "src\\assets\\msdf\\";
		const std::string font_name = font_path;
		std::string msdf_atlas_gen_path = std::string("\\") + root_fonts_dir + "msdf-atlas-gen.exe";
		const std::string font_path_full = root_fonts_dir + font_name;
		const std::string charset_path = root_fonts_dir + "charset.txt";
		const std::string out_image_path = font_path_full.substr(0, font_path_full.find_last_of('.')) + ".png";
		const std::string out_json_path = font_path_full.substr(0, font_path_full.find_last_of('.')) + ".json";

		std::string currentPath = std::filesystem::current_path().string();

		msdf_atlas_gen_path = currentPath + msdf_atlas_gen_path;
	
		std::string cmd_line =
			std::string("\"") + msdf_atlas_gen_path + std::string("\"")
            + " -font " + std::string("\"") + font_path_full + std::string("\"") +
			" -imageout " + std::string("\"")  + out_image_path + std::string("\"")  +
			std::string(" -dimensions 1024 1024") +
			" -json " + std::string("\"")  + out_json_path + std::string("\"")  +
			" -type msdf";

		// int result = std::system(cmd_line.c_str());
		// result = std::system("cd");
		SHELLEXECUTEINFO sei = { sizeof(SHELLEXECUTEINFO) };
		sei.lpFile = "cmd.exe";
		sei.lpParameters = (cmd_line).c_str();
		sei.nShow = SW_HIDE;  // Use SW_SHOW to display the command prompt

		SECURITY_ATTRIBUTES saAttr;
		saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
		saAttr.bInheritHandle = TRUE;
		saAttr.lpSecurityDescriptor = nullptr;
	
		HANDLE hReadPipe, hWritePipe;

		if (!CreatePipe(&hReadPipe, &hWritePipe, &saAttr, 0)) {
			std::cerr << "Error creating pipe." << std::endl;
			// return 1;
		}
	
		STARTUPINFO si = { sizeof(STARTUPINFO) };
		si.dwFlags = STARTF_USESTDHANDLES;
		si.hStdOutput = hWritePipe;

		PROCESS_INFORMATION pi;

	// 	std::to_wstring(cmd_line.c_str());
		// std::wstring_convert<std::codecvt_utf8<wchar_t>> converter;
		// std::wstring wideString = converter.from_bytes(cmd_line);

		// std::wstring cmd_line_wstr = cmd_line.w;
		if (CreateProcess(nullptr, const_cast<LPSTR>(cmd_line.c_str()), nullptr, nullptr, TRUE, 0, nullptr, nullptr, &si, &pi)) {
		// if (CreateProcess(nullptr, cmd_line.c_str(), nullptr, nullptr, TRUE, 0, nullptr, nullptr, &si, &pi)) {
			// Close the write end of the pipe as it's not needed in this process
			CloseHandle(hWritePipe);

			// Read the output from the pipe
			char buffer[4096];
			DWORD bytesRead;
			while (ReadFile(hReadPipe, buffer, sizeof(buffer), &bytesRead, nullptr) && bytesRead > 0) {
				std::cout.write(buffer, bytesRead);
			}

			// Wait for the child process to finish
			WaitForSingleObject(pi.hProcess, INFINITE);

			// Close process and thread handles
			CloseHandle(pi.hProcess);
			CloseHandle(pi.hThread);
		} else {
			std::cerr << "Error creating process." << std::endl;
			// return 1;
		}
		// if (ShellExecuteEx(&sei)) {
		// 	// The command was launched successfully
		// 	WaitForSingleObject(sei.hProcess, INFINITE);  // Wait for the command to finish
		// 	CloseHandle(sei.hProcess);  // Close the process handle
		// } else {
		// 	std::cerr << "Error launching the command." << std::endl;
		// 	// return 1;
		// }

		// if (result != 0) {
		// 	// throw std::runtime_error("Command execution failed");
		// 	wlog_err("ERROR READING/WRITING FONT");
		// }

		/* load font file */
		std::ifstream fontFile(font_path_full.c_str(), std::ios::binary);

		if (!fontFile.is_open()) {
			wlog_err("Error opening font file.");
		}

		fontFile.seekg(0, std::ios::end);
		long file_size = (long)fontFile.tellg(); /* how long is the file ? */
		fontFile.seekg(0, std::ios::beg); /* reset */

		std::vector<unsigned char> fontBuffer(file_size);

		if (!fontFile.read(reinterpret_cast<char*>(fontBuffer.data()), file_size)) {
			wlog_err("Error reading font file.");
		}
		/* prepare font */
		stbtt_fontinfo info;
		if (!stbtt_InitFont(&info, fontBuffer.data(), 0)) {
			wlog_err("Failed to initialize font.");
		}

		// float stb_scale = stbtt_ScaleForMappingEmToPixels(&info,);

		// this->metadata_buff = new Buffer(&w_glyphs[0], this->w_glyphs.size() * sizeof(WGlyph));
		this->metadata_buff = new Buffer(
			BuffDesc{
			.byte_len = (int)this->w_glyphs.size() * (int)sizeof(WGlyph),
				.upload_data = &w_glyphs[0],
				.immutable_subdata = false,
				.name = "Buff text metadata"
			}
		);

		// this->atlas_tex = new Texture(out_image_path);
		this->atlas_tex = WEngine->alloc_textures.push(out_image_path).ptr;


		glm::vec2 canvas_sz = glm::vec2(
			float(this->atlas_tex->resx),
			float(this->atlas_tex->resy)
		);

		std::ifstream json_file(out_json_path);
		json json_data;
		json_file >> json_data;
		json_file.close();




		this->ascender = json_data["metrics"]["ascender"].get<float>();
		this->descender = json_data["metrics"]["descender"].get<float>();
		this->em_size = json_data["metrics"]["emSize"].get<float>();
		this->line_height = json_data["metrics"]["lineHeight"].get<float>();
		this->distance_range = json_data["atlas"]["distanceRange"].get<float>();
		this->size = json_data["atlas"]["size"].get<float>();

		// float stb_scale = stbtt_ScaleForPixelHeight(&info, this->size);

		// for(auto sg: json_data.items()) {
		// 	std::cout << sg.key() << std::endl;
		// 	std::cout << sg.value() << std::endl;
		// }
		for (const auto& glyph : json_data["glyphs"]) {
			uint32_t unicode = glyph["unicode"].get<uint32_t>();

			float advance = glyph["advance"].get<float>();
			float planeBoundsleft = 0.0;
			float planeBoundsbottom = 0.0;
			float planeBoundsright = 0.0;
			float planeBoundstop = 0.0;
			float atlasBoundsleft = 0.0;
			float atlasBoundsbottom = 0.0;
			float atlasBoundsright = 0.0;
			float atlasBoundstop = 0.0;

			// for(auto sg: glyph.items()) {
			// 	std::cout << sg.key() << std::endl;
			// 	std::cout << sg.value() << std::endl;
			// }

			if(glyph.contains("planeBounds")) {
				if(glyph.contains("atlasBounds")) {
					// glyph.items()
					// for(auto sg : glyph) {
					// 	// sg.type_name()
					// 	sg.key
					// 	std::cout << sg << std::endl;
					// }
					// glyph.count()
					planeBoundsleft = glyph["planeBounds"]["left"].get<float>();
					planeBoundsbottom = glyph["planeBounds"]["bottom"].get<float>();
					planeBoundsright = glyph["planeBounds"]["right"].get<float>();
					planeBoundstop = glyph["planeBounds"]["top"].get<float>();
					atlasBoundsleft = glyph["atlasBounds"]["left"].get<float>() / canvas_sz.x;
					atlasBoundsbottom = glyph["atlasBounds"]["bottom"].get<float>() / canvas_sz.y;
					atlasBoundsright = glyph["atlasBounds"]["right"].get<float>() / canvas_sz.x;
					atlasBoundstop = glyph["atlasBounds"]["top"].get<float>() / canvas_sz.y;
					// int unicode_idx = glyph["unicode"].get<uint32_t>();
				}
			}
			this->w_glyphs[unicode] = {
					unicode , advance , planeBoundsleft , planeBoundsbottom ,
					planeBoundsright , planeBoundstop , atlasBoundsleft , atlasBoundsbottom ,
					atlasBoundsright , atlasBoundstop , {}
			};


			for(int k = 0; k < 255; k++) {
				int kern = stbtt_GetCodepointKernAdvance(&info, int(unicode), k);
				// float kern_f = float(kern) * stb_scale;
				// float kern_f = float(kern) * this->size;
				// float kern_f = float(kern) * this->em_size;
				// float kern_f = float(kern) / this->size;
				float kern_f = float(kern) / 32.0;
				kern_f *= 1.0/64.0;
				// this->w_glyphs[unicode].kern_advance[k] = kern_f/15.0f;
				this->w_glyphs[unicode].kern_advance[k] = kern_f;
				if(kern != 0) {
					// std::cout << "kern" << std::endl;
					// std::cout << kern_f << std::endl;
					// std::cout << "advance" << std::endl;
					// std::cout << advance << std::endl;
				}
			}
		}
	}


	TextDrawer::TextDrawer (Font* font): font(font){
		const size_t buff_len = 10000;
		this->text_buff = new Buffer({.byte_len = buff_len * sizeof(float), .immutable_subdata = false, .name = "Buff TextDrawer"});
		this->buff_ptr = (float*)this->text_buff->cpu_data;
	}

	void TextDrawer::flush() {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_DEPTH_TEST);

		// int char_offs = this->char_cnt_accum;

		this->font->shader_prog->use();
		this->font->shader_prog->setUniform("verts", this->text_buff);
		this->font->shader_prog->setUniform("font_tex", this->font->atlas_tex);

		// 6 verts per char
		// 16 floats * 4 bytes per char

		constexpr int verts_per_char = 6;
		constexpr int floats_per_vert = 16;
		constexpr int floats_per_char = verts_per_char * floats_per_vert;
		constexpr int bytes_per_vert = floats_per_vert * 4;

		// const int c
		int vert_cnt_accum = verts_per_char * this->char_cnt_accum;
		int vert_cnt_curr = verts_per_char * this->char_cnt;
		int byte_cnt_curr = vert_cnt_curr * bytes_per_vert;

		int float_offs_from_start = this->buff_ptr - (float*)this->text_buff->cpu_data;
		float_offs_from_start -= this->char_cnt * floats_per_char;
		int byte_offs_from_start = float_offs_from_start * 4;

		float* curr_buff_ptr = this->buff_ptr - vert_cnt_curr * floats_per_vert;

		this->text_buff->upload_sub_data(
			curr_buff_ptr, byte_cnt_curr, byte_offs_from_start
		);

		glDrawArrays(GL_TRIANGLES, vert_cnt_accum, vert_cnt_curr);
		glDisable(GL_BLEND);

		this->char_cnt_accum += this->char_cnt;
		this->char_cnt = 0;
	}

void TextDrawer::reset_end_of_frame() {
	this->char_cnt_accum = 0;
	this->char_cnt = 0;
	this->buff_ptr = (float*)this->text_buff->cpu_data;
}

void TextDrawer::draw(TextDescriptor& desc) {
		glm::vec2 starting_offs = desc.pos;
		glm::vec2 offs = starting_offs;
		float fs_scale = 1./ (font->ascender - font->descender);

		float space_advance = font->w_glyphs[int(' ')].advance * desc.whitespace_scale;

		float vertical_step = fs_scale * font->line_height * desc.sz;

		float aspect = float(WEngine->RESX)/float(WEngine->RESY);
		const float advance_mult = desc.sz * desc.kern_scale * 1.0f/aspect;
		const glm::vec2 quad_mult = glm::vec2(1.0f/aspect, 1.0f) * desc.sz * fs_scale;

		auto get_advance = [&](int a, int b) -> float {
			if(b >= 255 || a >= 255){
				return 0.0;
			}
			float advance = font->w_glyphs[a].advance;
			if(b < 0) {
				return advance;
			}
			float kern = font->w_glyphs[a].kern_advance[b];
			if(!desc.do_kerning) {
				kern = 0.0f;
			}
			return (advance + kern) * advance_mult;
		};

		std::array<float,50> row_widths = { };
		// if(this->centered) {
		float max_row_w = 0.0;
		{
			float curr_row_w = 0.0;
			// float total_height = vertical_step;
			float total_height = 0.0f;
			// float acc_offs = 0.0;
			int i = 0;
			int row = 0;
			for(auto ch : desc.text) {
				WGlyph glyph = font->w_glyphs[int(ch)];

				if(ch == '\r') {
				} else if(ch == '\n') {
					row_widths[row] = curr_row_w;
					total_height += vertical_step;
					if(curr_row_w > max_row_w) {
						max_row_w = curr_row_w;
					}
					curr_row_w = 0.0;
					row++;
				} else if(ch == ' ') {
					float advance = space_advance;
					curr_row_w += fs_scale * advance * desc.sz * 1.0f/aspect;
				} else {
						if(i < (int)desc.text.size() - 1) {
							curr_row_w += get_advance(int(desc.text[i]), int(desc.text[i + 1]));
						} else {
							curr_row_w += get_advance(int(desc.text[i]), -1)*0.1f;
						}
				}

				if(i == (int)desc.text.size() - 1) {
					row_widths[row] = curr_row_w;
					total_height -= vertical_step*0.5f;
				}
				i++;
			}
			if(curr_row_w > max_row_w) {
				max_row_w = curr_row_w;
			}

			if(desc.anchor != TextAnchor::TOP_LEFT) {
				glm::vec2 accum_offs = glm::vec2( max_row_w, total_height );

				glm::vec2 anchor_offs = glm::vec2(0.0f,0.0f);
				switch (desc.anchor) {
					case TextAnchor::CENTER:
						anchor_offs = glm::vec2(0,0);
						break;
					case TextAnchor::TOP:
						anchor_offs  = glm::vec2(0,1);
						break;
					case TextAnchor::TOP_RIGHT:
						anchor_offs  = glm::vec2(1,1);
						break;
					case TextAnchor::TOP_LEFT:
						anchor_offs  = glm::vec2(-1,1);
						break;
					case TextAnchor::BOT:
						anchor_offs  = glm::vec2(0,-1);
						break;
					case TextAnchor::BOT_RIGHT:
						anchor_offs  = glm::vec2(1,-1);
						break;
					case TextAnchor::BOT_LEFT:
						anchor_offs  = glm::vec2(-1,-1);
						break;
					case TextAnchor::RIGHT:
						anchor_offs  = glm::vec2(1,0);
						break;
					case TextAnchor::LEFT:
						anchor_offs = glm::vec2(-1,0);
						break;
				}
				anchor_offs *= glm::vec2(0.5f,-0.5f);
				anchor_offs += 0.5f;

				starting_offs.x -= accum_offs.x * anchor_offs.x;
				starting_offs.y += accum_offs.y * anchor_offs.y;
				// starting_offs.x -= max_row_w * 0.5;
				offs.x = starting_offs.x;
				offs.y = starting_offs.y;
			}
		}


		int i = -1;
		int chars = 0;
		int row = 0;

		auto center = [&](int _row) {
			offs.x += max_row_w*0.5;
			offs.x -= row_widths[_row]*0.5f;
		};

		if(desc.centered) {
			center(0);
		}
		for(auto ch : desc.text) {
			i++;
			WGlyph glyph = font->w_glyphs[int(ch)];

			// glm::vec2 uv_coords = glm::vec2(
			// 	glyph.atlas_bound_right - glyph.atlas_bound_left,
			// 	glyph.atlas_bound_top - glyph.atlas_bound_bot
			// )/canvas_sz;

			if(ch == '\r') {
			} else if(ch == '\n') {
				row++;
				offs.x = starting_offs.x;
				if(desc.centered) {
					center(row);
				}
				offs.y -= vertical_step;
			} else if(ch == ' ') {
				float advance = space_advance;
				if(i < desc.text.size() -1) {
					// !!!! TODO: Advance

				}
				offs.x += fs_scale * advance * desc.sz * 1.0f/aspect;
			} else {
				glm::vec2 bot_l_uv = glm::vec2(glyph.atlas_bound_left, glyph.atlas_bound_bot);
				glm::vec2 top_l_uv = glm::vec2(glyph.atlas_bound_left, glyph.atlas_bound_top);
				glm::vec2 bot_r_uv = glm::vec2(glyph.atlas_bound_right, glyph.atlas_bound_bot);
				glm::vec2 top_r_uv = glm::vec2(glyph.atlas_bound_right, glyph.atlas_bound_top);

				glm::vec2 bot_l_quad = glm::vec2(glyph.plane_bound_left, glyph.plane_bound_bot) * quad_mult + offs;
				glm::vec2 top_l_quad = glm::vec2(glyph.plane_bound_left, glyph.plane_bound_top) * quad_mult + offs;
				glm::vec2 bot_r_quad = glm::vec2(glyph.plane_bound_right, glyph.plane_bound_bot) * quad_mult + offs;
				glm::vec2 top_r_quad = glm::vec2(glyph.plane_bound_right, glyph.plane_bound_top) * quad_mult + offs;

				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.bg_col.x);
				*buff_ptr++ = float(desc.bg_col.y);
				*buff_ptr++ = float(desc.bg_col.z);
				*buff_ptr++ = float(desc.bg_col.w);
				*buff_ptr++ = bot_l_quad.x;
				*buff_ptr++ = bot_l_quad.y;
				*buff_ptr++ = bot_l_uv.x;
				*buff_ptr++ = bot_l_uv.y;
				*buff_ptr++ = desc.col.r;
				*buff_ptr++ = desc.col.g;
				*buff_ptr++ = desc.col.b;
				*buff_ptr++ = desc.col.w;

				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.bg_col.x);
				*buff_ptr++ = float(desc.bg_col.y);
				*buff_ptr++ = float(desc.bg_col.z);
				*buff_ptr++ = float(desc.bg_col.w);
				*buff_ptr++ = bot_r_quad.x;
				*buff_ptr++ = bot_r_quad.y;
				*buff_ptr++ = bot_r_uv.x;
				*buff_ptr++ = bot_r_uv.y;
				*buff_ptr++ = desc.col.r;
				*buff_ptr++ = desc.col.g;
				*buff_ptr++ = desc.col.b;
				*buff_ptr++ = desc.col.w;

				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.bg_col.x);
				*buff_ptr++ = float(desc.bg_col.y);
				*buff_ptr++ = float(desc.bg_col.z);
				*buff_ptr++ = float(desc.bg_col.w);
				*buff_ptr++ = top_r_quad.x;
				*buff_ptr++ = top_r_quad.y;
				*buff_ptr++ = top_r_uv.x;
				*buff_ptr++ = top_r_uv.y;
				*buff_ptr++ = desc.col.r;
				*buff_ptr++ = desc.col.g;
				*buff_ptr++ = desc.col.b;
				*buff_ptr++ = desc.col.w;

				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.bg_col.x);
				*buff_ptr++ = float(desc.bg_col.y);
				*buff_ptr++ = float(desc.bg_col.z);
				*buff_ptr++ = float(desc.bg_col.w);
				*buff_ptr++ = bot_l_quad.x;
				*buff_ptr++ = bot_l_quad.y;
				*buff_ptr++ = bot_l_uv.x;
				*buff_ptr++ = bot_l_uv.y;
				*buff_ptr++ = desc.col.r;
				*buff_ptr++ = desc.col.g;
				*buff_ptr++ = desc.col.b;
				*buff_ptr++ = desc.col.w;

				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.bg_col.x);
				*buff_ptr++ = float(desc.bg_col.y);
				*buff_ptr++ = float(desc.bg_col.z);
				*buff_ptr++ = float(desc.bg_col.w);
				*buff_ptr++ = top_r_quad.x;
				*buff_ptr++ = top_r_quad.y;
				*buff_ptr++ = top_r_uv.x;
				*buff_ptr++ = top_r_uv.y;
				*buff_ptr++ = desc.col.r;
				*buff_ptr++ = desc.col.g;
				*buff_ptr++ = desc.col.b;
				*buff_ptr++ = desc.col.w;

				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.outline);
				*buff_ptr++ = float(desc.bg_col.x);
				*buff_ptr++ = float(desc.bg_col.y);
				*buff_ptr++ = float(desc.bg_col.z);
				*buff_ptr++ = float(desc.bg_col.w);
				*buff_ptr++ = top_l_quad.x;
				*buff_ptr++ = top_l_quad.y;
				*buff_ptr++ = top_l_uv.x;
				*buff_ptr++ = top_l_uv.y;
				*buff_ptr++ = desc.col.r;
				*buff_ptr++ = desc.col.g;
				*buff_ptr++ = desc.col.b;
				*buff_ptr++ = desc.col.w;

				chars++;
				if(i < desc.text.size() - 1) {
					offs.x += get_advance(int(desc.text[i]), int(desc.text[i + 1]));
				}
			}
		}
		// for(int i = 0; i < )



		this->char_cnt += chars;
	}

