﻿#pragma once
#include <engine/Drawer2D.h>
#include <gl/Buffer.h>
#include <gl/ShaderProgram.h>
#include "stb_truetype.h"

struct Buffer;

namespace fs = std::filesystem;
using json = nlohmann::json;


struct WGlyph {
	uint32_t unicode;
	float advance;
	float plane_bound_left;
	float plane_bound_bot;
	float plane_bound_right;
	float plane_bound_top;
	float atlas_bound_left;
	float atlas_bound_bot;
	float atlas_bound_right;
	float atlas_bound_top;
	std::array<float, 255> kern_advance;
};

using WGlyphArray = std::array<WGlyph, 255>;

class Font {
public:
	WGlyphArray w_glyphs;
	Texture* atlas_tex = nullptr;
	Buffer* metadata_buff = nullptr;

	ShaderProgram* shader_prog = nullptr;

	float ascender = 0.0;
	float descender = 0.0;
	float em_size = 0.0;
	float line_height = 0.0;
	float distance_range = 0.0;
	float size = 0.0;

	Font(const std::string& font_path);
private:
	void load_from_msdf(const std::string& font_path);
};


// struct MDSFVert {
//
// };

// struct TexAnchorType {
// 	int values[2];
// };
enum class TextAnchor {
	CENTER ,
	TOP ,
	TOP_RIGHT ,
	TOP_LEFT ,
	BOT ,
	BOT_RIGHT ,
	BOT_LEFT ,
	RIGHT ,
	LEFT
};

struct UIElem {
	// glm::vec2 pos = glm::vec2(0);
	RectDescr rect;

	bool mouse_over = false;
	bool mouse_clicked = false;
	bool mouse_down = false;
	bool mouse_unclicked = false;

	std::function<void(UIElem*)> render;
	std::function<void(UIElem*)> on_mouse_not_over;
	std::function<void(UIElem*)> on_mouse_over;
	std::function<void(UIElem*)> on_mouse_clicked;
	std::function<void(UIElem*)> on_mouse_down;
	std::function<void(UIElem*)> on_mouse_unclicked;
};

struct TextDescriptor {
	glm::vec2 pos = glm::vec2(0);
	glm::vec4 col = glm::vec4(1.0);
	float sz = 0.194;
	float kern_scale = 0.722;
	float whitespace_scale = 2.667;
	bool do_kerning = true;
	bool outline = true;
	glm::vec4 bg_col = glm::vec4(0);
	TextAnchor anchor = TextAnchor::CENTER;
	bool centered = false;
	SIMDString<512> text ;
};

class TextDrawer {
public:
	Font* font;
	Buffer* text_buff;
	float* buff_ptr;

	int char_cnt = 0;
	int char_cnt_accum = 0;

	TextDrawer (Font* font);

	inline void update_verts() { }

	void flush();
	void reset_end_of_frame();

	void draw(TextDescriptor& desc);
};

// class Text {
// 	private:
// 	void fill_cpu_text_buff(const std::string& text) {
// 		for(int i = 0; i < 100 * 100; i++) {
// 				text_buff_arr[i] = 0;
// 		}
// 		int row = 0;
// 		int col = 0;
// 		for(char ch : text) {
// 			if(ch == '\n') {
// 				row++;
// 				col = 0;
// 			} else {
// 				uint32_t idx = row * 100 + col;
// 				text_buff_arr[idx] = ch;
// 				col++;
// 			}
// 		}
// 	}
// 	public:
// 	uint32_t* text_buff_arr;
// 	Buffer* text_buff;
// 	ShaderProgram* prog_quad_text;
//
// 	int text_w = 100;
// 	int text_h = 10;
//
// 	Text(std::string text, ShaderProgram* prog_quad_text): prog_quad_text(prog_quad_text) {
// 		this->text_buff_arr = new uint32_t[100 * 100];
//
// 		fill_cpu_text_buff(text);
// 		this->text_buff = new Buffer(text_buff_arr, 100 * 100 * sizeof(uint32_t));
// 	}
// 	void replace_text(const std::string& text) {
// 		fill_cpu_text_buff(text);
// 		this->text_buff->upload_smth(
// 			this->text_buff_arr,
// 			text.length() * sizeof(uint32_t)
// 		);
// 	}
// 	void draw(
// 		glm::vec2 pos = glm::vec2(0.,0.),
// 		glm::vec2 sz = glm::vec2(0.5,0.5),
// 		float rot = 0.0
// 		) {
// 			prog_quad_text->use();
// 			prog_quad_text->setUniform("text_buff", text_buff);
// 			prog_quad_text->setUniform("pos", pos);
// 			prog_quad_text->setUniform("sz", sz);
// 			prog_quad_text->setUniform("text_w", text_w);
// 			prog_quad_text->setUniform("text_h", text_h);
// 			prog_quad_text->setUniform("rot_amt", rot);
// 			glDrawArrays(GL_TRIANGLES, 0, 6);
// 	}
//
// };
