#include "pch.h"

#include "Buffer.h"

#include "engine/Engine.h"


Buffer::Buffer( BuffDesc desc ) {
	// for(int byte_len)
	// char8_t* data_byte_ptr = (char8_t*)data;
	// for(int i = 0; i < byte_len; i++) {
	// 	*data_byte_ptr++ = 0;
	// }
	this->byte_len = desc.byte_len;
	// this->elem_cnt = byte_len / 4;
	this->elem_cnt = byte_len / 4; // TODO: useless

	WEngine->buffers.push_back(this);

	glCreateBuffers(1, &this->pid);


	this->cpu_data = new char8_t[byte_len];

	if(desc.upload_data.has_value()) {
		memcpy(this->cpu_data, desc.upload_data.value(), byte_len);
	} else {
		for(int i = 0; i < this->byte_len; i++) {
			this->cpu_data[i] = 0;
		}
	}

	// uint32_t flags = GL_MAP_WRITE_BIT | GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT;
	// uint32_t flags = GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT;
	// uint32_t flags = GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT;
	// uint32_t flags = GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT;
	// uint32_t flags = GL_MAP_WRITE_BIT;
	uint32_t flags = 0;

	if(desc.mappable_write) {
		flags |= GL_MAP_WRITE_BIT;
	} else if(!desc.immutable_subdata) {
		flags |= GL_DYNAMIC_STORAGE_BIT;
	}
	if(desc.mappable_read) {
		flags |= GL_MAP_READ_BIT;
	}

	if(desc.persistennt) {
		flags |= GL_MAP_PERSISTENT_BIT;
	}

	void* data_to_upload = desc.upload_data.has_value() ? desc.upload_data.value() : this->cpu_data;

	glNamedBufferStorage(this->pid, byte_len, data_to_upload, flags);

	glGetNamedBufferParameterui64vNV(this->pid,GL_BUFFER_GPU_ADDRESS_NV, &this->bindless_address);

	GLenum residency_access;
	if(desc.read_from_shader && desc.write_from_shader) {
		residency_access = GL_READ_WRITE;
	} else if (desc.write_from_shader) {
		residency_access = GL_WRITE_ONLY;
	} else {
		residency_access = GL_READ_ONLY;
	}

	glMakeNamedBufferResidentNV(this->pid,residency_access);

	if(desc.name.length() > 0) {
		this->set_name(desc.name);
	}

	// this->mapped_buff = glMapNamedBufferRange(this->pid, 0, this->byte_len, flags);
}



void Buffer::bind_as_SSBO(int idx) {
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, idx, this->pid);
}

void Buffer::bind_as_MDI_buff() {
	glBindBuffer(GL_DRAW_INDIRECT_BUFFER, this->pid);
}

void Buffer::bind_as_indirect_compute_buff() {
	glBindBuffer(GL_DISPATCH_INDIRECT_BUFFER, this->pid);
}

void Buffer::bind_as_MDI_PARAM_buff() {
	glBindBuffer(GL_PARAMETER_BUFFER, this->pid);
}

void Buffer::upload_sub_data(void* data, int byte_len, int byte_offs) {
	glNamedBufferSubData(this->pid,byte_offs, byte_len, data);
}

void Buffer::set_name(std::string_view _name) {
	this->name = _name;
	glObjectLabel( GL_BUFFER, this->pid, -1, &this->name[0]);
}

void Buffer::flush() {
	// glFlushMappedNamedBufferRange(this->pid,0,this->byte_len);
}


void Buffer::clear() {
	uint8_t val[] = {0,0,0};
	glClearNamedBufferSubData( this->pid, GL_R8I, 0, this->byte_len, GL_RED_INTEGER, GL_BYTE, &val[0] );
}

void Buffer::destroy() {
	glUnmapNamedBuffer(this->pid);

	// delete this->cpu_data; // NOOOOOOOOOOOOOOOOOOO PLS NO uncomment ._.

	glMakeNamedBufferNonResidentNV(this->pid);
	unsigned buffers = {this->pid};
	glDeleteBuffers(1, &buffers);

	// WEngine->buffers.erase()
	WEngine->buffers.erase(std::remove(WEngine->buffers.begin(), WEngine->buffers.end(), this), WEngine->buffers.end());
}
