#include "pch.h"
#include "Framebuffer.h"

#include <print>
#include "engine/Engine.h"

Framebuffer::Framebuffer(
	FBDesc fb_desc
)
	: pongable(fb_desc._pongable), autopong(fb_desc.autopong), name(fb_desc.name),
	textures(fb_desc.textures), depth_texture(fb_desc.depth_texture), fb_desc(fb_desc) {

	this->create();
}

void Framebuffer::create() {
	const bool is_first_init = this->pid == 1<<20;

	if(is_first_init) {
		if (fb_desc.name.empty()) {
			fb_desc.name = "FB Unnamed";
		}

		int i = 0;
		for (auto tex : this->textures) {
			tex->set_name(this->name + " tex " + std::to_string(i++));
		}
	}

	wlog_info("Init framebuffer: ", this->name.c_str());
	if (fb_desc.layered) {
		for (int i = 0; i < 6; i++) {
			uint32_t local_pid;
			glCreateFramebuffers(1, &local_pid);

			if (this->textures.size() > 0) {
				uint32_t col_buffs[10];

				int col_attach_idx = 0;
				for (auto& tex : this->textures) {
					glNamedFramebufferTexture(local_pid, GL_COLOR_ATTACHMENT0 + col_attach_idx, tex->pid, 0);
					col_buffs[col_attach_idx] = GL_COLOR_ATTACHMENT0 + col_attach_idx;
					// std::println("amogus ");
					// this->arr[i] = elem;
					col_attach_idx++;
				}
				if (col_attach_idx > 0) {
					glNamedFramebufferDrawBuffers(local_pid, col_attach_idx, col_buffs);
				}
			}
			if (this->depth_texture) {
				glNamedFramebufferTextureLayer(
					local_pid,
					GL_DEPTH_ATTACHMENT,
					depth_texture->pid,
					0,
					i
				);
			}
			// this->pids_layered.push_back(local_pid);
			this->pids_layered.push_back(local_pid);
		}
		this->pid = this->generate_framebuffer(textures);
	} else {
		this->pid = this->generate_framebuffer(textures);
		if (this->pongable) {
			if(is_first_init) {
				this->textures_back = WArray<Texture*>({}, this->textures.sz);
				int i = 0;
				for (auto tex : this->textures) {
					Texture* cloned_tex = tex->clone();
					this->textures_back.add(cloned_tex);
					cloned_tex->set_name(this->name + " backtex " + std::to_string(i++));
				}
			}
			this->pid_back = this->generate_framebuffer(this->textures_back);
		}
	}

	if (depth_texture != nullptr) {
		glNamedFramebufferTexture(
			this->pid,
			GL_DEPTH_ATTACHMENT,
			depth_texture->pid,
			0
		);
		if (pongable) {
			if(is_first_init) {
				this->depth_texture_back = this->depth_texture->clone();
			}
			glNamedFramebufferTexture(
				this->pid_back,
				GL_DEPTH_ATTACHMENT,
				this->depth_texture_back->pid,
				0
			);
		}

		#ifdef EDITOR
		uint32_t status = glCheckNamedFramebufferStatus(this->pid, GL_FRAMEBUFFER);

		if (status != GL_FRAMEBUFFER_COMPLETE) {
			std::cout << "---- INVALID FRAMEBUFFER \n";
		}
		#endif
	}


	for (auto t : this->textures) {
		t->bound_fb = this;
	}
	for (auto t : this->textures_back) {
		t->bound_fb = this;
	}
	if (this->depth_texture) {
		this->depth_texture->bound_fb = this;
	}
	if (this->depth_texture_back) {
		this->depth_texture_back->bound_fb = this;
	}

	if (this->pid != 0) {
		glObjectLabel(GL_FRAMEBUFFER, this->pid, -1, &this->name[0]);
	}
	if (this->pid_back != 0) {
		std::string backbuff_name = this->name;
		backbuff_name += "_backbuff";
		glObjectLabel(GL_FRAMEBUFFER, this->pid, -1, &backbuff_name[0]);
	}

	// Globals::engine->framebuffers->add(this);
	if(is_first_init) {
		Globals::engine->framebuffers.push_back(this);
	}
}

uint32_t Framebuffer::generate_framebuffer(WArray<Texture*>& input_textures) {
	uint32_t local_pid;
	glCreateFramebuffers(1, &local_pid);

	uint32_t buffs[10];

	int i = 0;
	for (auto& tex : input_textures) {
		glNamedFramebufferTexture(local_pid, GL_COLOR_ATTACHMENT0 + i, tex->pid, 0);
		buffs[i] = GL_COLOR_ATTACHMENT0 + i;
		// wlog_info("Create entity nocol: {}", this->game_obj_id);
		// std::println("amogus ");
		// this->arr[i] = elem;
		i++;
	}
	glNamedFramebufferDrawBuffers(local_pid, i, buffs);

	return local_pid;
}

WArray<Texture*>* Framebuffer::get_front_textures() {
	return pong_idx == 0 ? &this->textures : &this->textures_back;
}

Texture* Framebuffer::get_front_textures_at(int idx) {
	return (*this->get_front_textures())[idx];
}

Texture* Framebuffer::get_back_textures_at(int idx) {
	return (*this->get_back_textures())[idx];
}

WArray<Texture*>* Framebuffer::get_back_textures() {
	return pong_idx == 0 ? &this->textures_back : &this->textures;
}

Texture* Framebuffer::get_front_depth() {
	// return this->depth_texture;
	return pong_idx == 0 ? this->depth_texture : this->depth_texture_back;
}

Texture* Framebuffer::get_back_depth() {
	return pong_idx == 0 ? this->depth_texture_back : this->depth_texture;
}

uint32_t Framebuffer::get_front_fb() {
	return pong_idx == 0u ? this->pid : this->pid_back;
}

uint32_t Framebuffer::get_back_fb() {
	return pong_idx == 0 ? this->pid_back : this->pid;
}

void Framebuffer::pong() {
	this->pong_idx = 1 - this->pong_idx;
}

Framebuffer::Framebuffer() = default;

Framebuffer* Framebuffer::create_default_framebuffer() {
	Framebuffer* f = new Framebuffer();
	f->pid = 0;

	auto t = new Texture();
	t->pid = 0;
	t->resx = WEngine->RESX;
	t->resy = WEngine->RESY;

	f->textures = WArray({t});
	return f;
}

void Framebuffer::bind() {
	glBindFramebuffer(GL_FRAMEBUFFER, this->get_front_fb());
	if(this == WEngine->default_fb && WEngine->fullscreen) {
  		glViewport(0, 0, (int)WEngine->desktop_fullscreen_res_x, (int)WEngine->desktop_fullscreen_res_y);
	} else {
		if (this->textures.size() > 0) {
			glViewport(0, 0, (int)this->textures[0]->resx, (int)this->textures[0]->resy);
		} else {
			glViewport(0, 0, (int)this->depth_texture->resx, (int)this->depth_texture->resy);
		}
	}
}

void Framebuffer::bind_layer(int layer) {
	if (this->textures.size() > 0) {
		glViewport(0, 0, (int)this->textures[0]->resx, (int)this->textures[0]->resy);
	} else {
		glViewport(0, 0, (int)this->depth_texture->resx, (int)this->depth_texture->resy);
	}
	// auto d = this->pids_layered.data();
	// auto fbid = d[layer];
	uint32_t fbid = this->pids_layered[layer];
	glBindFramebuffer(GL_FRAMEBUFFER, fbid);
}

void Framebuffer::clear_layer(int layer, float r, float g, float b, float a) {
	float clear_col[] = {r, g, b, a};
	// glClearNamedFramebufferfv(this->pids_layered[layer], GL_COLOR, 0, clear_col);
	auto d = this->pids_layered.data();
	auto fbid = d[layer];
	if(this->textures.size() > 0) {
		glClearNamedFramebufferfv(fbid, GL_COLOR, 0, clear_col);
	}
	if(this->depth_texture) {
		glClearNamedFramebufferfi(fbid, GL_DEPTH_STENCIL, 0, 1, 0);
	}
}


void Framebuffer::clear_back(float r, float g, float b, float a) {
	int i = 0;
	float clear_col[] = {r, g, b, a};
	for (auto tex : *this->get_back_textures()) {
		glClearNamedFramebufferfv(this->get_back_fb(), GL_COLOR, i, clear_col);
		i++;
	}
	if (this->get_back_depth() != nullptr) {
		glClearNamedFramebufferfi(this->get_back_fb(), GL_DEPTH_STENCIL, 0, 1, 0);
		// glClearNamedFramebufferfv(this->get_front_fb(), GL_DEPTH_STENCIL, 0, clear_col);
	}
}

void Framebuffer::clear(float r, float g, float b, float a) {
	int i = 0;
	float clear_col[] = {r, g, b, a};
	for (auto tex : *this->get_front_textures()) {
		glClearNamedFramebufferfv(this->get_front_fb(), GL_COLOR, i, clear_col);
		i++;
	}
	if (this->get_front_depth() != nullptr) {
		glClearNamedFramebufferfi(this->get_front_fb(), GL_DEPTH_STENCIL, 0, 1, 0);
		// glClearNamedFramebufferfv(this->get_front_fb(), GL_DEPTH_STENCIL, 0, clear_col);
	}
}

void Framebuffer::blit(Framebuffer* other,
	int bitmask /*= GL_COLOR_BUFFER_BIT*/,
	int interp /*= GL_NEAREST*/) {
	int resx = (int)this->textures[0]->resx;
	int resy = (int)this->textures[0]->resy;
	glBlitNamedFramebuffer(this->get_front_fb(), other->pid, 0, 0, resx, resy, 0, 0, resx,
		resy, bitmask, interp);
}
