#pragma once

#ifdef EDITOR
	#include "FileWatch.hpp"
#endif

#include "Texture.h"
#include "engine/Engine.h"

struct FBDesc {
	WArray<Texture*> textures;
	bool _pongable = false;
	bool autopong = true;
	Texture* depth_texture = nullptr;
	bool layered = false;
	std::string name = "";
};

class Framebuffer {
	public:
	bool pongable = false;
	bool autopong = false;
	int pong_idx = 0;
	
	FBDesc fb_desc;
	uint32_t pid = 1<<20;
	uint32_t pid_back = 0;

	// uint32_t pid = 0;
	// WArray<uint32_t*> pids_layered;
	// smallvec<uint32_t, 6> pids_layered = smallvec<uint32_t, 6>();
	std::vector<uint32_t> pids_layered;

	std::string name = std::string("");

	// TODO: move Array<Texture> to heap
	WArray<Texture*> textures;
	WArray<Texture*> textures_back;

	Texture* depth_texture = nullptr;
	Texture* depth_texture_back = nullptr;

	WArray<Texture*>* get_front_textures();
	Texture* get_front_textures_at(int idx);
	Texture* get_back_textures_at(int idx);
	WArray<Texture*>* get_back_textures();

	Texture* get_front_depth();
	Texture* get_back_depth();
	uint32_t get_front_fb();
	uint32_t get_back_fb();

	void destroy() {
		for(auto pid : this->pids_layered) {
			glDeleteFramebuffers(1, &pid);
		}
		glDeleteFramebuffers(1, &this->pid);
		if(this->pid_back != 0) {
			glDeleteFramebuffers(1, &this->pid_back);
		}
		// WEngine->framebuffers.erase()
		int idx = 0;
		for(auto fb : WEngine->framebuffers) {
			if(fb == this) {
				break;
			}
			idx++;
		}
		WEngine->framebuffers.erase(WEngine->framebuffers.begin() + idx);
	}

	void pong();

	Framebuffer();
  static Framebuffer* create_default_framebuffer();
  Framebuffer(FBDesc fb_desc);

	void create();

  void bind();
	void bind_layer(int layer);
	void clear_layer(int layer, float r = 0, float g = 0, float b = 0, float a = 0);

	void clear(float r = 0, float g = 0, float b = 0, float a = 0);
	void clear_back(float r = 0, float g = 0, float b = 0, float a = 0);

  void blit(Framebuffer* other, int bitmask = GL_COLOR_BUFFER_BIT, int interp = GL_NEAREST);

	void resize(uint32_t resx, uint32_t resy) {
		for(Texture* tex : this->textures) {
			tex->resize(resx, resy);
		}
		for(Texture* tex : this->textures_back) {
			tex->resize(resx, resy);
		}
		
		for(auto pid : this->pids_layered) {
			glDeleteFramebuffers(1, &pid);
		}
		glDeleteFramebuffers(1, &this->pid);
		if(this->pid_back != 0) {
			glDeleteFramebuffers(1, &this->pid_back);
		}

		this->create();
		// if(this.)
		// this->textures
		// for()
		
	}
private:

	uint32_t generate_framebuffer(WArray<Texture*>& input_textures);
};
