#include "pch.h"
#include "Shader.h"
#include "ShaderManager.h"
#include "engine/Engine.h"


void Shader::create() {
	#ifdef EDITOR
	WEngine->shader_manager->shaders.add(this);
	#endif
	// this->error = new char[1000];
	this->error.reserve(1000);
	// this->shader_type = shader_type;
	// this->code = nullptr;
	this->compile();
}


Shader* Shader::clone_with_new_defines(WDumbMap<SIMDString<>, 32> defines_map) {
	return new Shader(this->path, defines_map);
}

Shader::Shader(std::string_view path, WDumbMap<SIMDString<>, 32> defines_map)
	: path(path), defines_map(defines_map)
// , shader_type(shader_type)
{
	wlog_info("{} {}", "Initializing shader: ", path);
	if(path.ends_with(".comp")) {
		this->shader_type = ShaderType::Compute;
	} else if(path.ends_with(".frag")) {
		this->shader_type = ShaderType::Fragment;
	} else if(path.ends_with(".geom")) {
		this->shader_type = ShaderType::Geometry;
	} else if(path.ends_with(".vert")) {
		this->shader_type = ShaderType::Vertex;
	} else {
		assert(false);
	}
	create();
}

void Shader::compile() {
	this->code = ShaderManager::get_shader_code(this);
	// delete prev iteration of this shader, if doing live reloading
	const char* code_cstr = this->code.c_str();

	int new_gl_id = glCreateShader((int)this->shader_type);
	// glShaderSource(new_gl_id, 1, &this->code[0], NULL);
	glShaderSource(new_gl_id, 1, &code_cstr, NULL);
	glCompileShader(new_gl_id);

	// glGetShaderI

	int32_t compile_status;
	glGetShaderiv(new_gl_id, GL_COMPILE_STATUS, &compile_status);

	if (compile_status == 0) {
		// char* error_ptr = this->error.c_str();
		glGetShaderInfoLog(new_gl_id, 1000 - 1, NULL, this->error.data());
		wlog_err("compile error");
		wlog_err("{}\n\n", this->error.c_str());
		// printf("compile error");
		// fprintf(stderr, "%s\n\n", this->error.c_str());
		if (new_gl_id > 0) {
			glDeleteShader(new_gl_id);
		}
		compilation_success = false;
	} else {
		if (pid > 0) {
			glDeleteShader(pid);
		}
		this->pid = new_gl_id;

		#ifdef GL_DEBUG_ENABLED
		glObjectLabel( GL_SHADER, this->pid, -1, &this->path[0]);
		#endif

		compilation_success = true;
	}
}
