#include "pch.h"

#include "ShaderProgram.h"

#include <print>
#include "ShaderManager.h"
#include "Shader.h"
#include "engine/Engine.h"
#include "gl/Texture.h"
#include "gl/Buffer.h"


ShaderProgram::ShaderProgram(
	ShaderProgDescr shader_prog_descr,
	bool gen_shadow_map_shader,
	bool is_shadow_map_shader
) {

	gen_shadow_map_shader = false;
	if(gen_shadow_map_shader) {
		ShaderProgDescr shader_prog_descr_shadow_map = shader_prog_descr;
		shader_prog_descr_shadow_map.shader_c = WEngine->shadowmap_geom_shader;

		this->shadow_map_variation = new ShaderProgram( shader_prog_descr_shadow_map, false, true );
	} else if(is_shadow_map_shader){
		WDumbMap<SIMDString<>, 32> _defines_map_shadow_map_shader;
		_defines_map_shadow_map_shader.set("SHADOW_MAP", "");
		shader_prog_descr.shader_a = shader_prog_descr.shader_a->clone_with_new_defines(_defines_map_shadow_map_shader);
		shader_prog_descr.shader_b = WEngine->passthrough_frag_shader;
	}

	Shader* shaders[] = {
		shader_prog_descr.shader_a,
		shader_prog_descr.shader_b,
		shader_prog_descr.shader_c,
		shader_prog_descr.shader_d
	};

	auto add_shader_prog_to_shader = [&](Shader* shader) {
		shader->shader_programs.push_back(this);
		// auto it = std::find(shader->shader_programs.begin(), shader->shader_programs.end(), this);
		// if(shader->shader_programs.)
	};

	for (auto shader : shaders) {
		if(shader != nullptr) {
			add_shader_prog_to_shader(shader);
			if(shader->shader_type == ShaderType::Compute) {
				this->comp_shader = shader;
			} else if(shader->shader_type == ShaderType::Fragment) {
				this->frag_shader = shader;
			} else if(shader->shader_type == ShaderType::Vertex) {
				this->vert_shader = shader;
			} else if(shader->shader_type == ShaderType::Geometry) {
				this->geom_shader = shader;
			}
		}
	}
	this->error = new char[1000];
	#ifdef EDITOR
	WEngine->shader_manager->shader_programs.add(this);
	#endif
	link();
}

void ShaderProgram::setUniform(const char* name, Buffer* uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, float uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::mat4 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::vec2 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::vec3 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::vec4 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::ivec2 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::ivec3 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, glm::ivec4 uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, Texture* uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, int uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}

void ShaderProgram::setUniform(const char* name, bool uniform_val) {
	this->setUniformInternal(name, uniform_val, get_uniform_loc(name));
}


void ShaderProgram::use() {
	glUseProgram(this->pid);
}

void ShaderProgram::setUniformInternal(const char* name, Buffer* buff, int loc) {
	// glProgramUniformui64NV(this->pid, glGetUniformLocation(this->pid, name), buff->bindless_address );
	glProgramUniform1ui64NV(this->pid, loc, buff->bindless_address);
}

void ShaderProgram::setUniformInternal(const char* name, int32_t v, int loc) {
	glProgramUniform1i(this->pid, loc, v);
}

void ShaderProgram::setUniformInternal(const char* name, bool v, int loc) {
	glProgramUniform1i(this->pid, loc, int(v));
}

void ShaderProgram::setUniformInternal(const char* name, glm::mat4 v, int loc) {
	glProgramUniformMatrix4fv(this->pid, loc, 1, GL_FALSE, &v[0][0]);
}

void ShaderProgram::setUniformInternal(const char* name, glm::vec2 v, int loc) {
	glProgramUniform2fv(this->pid, loc, 1, &v[0]);
}

void ShaderProgram::setUniformInternal(const char* name, glm::vec3 v, int loc) {
	glProgramUniform3fv(this->pid, loc, 1, &v[0]);
}

void ShaderProgram::setUniformInternal(const char* name, glm::vec4 v, int loc) {
	glProgramUniform4fv(this->pid, loc, 1, &v[0]);
}

void ShaderProgram::setUniformInternal(const char* name, glm::ivec2 v, int loc) {
	glProgramUniform2iv(this->pid, loc, 1, &v[0]);
}

void ShaderProgram::setUniformInternal(const char* name, glm::ivec3 v, int loc) {
	glProgramUniform3iv(this->pid, loc, 1, &v[0]);
}

void ShaderProgram::setUniformInternal(const char* name, glm::ivec4 v, int loc) {
	glProgramUniform4iv(this->pid, loc, 1, &v[0]);
}

void ShaderProgram::dispatch_compute(int resx, int resy, int resz, bool auto_wkg_sz) {
	if(auto_wkg_sz) {
		if (this->wkg_sz.x < 0 || this->wkg_sz.y < 0 || this->wkg_sz.z < 0) {
			glGetProgramiv(this->pid, GL_COMPUTE_WORK_GROUP_SIZE, &this->wkg_sz.x);
		}
		int disp_x = (resx + this->wkg_sz.x - 1) / (this->wkg_sz.x);
		int disp_y = (resy + this->wkg_sz.y - 1) / (this->wkg_sz.y);
		int disp_z;
		if(resz <= 1) {
			disp_z = 1;
		} else {
			disp_z = (resz + this->wkg_sz.z - 1) / (this->wkg_sz.z);
		}

		// int wg_sz = 16;
		glDispatchCompute( disp_x, disp_y, disp_z );
	} else {
		glDispatchCompute( resx, resy, resz );
	}
}

void ShaderProgram::setUniformInternal(const char* name, Texture* v, int loc) {
	glProgramUniform1i(this->pid, loc, v->bindless_idx);
}

void ShaderProgram::setUniformInternal(const char* name, float v, int loc) {
	glProgramUniform1f(this->pid, loc, v);
}


void ShaderProgram::link() {
	int new_gl_id;
	if (this->vert_shader != nullptr) {
		if (!vert_shader->compilation_success || !frag_shader->compilation_success || (geom_shader && !geom_shader->compilation_success)) {
			return;
		}
		new_gl_id = glCreateProgram();
		glAttachShader(new_gl_id, vert_shader->pid);
		glAttachShader(new_gl_id, frag_shader->pid);
		if(geom_shader) {
			glAttachShader(new_gl_id, geom_shader->pid);
		}
	} else {
		if (!comp_shader->compilation_success) {
			return;
		}
		new_gl_id = glCreateProgram();
		glAttachShader(new_gl_id, comp_shader->pid);
	}

	glLinkProgram(new_gl_id);

	#ifdef EDITOR
	int gl_temp_link_status;
	glGetProgramiv(new_gl_id, GL_LINK_STATUS, &gl_temp_link_status);
	if (gl_temp_link_status == 0) {
		this->link_success = false;
		glGetProgramInfoLog(new_gl_id, 1000 - 1, NULL, this->error);
		// __debugbreak();
		std::println(stderr, "Error linking shader program. ");
		std::println(stderr, "{}\n", this->error);
	} else {
		glValidateProgram(new_gl_id);
		this->link_success = true;
		GLint validated = false;
		glGetProgramiv(new_gl_id, GL_VALIDATE_STATUS, &validated);
		if (validated == GL_FALSE) {
			char val_log[1000];
			glGetProgramInfoLog(new_gl_id, 1000 - 1, NULL, &val_log[0]);
			std::cout << "-- SHADER VALIDATION ERROR --" << "\n";
			std::cout << val_log << "\n";
		}
	}
	if (!this->link_success) {
		// Delete new program.
		if (new_gl_id > 0) {
			glDeleteProgram(new_gl_id);
		}
	} else {
		// Replace old pid with new pid.
		glDeleteProgram(this->pid);
		this->pid = new_gl_id;
		// error = null;
		this->link_success = true;

		if(this->comp_shader != nullptr){
			glGetProgramiv(this->pid, GL_COMPUTE_WORK_GROUP_SIZE, &this->wkg_sz.x);
		}

		#ifdef GL_DEBUG_ENABLED
		SIMDString<128> shaderprog_name;
		if(this->vert_shader) {
			shaderprog_name += this->vert_shader->path;
			shaderprog_name += " ";
		}
		if(this->frag_shader) {
			shaderprog_name += this->frag_shader->path;
			shaderprog_name += " ";
		}
		if(this->geom_shader) {
			shaderprog_name += this->geom_shader->path;
			shaderprog_name += " ";
		}
		if(this->comp_shader) {
			shaderprog_name += this->comp_shader->path;
			shaderprog_name += " ";
		}
		if(this->shadow_map_variation) {
			shaderprog_name += "HAS SHADOWMAP";
		}

		glObjectLabel( GL_PROGRAM, this->pid, -1, &shaderprog_name[0]);
		#endif

		// Delete new shaders.
		if (this->vert_shader != nullptr) {
			glDetachShader(this->pid, vert_shader->pid);
			glDetachShader(this->pid, frag_shader->pid);
			if(this->geom_shader) {
				glDetachShader(this->pid, this->geom_shader->pid);
			}
		} else {
			glDetachShader(this->pid, comp_shader->pid);
		}
	}
	#else
  this->pid = new_gl_id;
	this->link_success = true;
	#endif
}
