#pragma once
#include "WDumbMap.h"

struct Shader;
class Texture;
class Buffer;


typedef std::variant<int, float, glm::mat4, glm::vec2, Texture*, Buffer*, bool> UniformValue;

struct ShaderProgDescr {
  Shader* shader_a = nullptr;
  Shader* shader_b = nullptr;
  Shader* shader_c = nullptr;
  Shader* shader_d = nullptr;
};

class ShaderProgram {
private:
  void setUniformInternal(const char* name, Buffer* buff, int loc);
  void setUniformInternal(const char* name, float v, int loc);
	void setUniformInternal(const char* name, glm::mat4 v, int loc);
	void setUniformInternal(const char* name, glm::vec2 v, int loc);
	void setUniformInternal(const char* name, glm::vec3 v, int loc);
	void setUniformInternal(const char* name, glm::vec4 v, int loc);
	void setUniformInternal(const char* name, glm::ivec2 v, int loc);
	void setUniformInternal(const char* name, glm::ivec3 v, int loc);
	void setUniformInternal(const char* name, glm::ivec4 v, int loc);
  void setUniformInternal(const char* name, Texture* v, int loc);
  void setUniformInternal(const char* name, int v, int loc);
  void setUniformInternal(const char* name, bool v, int loc);

	inline int get_uniform_loc(const char* name) {
		int loc;
		if(this->uniforms_map.contains(name)) {
			loc = this->uniforms_map.get(name);
		} else {
			loc = glGetUniformLocation(this->pid, name);
			this->uniforms_map.set(name, loc);
		}
		return loc;
	}

public:
  int pid = 0;

  char* error = nullptr;
  bool link_success = false;

  Shader* vert_shader = nullptr;
  Shader* frag_shader = nullptr;
  Shader* comp_shader = nullptr;
  Shader* geom_shader = nullptr;

	WDumbMap<int, 32> uniforms_map;
	WDumbMap<SIMDString<>, 32> defines_map;

	ShaderProgram* shadow_map_variation = nullptr;

	glm::ivec3 wkg_sz = glm::ivec3(-1,-1,-1);

  ShaderProgram(ShaderProgDescr shader_prog_descr, bool gen_shadow_map_shader = false, bool is_shadow_map_shader = false);


	void setUniform(const char* name, Buffer* uniform_val);
  void setUniform(const char* name, float uniform_val);
  void setUniform(const char* name, glm::mat4 uniform_val);
  void setUniform(const char* name, glm::vec2 uniform_val);
	void setUniform(const char* name, glm::vec3 uniform_val);
	void setUniform(const char* name, glm::vec4 uniform_val);
  void setUniform(const char* name, glm::ivec2 uniform_val);
	void setUniform(const char* name, glm::ivec3 uniform_val);
	void setUniform(const char* name, glm::ivec4 uniform_val);
  void setUniform(const char* name, Texture* uniform_val);
  void setUniform(const char* name, int uniform_val);
	void setUniform(const char* name, bool uniform_val);

  void use();
	void dispatch_compute(int resx, int resy, int resz = 0, bool auto_wkg_sz = true);

  void link();
};
