#include "pch.h"
#include "Texture.h"
#include "Buffer.h"
#include "engine/Engine.h"


int get_components_cnt(Format format) {
	int comp_cnt = 0;
	switch (format) {
		case Format::UNSET:
			break;
		case Format::RGBA:
		case Format::RGBA_INTEGER:
		case Format::BGRA: {
			comp_cnt = 4;
			break;
		}
		case Format::RGB_INTEGER:
		case Format::RGB: {
			comp_cnt = 3;
			break;
		}
		case Format::RG_INTEGER:
		case Format::RG: {
			comp_cnt = 2;
			break;
		}
		case Format::DEPTH_COMPONENT:
		case Format::RED:
		case Format::RED_INTEGER: {
			comp_cnt = 1;
			break;
		}
	}
	return comp_cnt;
}

std::string get_name_from_internal_format(InternalFormat internal_format) {
	switch (internal_format) {
		case InternalFormat::UNSET:
			return "UNSET";
		case InternalFormat::RGBA32F:
			return "RGBA32F";
		case InternalFormat::RGB32F:
			return "RGB32F";
		case InternalFormat::RGBA16F:
			return "RGBA16F";
		case InternalFormat::RGB16F:
			return "RGB16F";
		case InternalFormat::RGBA:
			return "RGBA";
		case InternalFormat::RGBA8:
			return "RGBA8";
		case InternalFormat::SRGB8_ALPHA8:
			return "SRGB8_ALPHA8";
		case InternalFormat::RGB:
			return "RGB";
		case InternalFormat::RGB8:
			return "RGB8";
		case InternalFormat::SRGB8:
			return "SRGB8";
		case InternalFormat::RG:
			return "RG";
		case InternalFormat::R:
			return "R";
		case InternalFormat::RG16F:
			return "RG16F";
		case InternalFormat::R16F:
			return "R16F";
		case InternalFormat::R32F:
			return "R32F";
		case InternalFormat::R32UI:
			return "R32UI";
		case InternalFormat::R16UI:
			return "R16UI";
		case InternalFormat::RG16UI:
			return "RG16UI";
		case InternalFormat::RGB16UI:
			return "RGB16UI";
		case InternalFormat::RGBA16UI:
			return "RGBA16UI";
		case InternalFormat::R8UI:
			return "R8UI";
		case InternalFormat::R32I:
			return "R32I";
		case InternalFormat::R16I:
			return "R16I";
		case InternalFormat::RG16I:
			return "RG16I";
		case InternalFormat::R8I:
			return "R8I";
		case InternalFormat::R11F_G11F_B10F:
			return "R11F_G11F_B10F";
		case InternalFormat::DEPTH_COMPONENT:
			return "DEPTH_COMPONENT";
		case InternalFormat::DEPTH_COMPONENT16:
			return "DEPTH_COMPONENT16";
		case InternalFormat::DEPTH_COMPONENT24:
			return "DEPTH_COMPONENT24";
		case InternalFormat::DEPTH_COMPONENT32:
			return "DEPTH_COMPONENT32";
		case InternalFormat::DEPTH_COMPONENT32F:
			return "DEPTH_COMPONENT32F";
	}
	return "errname";
}

int get_bytes_per_component(InternalFormat internal_format) {
	int bytes = 0;

	switch(internal_format) {
		case InternalFormat::RGB:
		case InternalFormat::RGB8:
		case InternalFormat::SRGB8:
		case InternalFormat::RGBA:
		case InternalFormat::RGBA8:
		case InternalFormat::RG:
		case InternalFormat::R8I:
		case InternalFormat::R8UI:
		case InternalFormat::R:
		case InternalFormat::SRGB8_ALPHA8:{
			bytes = 1;
			break;
		}

		case InternalFormat::R11F_G11F_B10F:
		case InternalFormat::RGBA16F:
		case InternalFormat::R16I:
		case InternalFormat::R16F:
		case InternalFormat::RG16F:
		case InternalFormat::RG16I:
		case InternalFormat::RGB16F:
		case InternalFormat::RGB16UI:
		case InternalFormat::RGBA16UI:
		case InternalFormat::R16UI:
		case InternalFormat::DEPTH_COMPONENT16:
		case InternalFormat::RG16UI: {
			bytes = 2;
			break;
		}

		case InternalFormat::R32F:
		case InternalFormat::R32I:
		case InternalFormat::RGB32F:
		case InternalFormat::RGBA32F:
		case InternalFormat::R32UI:
		case InternalFormat::DEPTH_COMPONENT32:
		case InternalFormat::DEPTH_COMPONENT32F: {
			bytes = 4;
			break;
		}

		case InternalFormat::DEPTH_COMPONENT:
		case InternalFormat::DEPTH_COMPONENT24:
		{
			break;
		}
		default: {

		};
	}
	return bytes;
}
inline std::pair<Format, Type> get_format_and_type(InternalFormat internal_format) {
	Format format = Format::UNSET;
	Type type = Type::UNSET;

	switch(internal_format) {
		case InternalFormat::RGBA32F:
		case InternalFormat::RGBA16F:
		case InternalFormat::RGBA:
		case InternalFormat::RGBA8:
		case InternalFormat::SRGB8_ALPHA8:{
			format = Format::RGBA;
			break;
		}

		case InternalFormat::R11F_G11F_B10F:
		case InternalFormat::RGB16F:
		case InternalFormat::RGB:
		case InternalFormat::RGB8:
		case InternalFormat::SRGB8: {
			format = Format::RGB;
			break;
		}

		case InternalFormat::RG16F:
		case InternalFormat::RG:{
			format = Format::RG;
			break;
		}

		case InternalFormat::R16F:
		case InternalFormat::R32F:
		case InternalFormat::R:{
			format = Format::RED;
			break;
		}

		case InternalFormat::DEPTH_COMPONENT:
		case InternalFormat::DEPTH_COMPONENT16:
		case InternalFormat::DEPTH_COMPONENT24:
		case InternalFormat::DEPTH_COMPONENT32:
		case InternalFormat::DEPTH_COMPONENT32F: {
			format = Format::DEPTH_COMPONENT;
			break;
		}
		case InternalFormat::R8I:
		case InternalFormat::R16I:
		case InternalFormat::R32I:
		case InternalFormat::R8UI:
		case InternalFormat::R16UI:
		case InternalFormat::R32UI: {
			format = Format::RED_INTEGER;
			break;
		}

		case InternalFormat::RG16I:
		case InternalFormat::RG16UI: {
			format = Format::RG_INTEGER;
			break;
		}
		case InternalFormat::RGB16UI: {
			format = Format::RGB_INTEGER;
			break;
		}
		case InternalFormat::RGBA16UI: {
			format = Format::RGBA_INTEGER;
			break;
		}
		
		default: ;
	}

	switch(internal_format) {
		// ------- FLOAT
		case InternalFormat::RGBA32F:
		case InternalFormat::RGB32F:
		case InternalFormat::RGBA16F:
		case InternalFormat::R11F_G11F_B10F:
		case InternalFormat::RGB16F:
		case InternalFormat::R16F:
		case InternalFormat::R32F:
		case InternalFormat::RG16F:
		case InternalFormat::DEPTH_COMPONENT32F: {
			type = Type::FLOAT;
			break;
		}

		// ------- 8 BIT
		case InternalFormat::R8UI:
		case InternalFormat::R:
		case InternalFormat::RG:
		case InternalFormat::RGBA:
		case InternalFormat::RGB:
		case InternalFormat::RGBA8:
		case InternalFormat::SRGB8_ALPHA8:
		case InternalFormat::RGB8:
		case InternalFormat::SRGB8: {
			type = Type::UNSIGNED_BYTE;
			break;
		}

		case InternalFormat::R8I: {
			type = Type::BYTE;
			break;
		}

		// ------- 16 BIT
		case InternalFormat::R16UI:
		case InternalFormat::RG16UI:
		case InternalFormat::RGB16UI:
		case InternalFormat::RGBA16UI:
		case InternalFormat::DEPTH_COMPONENT16: {
			type = Type::UNSIGNED_SHORT;
			break;
		}

		case InternalFormat::RG16I:
		case InternalFormat::R16I:  {
			type = Type::SHORT;
			break;
		}

		// ------- 32 BIT
		case InternalFormat::R32UI:
		case InternalFormat::DEPTH_COMPONENT32:
		case InternalFormat::DEPTH_COMPONENT24: {
			type = Type::UNSIGNED_INT;
			break;
		}

		case InternalFormat::R32I: {
			type = Type::INT;
			break;
		}

		case InternalFormat::DEPTH_COMPONENT: {
			format = Format::DEPTH_COMPONENT;
			break;
		}
		default: ;
	}

	return {format, type};
}

Texture::Texture(std::string file_name) {
	int width, height, nrChannels;

	bool is_hdri = file_name.ends_with(".hdr");

	stbi_set_flip_vertically_on_load(true);

	if(!is_hdri) {
		this->internal_format = InternalFormat::RGBA8;

		unsigned char *data = stbi_load(file_name.c_str(), &width, &height, &nrChannels, 0);

		if (nrChannels == 3) {
			this->internal_format = InternalFormat::RGB8;
			format = Format::RGB;
		}
		
		this->create({.resx = (uint32_t)width, .resy = (uint32_t)height, .resz =1, .internal_format = this->internal_format, .name = file_name});
		this->upload_data(data);
		stbi_image_free(data);
	} else {
		float *data = stbi_loadf(file_name.c_str(), &width, &height, &nrChannels, 0);
		TextureDesc tex_desc = {(uint32_t)width, (uint32_t)height, 1, InternalFormat::RGB16F, };
		tex_desc.is_mipmapped = true;
		this->create(tex_desc);
		this->upload_data(data);
		// glTextureParameteri(this->pid, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		// glTextureParameteri(this->pid, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		this->generate_mipmap();

		// glTextureParameteri(pid, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
		stbi_image_free(data);
	}

	// this->name = file_name;
	//
	// if(this->name.size() > 0) {
	// 	glObjectLabel( GL_TEXTURE, this->pid, -1, &this->name[0]);
	// }
	this->set_name(file_name);


	wlog_info("Loaded image file: ", file_name.c_str());
	// glTextureSubImage2D

	// glTexture
}

void Texture::set_name(std::string_view _name) {
	this->name = _name;
	glObjectLabel( GL_TEXTURE, this->pid, -1, &this->name[0]);
}

void Texture::upload_data(void* data) {
	if(this->resz == 1) {
		glTextureSubImage2D(
			this->pid,
			0,
			0,
			0,
			(int)this->resx,
			(int)this->resy,
			static_cast<GLenum>(this->format),
			static_cast<GLenum>(this->type),
			data
		);
	} else {
		glTextureSubImage3D(
			this->pid,
			0,
			0,
			0,
			0,
			(int)this->resx,
			(int)this->resy,
			(int)this->resz,
			static_cast<GLenum>(this->format),
			static_cast<GLenum>(this->type),
			data
		);
	}
}

void Texture::destroy() {
	if(this->tex_handle != std::numeric_limits<uint64_t>::max()) {
		glMakeImageHandleNonResidentARB(this->img_handle);
	}

	glMakeTextureHandleNonResidentARB(this->tex_handle);

	GLuint textures = {this->pid};
	glDeleteTextures(1, &textures);
}

void Texture::generate_mipmap() { glGenerateTextureMipmap(this->pid); }

void Texture::create(TextureDesc texture_desc) {
	this->_texture_desc = texture_desc;
	
	this->is_cubemap = texture_desc.is_cubemap;
	this->resx = texture_desc.resx;
	this->resy = texture_desc.resy;
	this->resz = texture_desc.resz;
	this->internal_format = texture_desc.internal_format;
	if(!texture_desc.name.empty()) {
		this->name = texture_desc.name;
	}

	auto format_and_type = get_format_and_type(this->internal_format);
	this->format = format_and_type.first;
	this->type = format_and_type.second;

	this->components = get_components_cnt(this->format);
	this->bytes_per_comp = get_bytes_per_component(this->internal_format);

	this->internal_format_name = get_name_from_internal_format(this->internal_format);
	// this->format = texture_desc.format;
	// this->type = texture_desc.type;

	if(is_cubemap) {
		glCreateTextures(GL_TEXTURE_CUBE_MAP, 1, &pid);
	} else {
		if(this->resz > 1){
			glCreateTextures(GL_TEXTURE_3D, 1, &pid);
		} else {
			glCreateTextures(GL_TEXTURE_2D, 1, &pid);
		}
	}


	glTextureParameteri(this->pid, GL_TEXTURE_MIN_FILTER, static_cast<GLint>(texture_desc.min_filter));
	glTextureParameteri(this->pid, GL_TEXTURE_MAG_FILTER, static_cast<GLint>(texture_desc.mag_filter));
	glTextureParameteri(pid, GL_TEXTURE_WRAP_S, static_cast<GLint>(texture_desc.wrap_s));
	glTextureParameteri(pid, GL_TEXTURE_WRAP_T, static_cast<GLint>(texture_desc.wrap_t));

	if(!is_cubemap) {
		int mip_levels = 1;
		if(texture_desc.is_mipmapped) {
			if(texture_desc.mip_levels != std::nullopt) {
				mip_levels = texture_desc.mip_levels.value();
			} else {
				float max_res = float(glm::max(this->resx, this->resy));
				mip_levels = int(floor(log2(max_res))) + 1;
			}
		}

		if(this->resz > 1) {
			glTextureStorage3D(pid, mip_levels, static_cast<GLenum>(internal_format), (int)this->resx, (int)this->resy, (int)this->resz);
		} else {
			glTextureStorage2D(pid, mip_levels, static_cast<GLenum>(internal_format), (int)this->resx, (int)this->resy);
		}
	} else {
		// glTextureStorage2D(pid, 1, static_cast<GLenum>(internal_format), (int)this->resx, (int)this->resy);
		if(this->resz > 1) {
			glTextureStorage3D(pid, 1, static_cast<GLenum>(internal_format), (int)this->resx, (int)this->resy, (int)this->resz);
		} else {
			glTextureStorage2D(pid, 1, static_cast<GLenum>(internal_format), (int)this->resx, (int)this->resy);
		}
		static auto empty = std::vector<float> (1000 * 1000 * 4 * 8);
		for(int i = 0; i < 6; i++) {
			glTextureSubImage3D(
				this->pid, 0, 0, 0, i,
				(int)this->resx, (int)this->resy, 1,
				static_cast<GLenum>(this->format), static_cast<GLenum>(this->type), &empty[0]
				);
		}
	}


	// glTextureStorage2D(pid, 1, int(internal_format), this->resx, this->resy);

	// glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_DEPTH_COMPONENT,
	// 									SHADOW_WIDTH, SHADOW_HEIGHT, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

	this->tex_handle = glGetTextureHandleARB(pid);
	// this->tex_handle = glGetTextureSamplerHandleARB(pid);
	// glGetTextureSamplerHandleARB()

	// glGetTextureSamplerHandleNV

	if (
		this->internal_format != InternalFormat::DEPTH_COMPONENT &&
		this->internal_format != InternalFormat::DEPTH_COMPONENT24 &&
		this->internal_format != InternalFormat::DEPTH_COMPONENT16 &&
		this->internal_format != InternalFormat::DEPTH_COMPONENT32F &&
		this->internal_format != InternalFormat::RGB8 &&
		this->internal_format != InternalFormat::SRGB8 &&
		this->internal_format != InternalFormat::SRGB8_ALPHA8 &&
		this->internal_format != InternalFormat::RGB16F &&
		this->internal_format != InternalFormat::DEPTH_COMPONENT32
	) {
		this->img_handle = glGetImageHandleARB(
			pid,
			0,
			this->resz == 1 ? false : true,
			0,
			(int)this->internal_format
		);
		// if(texture_desc.tex_access == T)
		glMakeImageHandleResidentARB(this->img_handle, static_cast<GLenum>(texture_desc.tex_access));
	}
	glMakeTextureHandleResidentARB(this->tex_handle);

	// if(glTextureCom)
	// TODO: FIX ALIASING
	// uint64_t* shared_ssbo_mapped_buff = (uint64_t*)(Globals::engine->shared_ssbo->mapped_buff);
	uint64_t* shared_ssbo_mapped_buff = (uint64_t*)(Globals::engine->shared_ssbo->cpu_data);

	// TODO: Find free slot
	// uint32_t tex_slot = Globals::engine->alloc_textures.alloc_cnt;
	uint32_t tex_slot;
	if(this->bindless_idx < 0) {
		 tex_slot = Globals::engine->alloc_textures.get_next_slot();
	} else {
		 tex_slot = this->bindless_idx;
	}

	// TODO: Support more than 64 and/or bleep bloop error.
	shared_ssbo_mapped_buff[0 + tex_slot] = tex_handle;
	shared_ssbo_mapped_buff[256 + tex_slot] = img_handle;
	

	this->bindless_idx = tex_slot;

	Globals::engine->shared_ssbo->flush(); // does nuthin lmao

	this->clear();

	#ifdef EDITOR
	// if(!texture_desc.is_mipmapped && !texture_desc.is_cubemap
	// 	&& texture_desc.internal_format != InternalFormat::DEPTH_COMPONENT
	// 	&& texture_desc.internal_format != InternalFormat::DEPTH_COMPONENT16
	// 	&& texture_desc.internal_format != InternalFormat::DEPTH_COMPONENT24
	// 	&& texture_desc.internal_format != InternalFormat::DEPTH_COMPONENT32
	// 	&& texture_desc.internal_format != InternalFormat::DEPTH_COMPONENT32F
	// 	) {
	// 	// Create a framebuffer object
	// 	GLuint framebuffer;
	// 	glCreateFramebuffers(1, &framebuffer);
	// 	glNamedFramebufferTexture(framebuffer, GL_COLOR_ATTACHMENT0, this->pid, 0);
	// 	uint32_t buffs[] = {GL_COLOR_ATTACHMENT0};
	// 	glNamedFramebufferDrawBuffers(framebuffer, 1, buffs);
	//
	// 	uint32_t status = glCheckNamedFramebufferStatus(this->pid, GL_FRAMEBUFFER);
	// 	if (status != GL_FRAMEBUFFER_COMPLETE) {
	// 		std::cout << "---- INCOMPLETE TEXTURE \n";
	// 	}
	//
	// 	glDeleteFramebuffers(1, &framebuffer);
	// }
	#endif
}

Texture::Texture( TextureDesc _texture_desc ) {
	this->create(_texture_desc);
}

Texture::Texture() {}

Texture* Texture::clone() {
	TextureDesc tex_desc = this->_texture_desc;
	return WEngine->alloc_textures.push(tex_desc).ptr;
}

void Texture::clear() {
	glClearTexSubImage(
		this->pid, 0,
		0,0,0,
		this->resx, this->resy, this->resz, (GLenum)this->format,static_cast<GLenum>(this->type), nullptr
	);
	// glClearTexImage(
	// 	this->pid, 0, static_cast<GLenum>(this->format), static_cast<GLenum>(this->type), nullptr
	// );
	// glClearTexImage(((GLTexture*)tex)->ID, 0, ((GLTexture*)tex)->srcFormat, ((GLTexture*)tex)->gl_type, nullptr);
}
