#pragma once

enum class TextureWrapping : GLint{
	Repeat = GL_REPEAT,
	MirroredRepeat = GL_MIRRORED_REPEAT,
	MirroredClampToEdge = GL_MIRROR_CLAMP_TO_EDGE,
};

enum class TextureFilter : GLenum{
	LinearMipLinear = GL_LINEAR_MIPMAP_LINEAR,
	Linear = GL_LINEAR,
	Nearest = GL_NEAREST,
};

enum class InternalFormat : GLenum {
	UNSET           = 0,
	RGBA32F            = GL_RGBA32F,
	RGB32F            = GL_RGB32F,
	RGBA16F            = GL_RGBA16F,
	RGB16F             = GL_RGB16F,
	RGBA               = GL_RGBA,
	RGBA8              = GL_RGBA8,
	SRGB8_ALPHA8       = GL_SRGB8_ALPHA8,
	RGB                = GL_RGB,
	RGB8               = GL_RGB8,
	SRGB8              = GL_SRGB8,
	RG                 = GL_RG,
	R                  = GL_R,
	RG16F              = GL_RG16F,
	R32F               = GL_R32F,
	R16F               = GL_R16F,
	R32UI              = GL_R32UI,
	R16UI              = GL_R16UI,
	RG16UI             = GL_RG16UI,
	RGB16UI             = GL_RGB16UI,
	RGBA16UI             = GL_RGBA16UI,
	R8UI               = GL_R8UI,
	R32I               = GL_R32I,
	R16I               = GL_R16I,
	RG16I              = GL_RG16I,
	R8I                = GL_R8I,
	R11F_G11F_B10F     = GL_R11F_G11F_B10F,
	DEPTH_COMPONENT    = GL_DEPTH_COMPONENT,
	DEPTH_COMPONENT16  = GL_DEPTH_COMPONENT16,
	DEPTH_COMPONENT24  = GL_DEPTH_COMPONENT24,
	DEPTH_COMPONENT32  = GL_DEPTH_COMPONENT32,
	DEPTH_COMPONENT32F = GL_DEPTH_COMPONENT32F
	//        DEPTH_COMPONENT16( GL14.GL_DEPTH_COMPONENT16),
};

enum class Format : GLenum {
	UNSET           = 0,
	RGBA            = GL_RGBA,
	BGRA            = GL_BGRA,
	RGB             = GL_RGB,
	RG              = GL_RG,
	RED             = GL_RED,
	RED_INTEGER     = GL_RED_INTEGER,
	RG_INTEGER      = GL_RG_INTEGER,
	RGB_INTEGER     = GL_RGB_INTEGER,
	RGBA_INTEGER    = GL_RGBA_INTEGER,
	DEPTH_COMPONENT = GL_DEPTH_COMPONENT
};

enum class Type : GLenum {
	UNSET          = 0,
	FLOAT          = GL_FLOAT,
	UNSIGNED_BYTE  = GL_UNSIGNED_BYTE,
	UNSIGNED_INT   = GL_UNSIGNED_INT,
	UNSIGNED_SHORT = GL_UNSIGNED_SHORT,
	SHORT          = GL_SHORT,
	BYTE           = GL_BYTE,
	INT            = GL_INT,
};


enum class TexAccess: GLenum {
	READ = GL_READ_ONLY,
	WRITE = GL_WRITE_ONLY,
	RW = GL_READ_WRITE
};
struct TextureDesc {
	uint32_t resx;
	uint32_t resy;
	uint32_t resz = 0;
	InternalFormat internal_format;
	bool is_cubemap = false;
	bool is_mipmapped = false;
	std::optional<int> mip_levels = std::nullopt;
	TextureFilter min_filter = TextureFilter::LinearMipLinear;
	TextureFilter mag_filter = TextureFilter::Linear;
	TextureWrapping wrap_s = TextureWrapping::MirroredRepeat;
	TextureWrapping wrap_t = TextureWrapping::MirroredRepeat;
	TexAccess tex_access = TexAccess::READ;
	std::string name = "";
};

class Texture {
	public:
		uint32_t pid = 0;
		uint32_t resx = 0;
		uint32_t resy = 0;
		uint32_t resz = 0;
		uint64_t tex_handle = 0;
		uint64_t img_handle = std::numeric_limits<uint64_t>::max();

		TextureDesc _texture_desc;
	
		bool is_model_texture = false;
		bool is_cubemap = false;
		int bindless_idx = -1;

		int generation = -1;
		int alloc_slot = -1;

		int bytes_per_comp = 0;
		int components = 0;

		InternalFormat internal_format = InternalFormat::UNSET;
		Format format = Format::UNSET;
		Type type = Type::UNSET;

		Framebuffer* bound_fb = nullptr;

		std::string name = std::string("");
		std::string internal_format_name;


		Texture();
		Texture(std::string file_name);
		void set_name(std::string_view _name);
		void create(TextureDesc texture_dex);
		Texture(TextureDesc texture_dex);
		Texture* clone();
		void clear();
		void upload_data(void* data);

		void destroy();

		void generate_mipmap();
		void bind_to_texture_unit(int unit) {
			glBindTextureUnit(unit, this->pid);
		}
		void bind_to_image_unit(int unit) {
			glBindImageTexture(unit, this->pid, 0, 0, 0, GL_READ_WRITE, (GLenum)this->internal_format);
			// glBindTextureUnit(unit, this->pid);
		}
		void resize(uint32_t resx, uint32_t resy) {
			this->destroy();
			this->create(this->_texture_desc);
		}
};
