#include "pch.h"
#define WGL_FILE

#include "WGL.h"

#include <dwmapi.h>

#include "engine/IO.h"
#include "engine/Engine.h"
#include "engine/EditorNew.h"


extern LRESULT ImGui_ImplWin32_WndProcHandler(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

WGL::WGL() { this->init_modern_gl(); }

LRESULT CALLBACK WGL::WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	if(msg == 7) {
		WEngine->io->window_focused = true;
	} else if(msg == 8) {
		WEngine->io->window_focused = false;
	}
	// if(msg == 7 || msg == 8) {
	// 	wlog_info("{}", msg);
	// }

	if (GetActiveWindow() != hwnd)
		return DefWindowProc(hwnd, msg, wParam, lParam);

	// if(msg == WM_KEYDOWN && wParam != (WPARAM)Key::Tab) {
	if(wParam != (WPARAM)Key::Tab) {
		if (ImGui_ImplWin32_WndProcHandler(hwnd, msg, wParam, lParam))
			return true;
	}

	if (Globals::engine != nullptr && Globals::engine->editor != nullptr) {
		if (!Globals::engine->io->handle_input(msg, wParam)) {
			switch (msg) {
				default:
					return DefWindowProc(hwnd, msg, wParam, lParam);
			}
		}
	}
	switch (msg) {
		case 16:
			Globals::engine->editor->finished = true;
			return true;
		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	#ifdef EDITOR
	#else
	switch (msg) {
		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	#endif

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

// void GLAPIENTRY
// MessageCallback( GLenum source,
//                  GLenum type,
//                  GLuint id,
//                  GLenum severity,
//                  GLsizei length,
//                  const GLchar* message,
//                  const void* userParam )
// {
//     fprintf( stderr, "GL CALLBACK: %s type = 0x%x, severity = 0x%x, message = %s\n",
//              ( type == GL_DEBUG_TYPE_ERROR ? "** GL ERROR **" : "" ),
//               type, severity, message );
// }


// Data stored per platform window
struct WGL_WindowData {
	HDC hDC;
};

bool CreateDeviceWGL(HWND hWnd, WGL_WindowData* data);
void CleanupDeviceWGL(HWND hWnd, WGL_WindowData* data);
void ResetDeviceWGL();

// Data
static HGLRC g_hRC;
static WGL_WindowData g_MainWindow;
static int g_Width;
static int g_Height;

static void Hook_Renderer_CreateWindow(ImGuiViewport* viewport) {
	assert(viewport->RendererUserData == NULL);

	WGL_WindowData* data = IM_NEW(WGL_WindowData);
	CreateDeviceWGL((HWND)viewport->PlatformHandle, data);
	viewport->RendererUserData = data;
}

static void Hook_Renderer_DestroyWindow(ImGuiViewport* viewport) {
	if (viewport->RendererUserData != NULL) {
		WGL_WindowData* data = (WGL_WindowData*)viewport->RendererUserData;
		CleanupDeviceWGL((HWND)viewport->PlatformHandle, data);
		IM_DELETE(data);
		viewport->RendererUserData = NULL;
	}
}

static void Hook_Platform_RenderWindow(ImGuiViewport* viewport, void*) {
	// Activate the platform window DC in the OpenGL rendering context
	if (WGL_WindowData* data = (WGL_WindowData*)viewport->RendererUserData)
		wglMakeCurrent(data->hDC, WEngine->wgl->glContext);
}

static void Hook_Renderer_SwapBuffers(ImGuiViewport* viewport, void*) {
	if (WGL_WindowData* data = (WGL_WindowData*)viewport->RendererUserData)
		::SwapBuffers(data->hDC);
}

void WGL::init_modern_gl() {
	// ----------------- RenderDoc ----------------- //
	//LoadModule("renderdoc.dll",0);
	// TODO - IF EDITOR

	// ----------------- OPENGL ----------------- //
	//wglCreateContextAttribsARB = (wglCreateContextAttribsARB_type*)wglGetProcAddress(
	//	"wglCreateContextAttribsARB");
	////wglChoosePixelFormatARB = (wglChoosePixelFormatARB_type*)wglGetProcAddress(
	////	"wglChoosePixelFormatARB");
	//PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = nullptr;
	//wglChoosePixelFormatARB = reinterpret_cast<PFNWGLCHOOSEPIXELFORMATARBPROC>(wglGetProcAddress("wglChoosePixelFormatARB"));

	// TODO: move this block into editor?
	//state = State();
	//editor = Editor(state);
	// create temporary window
	HINSTANCE hInstance = GetModuleHandle(0);

	LPCSTR CLASS_NAME = "LeviathanClass";


	WNDCLASS wc = {};
	wc.cbClsExtra = 0;                        // No Extra Window Data
	wc.cbWndExtra = 0;                        // No Extra Window Data
	wc.hIcon = LoadIcon(NULL, IDI_WINLOGO);   // Load The Default Icon
	wc.hCursor = LoadCursor(NULL, IDC_ARROW); // Load The Arrow Pointer
	wc.hbrBackground = NULL;                  // No Background Required For GL
	wc.lpszMenuName = NULL;                   // We Don't Want A Menu
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.hInstance = hInstance;
	wc.lpszClassName = CLASS_NAME;

	RegisterClass(&wc);

	HWND fakeWND = CreateWindow(
		wc.lpszClassName, "Fake Window",
		WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_SYSMENU,
		0, 0,             // position x, y
		1, 1,             // width, height
		NULL, NULL,       // parent window, menu
		hInstance, NULL); // instance, param

	HDC fakeDC = GetDC(fakeWND); // Device Context

	PIXELFORMATDESCRIPTOR fakePFD;
	ZeroMemory(&fakePFD, sizeof(fakePFD));
	fakePFD.nSize = sizeof(fakePFD);
	fakePFD.nVersion = 1;
	fakePFD.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	fakePFD.iPixelType = PFD_TYPE_RGBA;
	fakePFD.cColorBits = 32;
	fakePFD.cAlphaBits = 8;
	fakePFD.cDepthBits = 24;

	const int fakePFDID = ChoosePixelFormat(fakeDC, &fakePFD);
	if (fakePFDID == 0) {
		return;
	}

	if (SetPixelFormat(fakeDC, fakePFDID, &fakePFD) == false) {
		return;
	}

	HGLRC fakeRC = wglCreateContext(fakeDC); // Rendering Contex

	if (fakeRC == 0) {
		return;
	}

	if (wglMakeCurrent(fakeDC, fakeRC) == false) {
		return;
	}

	glewExperimental = true;
	if (glewInit() != GLEW_OK) {
		MessageBox(NULL, "Error Loading Glew", "", 0x00000000L);
		ExitProcess(0); // I don't think I close properly here
	}

	PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = nullptr;
	wglChoosePixelFormatARB = reinterpret_cast<PFNWGLCHOOSEPIXELFORMATARBPROC>(wglGetProcAddress(
		"wglChoosePixelFormatARB"));
	if (wglChoosePixelFormatARB == nullptr) {
		return;
	}

	PFNWGLCREATECONTEXTATTRIBSARBPROC wglCreateContextAttribsARB = nullptr;
	wglCreateContextAttribsARB = reinterpret_cast<PFNWGLCREATECONTEXTATTRIBSARBPROC>(wglGetProcAddress(
		"wglCreateContextAttribsARB"));
	if (wglCreateContextAttribsARB == nullptr) {
		return;
	}

	ImGui_ImplWin32_EnableDpiAwareness();

	RECT wr = {0, 0, (long)WEngine->RESX, (long)WEngine->RESY}; // set the size, but not the position
	AdjustWindowRect(&wr, WS_OVERLAPPEDWINDOW, FALSE);          // adjust the size

	
	int screenWidth = GetSystemMetrics(SM_CXSCREEN);
	int screenHeight = GetSystemMetrics(SM_CYSCREEN);
	int windowWidth = WEngine->RESX; // Adjust as needed
	int windowHeight = WEngine->RESY; // Adjust as needed
	
	int xPos = (screenWidth - windowWidth) / 2;
	int yPos = (screenHeight - windowHeight) / 2;

	this->hwnd = CreateWindow(CLASS_NAME, 0,
		// WS_POPUP | WS_VISIBLE | WS_POPUP | WS_VISIBLE | WS_CAPTION,
		WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_SYSMENU,
		xPos, yPos, wr.right - wr.left, wr.bottom - wr.top, 0, 0, 0, 0);
	
	// SendMessage(hwnd, WM_SETICON, 0, NULL);
	// SendMessage(hwnd, WM_SETICON, 1, NULL);
	
	
	BOOL USE_DARK_MODE = true;
	BOOL SET_IMMERSIVE_DARK_MODE_SUCCESS = SUCCEEDED(DwmSetWindowAttribute(
			this->hwnd, 20,
			&USE_DARK_MODE, sizeof(USE_DARK_MODE)));
	
	
	
	// SetWindowPos(this->hwnd, 0, xPos, yPos, (long)WEngine->RESX, (long)WEngine->RESY, SWP_NOMOVE | SWP_NOSIZE);

	hDC = GetDC(hwnd);

	// --------------------------------------------------------- //

	const int pixelAttribs[] = {
		WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
		WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
		WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
		WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
		WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB, 32,
		WGL_ALPHA_BITS_ARB, 8,
		WGL_DEPTH_BITS_ARB, 24,
		WGL_STENCIL_BITS_ARB, 8,
		WGL_SAMPLE_BUFFERS_ARB, GL_TRUE,
		WGL_SAMPLES_ARB, 4,
		0
	};
	int pixelFormatID;
	UINT numFormats;
	const bool status = wglChoosePixelFormatARB(hDC, pixelAttribs, NULL, 1, &pixelFormatID, &numFormats);

	if (status == false || numFormats == 0) {
		return;
	}

	PIXELFORMATDESCRIPTOR PFD;
	DescribePixelFormat(hDC, pixelFormatID, sizeof(PFD), &PFD);
	SetPixelFormat(hDC, pixelFormatID, &PFD);

	const int major_min = 4, minor_min = 6;
	const int contextAttribs[] = {
		WGL_CONTEXT_MAJOR_VERSION_ARB, major_min,
		WGL_CONTEXT_MINOR_VERSION_ARB, minor_min,
		WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
		//		WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_DEBUG_BIT_ARB,
		0
	};

	this->glContext = wglCreateContextAttribsARB(hDC, 0, contextAttribs);
	if (this->glContext == NULL) {
		return;
	}

	// Delete temp //
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(fakeRC);
	ReleaseDC(fakeWND, fakeDC);
	DestroyWindow(fakeWND);
	
	// --------------------------------------------------------- //


	g_MainWindow.hDC = hDC;

	if (!wglMakeCurrent(hDC, this->glContext)) {
		return;
	}


	// EnableMenuItem(GetSystemMenu(this->hwnd, FALSE), SC_CLOSE,
	// 						 MF_BYCOMMAND | MF_ENABLED);
	SetWindowLong(hwnd, GWL_STYLE, GetWindowLong(hwnd, GWL_STYLE) | WS_MINIMIZEBOX);
	
	// // Switch to fullscreen mode
	// DEVMODE devMode = { 0 };
	// devMode.dmSize = sizeof(DEVMODE);
	// devMode.dmPelsWidth = screenWidth;
	// devMode.dmPelsHeight = screenHeight;
	// devMode.dmBitsPerPel = 32;
	// devMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	//
	// // Change display settings to fullscreen
	// if (ChangeDisplaySettings(&devMode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) {
	// 	MessageBox(NULL, "Failed to switch to fullscreen mode!", "Error", MB_ICONERROR);
	// 	return;
	// }
	

	


	
	g_hRC = this->glContext;

	// SetThreadDpiAwarenessContext(DPI_AWARENESS_INVALID);
	// SetProcessDpiAwareness(PROCESS_PER_MONITOR_DPI_AWARE)
	SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2);

	TracyGpuContext;
	// ________ NEW SHIT ________ //


	// #if FULLSCREEN
	// 	ChangeDisplaySettings(&screenSettings, CDS_FULLSCREEN);
	// 	ShowCursor(0);
	// 	const HDC hDC = GetDC(CreateWindow((LPCSTR)0xC018, 0, WS_POPUP | WS_VISIBLE | WS_MAXIMIZE, 0, 0, 0, 0, 0, 0, 0, 0));
	// 	SetPixelFormat(hDC, ChoosePixelFormat(hDC, &pfd), &pfd);
	// 	wglMakeCurrent(hDC, wglCreateContext(hDC));
	// #else

	#ifdef GL_DEBUG_ENABLED
	glEnable(GL_DEBUG_OUTPUT);
	glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
	glDebugMessageCallback(Editor::GLErrorCallback, 0);
	#endif
	// glDebugMessageCallback(MessageCallback, 0);

	// initalize IMGUI context
	IMGUI_CHECKVERSION();
	ImGui::CreateContext();
	WEngine->editor->imgui_io = &ImGui::GetIO();
	// ImGuiIO& io = ImGui::GetIO();
	(void)WEngine->editor->imgui_io;
	// WEngine->editor->imgui_io->ConfigFlags |= ImGuiConfigFlags_ViewportsEnable;
	WEngine->editor->imgui_io->ConfigFlags |= ImGuiConfigFlags_DockingEnable;
	WEngine->editor->imgui_io->ConfigFlags |= ImGuiConfigFlags_NavNoCaptureKeyboard;

	ImGui::StyleColorsDark();
	ImGuiStyle& style = ImGui::GetStyle();
	style.WindowRounding = 0.0f;
	WEngine->editor->setImGuiTheme();


	//editor->setImGuiTheme();

	//_hwnd = hwnd;

	// SetWindowText(this->hwnd, reinterpret_cast<LPCSTR>(glGetString(GL_VERSION)));
	SetWindowTextA(this->hwnd, "Levwrightian");
	

	// SetWindowTextA(
	// 	[in]           HWND   hWnd,
	// 	[in, optional] LPCSTR lpString
	// )
					

	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);
	//glGetIntegerv()


	// ImGui_ImplWin32_Init(this->hwnd);
	ImGui_ImplWin32_InitForOpenGL(this->hwnd);
	ImGui_ImplOpenGL3_Init("#version 430"); // changed from 410 to 430. probably wont breaj everything... right?

	// tracy::GetProfiler().PlotData();

	// tracy::GetProfilerData([](const Tracy::Profiler::ZoneData* zones, uint64_t zoneCount) {
	// 	 // Process the zone data and update your engine's UI
	// 	 // (Example: Print zone names to console)
	// 	 for (uint64_t i = 0; i < zoneCount; ++i) {
	// 			 printf("Zone: %s\n", zones[i].name);
	// 	 }
	// });
	// tracy::GetProfiler().
	// ImGui::SetNextWindowPos(ImVec2(0,0));
	// ImGui::SetNextWindowSize(windowSize);
	// tracy::GetProfiler().GetNextZoneId();
	// tracy::ScopedZone

	

	int w = GetSystemMetrics(SM_CXSCREEN);
	int h = GetSystemMetrics(SM_CYSCREEN);
	WEngine->desktop_fullscreen_res_x = w;
	WEngine->desktop_fullscreen_res_y = h;

	
	int extendedStyle = GetWindowLong(hwnd, GWL_EXSTYLE);
	SetWindowLong(hwnd, GWL_EXSTYLE, extendedStyle | WS_EX_DLGMODALFRAME);
	SendMessage(hwnd, WM_SETICON, ICON_BIG, (LPARAM)NULL);
	SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM)NULL);
	// {
	// 	TCHAR className[256];
	// 	GetClassName(hwnd, className, sizeof(className) / sizeof(TCHAR));
	//
	// 	// Get the HINSTANCE of the window
	// 	HINSTANCE hInstance = (HINSTANCE)GetWindowLongPtr(hwnd, GWLP_HINSTANCE);
	//
	// 	// Get the current WNDCLASS structure for the window class
	// 	WNDCLASS wc;
	// 	if (GetClassInfo(hInstance, className, &wc)) {
	// 		// Set the icon handles to NULL
	// 		wc.hIcon = NULL;
	// 		// wc.hIconSm = NULL;
	//
	// 		// Register the updated WNDCLASS structure
	// 		if (!RegisterClass(&wc)) {
	// 			// Handle error (optional)
	// 			MessageBox(NULL, TEXT("Failed to register window class"), TEXT("Error"), MB_OK | MB_ICONERROR);
	// 			return;
	// 		}
	//
	// 		// Update the class long for the window to remove the icon
	// 		SetClassLongPtr(hwnd, GCLP_HICON, 0);
	// 		SetClassLongPtr(hwnd, GCLP_HICONSM, 0);
	//
	// 		// Force the window to redraw without the icon
	// 		// SetWindowPos(hwnd, NULL, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER | SWP_FRAMECHANGED);
	// 	} else {
	// 		// Handle error (optional)
	// 		MessageBox(NULL, TEXT("Failed to get window class info"), TEXT("Error"), MB_OK | MB_ICONERROR);
	// 	}
	// }
	
	
	if(WEngine->fullscreen) {
		SetWindowLongPtr(this->hwnd, GWL_STYLE, WS_VISIBLE | WS_POPUP);
		SetWindowPos(this->hwnd, HWND_TOP, 0, 0, w, h, SWP_FRAMECHANGED );
	} else {
		if(WEngine->borderless) {
			SetWindowLongPtr(this->hwnd, GWL_STYLE, WS_VISIBLE | WS_POPUP);
		}
		SetWindowPos(this->hwnd, HWND_TOP, xPos, yPos, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED);
	}

	ShowWindow(hwnd, SW_MINIMIZE);
	ShowWindow(hwnd, SW_RESTORE);
	SetForegroundWindow(hwnd);
	SetActiveWindow(hwnd);
	// RedrawWindow(hwnd, NULL, 0, RDW_FRAME | RDW_INVALIDATE | RDW_ALLCHILDREN);
	


	if (WEngine->editor->imgui_io->ConfigFlags & ImGuiConfigFlags_ViewportsEnable) {
		ImGuiPlatformIO& platform_io = ImGui::GetPlatformIO();
		IM_ASSERT(platform_io.Renderer_CreateWindow == NULL);
		IM_ASSERT(platform_io.Renderer_DestroyWindow == NULL);
		IM_ASSERT(platform_io.Renderer_SwapBuffers == NULL);
		IM_ASSERT(platform_io.Platform_RenderWindow == NULL);
		platform_io.Renderer_CreateWindow = Hook_Renderer_CreateWindow;
		platform_io.Renderer_DestroyWindow = Hook_Renderer_DestroyWindow;
		platform_io.Renderer_SwapBuffers = Hook_Renderer_SwapBuffers;
		platform_io.Platform_RenderWindow = Hook_Platform_RenderWindow;
	}

	return;
}


// Helper functions
bool CreateDeviceWGL(HWND hWnd, WGL_WindowData* data) {
	HDC hDc = ::GetDC(hWnd);
	PIXELFORMATDESCRIPTOR pfd = {0};
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;

	const int pf = ::ChoosePixelFormat(hDc, &pfd);
	if (pf == 0)
		return false;
	if (::SetPixelFormat(hDc, pf, &pfd) == FALSE)
		return false;
	::ReleaseDC(hWnd, hDc);

	data->hDC = ::GetDC(hWnd);
	// wglGetCurrentContext()
	// if (!g_hRC)
	// 	g_hRC = wglCreateContext(data->hDC);
	return true;
}


void CleanupDeviceWGL(HWND hWnd, WGL_WindowData* data) {
	wglMakeCurrent(NULL, NULL);
	::ReleaseDC(hWnd, data->hDC);
}
