﻿#pragma once


#include <cmath>
#include <cstdint>
#include <initializer_list>

// constexpr float pi = 3.14159265358979323846;
// constexpr float tau = 2.0 * pi;

#define min(a,b) (a < b ? a : b)
#define max(a,b) (a > b ? a : b)

template <typename genType>
inline genType clamp(genType x, genType edge0, genType edge1) {
	return max(min(x,edge1), edge0);
}

template <typename genType>
inline genType smoothstep(genType edge0, genType edge1, genType x) {
	genType t;  /* Or genDType t; */
	t = clamp<genType>((x - edge0) / (edge1 - edge0), 0.0, 1.0);
	return t * t * (3.0 - 2.0 * t);
}

struct Random {
	std::random_device rd;
	std::mt19937 mt;

	std::uniform_real_distribution<float> dist;

	Random() {
		this->mt = std::mt19937(this->rd());
		this->dist = std::uniform_real_distribution<float>(0.0f, 1.0f);
	}

	int rand_int(int min, int max) {
		return min + int(rand_f()*(float(max - min) - 0.0001));
	}
	float rand_f() {
		return dist(mt);
	}
	glm::vec2 rand_v2() {
		return glm::vec2(rand_f(),rand_f());
	}

	glm::vec3 rand_v3() {
		return glm::vec3(rand_f(),rand_f(),rand_f());
	}

};

		// inline glm::mat3 rot_x(float rad) {
		// 	this->rot_axis(0, rad);
		// }
		//
		// inline glm::mat3  rot_y(float rad) {
		// 	this->rot_axis(1, rad);
		// }
		//
		// inline glm::mat3  rot_z(float rad) {
		// 	this->rot_axis(2, rad);
		// }

		inline glm::mat3 get_rot_mat(int axis, float radians) {
			glm::mat3 r = glm::mat3(
				1,0,0,
				0,1,0,
				0,0,1
			);

			float r1 = cos(radians);
			float r2 = -sin(radians);
			float r3 = -r2;

			float* a = &r[0][0];
			if (axis == 0) {
				a[3 * 1 + 1] *= r1;
				a[3 * 1 + 2] = r2;
				a[3 * 2 + 1] = r3;
				a[3 * 2 + 2] = r1;
			} else if (axis == 1) {
				a[3 * 0 + 0] = r1;
				a[3 * 0 + 2] = -r2;
				a[3 * 2 + 0] = -r3;
				a[3 * 2 + 2] = r1;
			} else if (axis == 2) {
				a[3 * 0 + 0] = r1;
				a[3 * 0 + 1] = r2;
				a[3 * 1 + 0] = r3;
				a[3 * 1 + 1] = r1;
			}
			return r;
		}


template <typename T, int sz>
class Vector {
	public:
		T* vals;
		T* x;
		T* y;
		T* z;

		Vector(Vector& other) {
			this->vals = new T[sz];

			for (int i = 0; i < 3; i++) {
				this->vals[i] = other.vals[i];
			}
		}

		Vector(std::initializer_list<T> input_vals) {
			// sz = input_vals.size();

			this->vals = new T[sz];

			int i = 0;
			for (T elem : input_vals) {
				this->vals[i] = elem;
				if (i == 0) {
					this->x = &this->vals[i];
				} else if (i == 1) {
					this->y = &this->vals[i];
				} else if (i == 2) {
					this->z = &this->vals[i];
				}
				i++;
			}
		}

		Vector operator+(Vector const& other) {
			Vector t(*this);
			t += other;
			return t;
		}

		void operator+=(Vector<T, sz> const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] += other.vals[i];
			}
		}

		void operator*=(Vector<T, sz> const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] *= other.vals[i];
			}
		}

		void operator/=(Vector<T, sz> const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] /= other.vals[i];
			}
		}

		void operator-=(Vector<T, sz> const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] -= other.vals[i];
			}
		}

		Vector operator+(T const& other) {
			Vector t(*this);
			t += other;
			return t;
		}

		Vector operator-(T const& other) {
			Vector t(*this);
			t -= other;
			return t;
		}

		Vector operator*(T const& other) {
			Vector t = Vector(*this);
			t *= other;
			return t;
		}

		Vector operator/(T const& other) {
			Vector t(*this);
			t /= other;
			return t;
		}

		void operator+=(T const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] += other;
			}
		}

		void operator*=(T const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] *= other;
			}
		}

		void operator/=(T const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] /= other;
			}
		}

		void operator-=(T const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] -= other;
			}
		}

		T& operator[](int i) const { return &this->vals[i]; }
		T operator[](int i) { return this->vals[i]; }
		// T operator[](int i) { return *this->vals[i]; }

		Vector& operator=(Vector const& other) {
			for (int i = 0; i < sz; i++) {
				this->vals[i] = other.vals[i];
			}
			return *this;
		}

		void normalize() {
			*this /= this->length();
		}

		T length() {
			return sqrt(this->len_sq());
		}

		T len_sq() {
			T res = 0;
			for (int i = 0; i < sz; i++) {
				res += this->vals[i] * this->vals[i];
			}
			return res;
		}

		T dot(Vector const& other) {
			T res = 0;
			for (int i = 0; i < sz; i++) {
				res += this->vals[i] * other.vals[i];
			}
			return res;
		}

		void cross(Vector const& other) {
			T _x = (*this->y) * (*other.z) - (*this->z) * (*other.y);
			T _y = (*this->z) * (*other.x) - (*this->x) * (*other.z);
			T _z = (*this->x) * (*other.y) - (*this->y) * (*other.x);
			this->vals[0] = _x;
			this->vals[1] = _y;
			this->vals[2] = _z;
		}

		void rot_x(float rad) {
			this->rot_axis(0, rad);
		}

		void rot_y(float rad) {
			this->rot_axis(1, rad);
		}

		void rot_z(float rad) {
			this->rot_axis(2, rad);
		}

		void rot_axis(uint32_t axis, float rad) {
			int ax_a;
			int ax_b;
			if (axis == 0) {
				ax_a = 1;
				ax_b = 2;
			} else if (axis == 1) {
				ax_a = 0;
				ax_b = 2;
			} else {
				ax_a = 0;
				ax_b = 1;
			}
			T v_a = (*this)[ax_a];
			T v_b = (*this)[ax_b];

			T ret_a = cos(rad) * v_a - sin(rad) * v_b;
			T ret_b = sin(rad) * v_a + cos(rad) * v_b;
			this->vals[ax_a] = ret_a;
			this->vals[ax_b] = ret_b;
		}

		~Vector() {
			delete[] vals;
		}
};

typedef Vector<float, 3> Vec3;
typedef Vector<int, 3> Vec3i;


template <typename T>
class Matrix {
	public:
		T* vals;

		Matrix() {
			// *this = Matrix({});
			this->vals = new T[16];
			this->identity();
		}
		Matrix(std::initializer_list<T> input_vals) {
			this->vals = new T[16];

			this->identity();

			int i = 0;
			for (T elem : input_vals) {
				this->vals[i] = elem;
				i++;
			}
		}

		void copy_from_other(Matrix& other) {
			for (int i = 0; i < 16; i++) {
				this->vals[i] = other.vals[i];
			}
		}

		void identity() {
			int i = 0;
			for (; i < 16; i++) {
				vals[i] = 0.0;
			}

			vals[0] = 1;
			vals[3 + 2] = 1;
			vals[10] = 1;
			vals[15] = 1;
		}

		~Matrix() {
			delete[] this->vals;
		}

		T& at(size_t col, size_t row) {
			return vals[col * 4 + row];
		}


		void inverse() {
			T a00 = this->vals[0];
			T a01 = this->vals[1];
			T a02 = this->vals[2];
			T a03 = this->vals[3];
			T a10 = this->vals[4];
			T a11 = this->vals[5];
			T a12 = this->vals[6];
			T a13 = this->vals[7];

			T a20 = this->vals[8];
			T a21 = this->vals[9];
			T a22 = this->vals[10];
			T a23 = this->vals[11];

			T a30 = this->vals[12];
			T a31 = this->vals[13];
			T a32 = this->vals[14];
			T a33 = this->vals[15];

			T b00 = a00 * a11 - a01 * a10;
			T b01 = a00 * a12 - a02 * a10;
			T b02 = a00 * a13 - a03 * a10;
			T b03 = a01 * a12 - a02 * a11;
			T b04 = a01 * a13 - a03 * a11;
			T b05 = a02 * a13 - a03 * a12;
			T b06 = a20 * a31 - a21 * a30;
			T b07 = a20 * a32 - a22 * a30;
			T b08 = a20 * a33 - a23 * a30;
			T b09 = a21 * a32 - a22 * a31;
			T b10 = a21 * a33 - a23 * a31;
			T b11 = a22 * a33 - a23 * a32;

			const T det = b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;

			if (det == 0.0) {
				return;
			}

			int idx = 0;
			this->vals[idx++] = (a11 * b11 - a12 * b10 + a13 * b09) / det;
			this->vals[idx++] = (a02 * b10 - a01 * b11 - a03 * b09) / det;
			this->vals[idx++] = (a31 * b05 - a32 * b04 + a33 * b03) / det;
			this->vals[idx++] = (a22 * b04 - a21 * b05 - a23 * b03) / det;
			this->vals[idx++] = (a12 * b08 - a10 * b11 - a13 * b07) / det;
			this->vals[idx++] = (a00 * b11 - a02 * b08 + a03 * b07) / det;
			this->vals[idx++] = (a32 * b02 - a30 * b05 - a33 * b01) / det;
			this->vals[idx++] = (a20 * b05 - a22 * b02 + a23 * b01) / det;
			this->vals[idx++] = (a10 * b10 - a11 * b08 + a13 * b06) / det;
			this->vals[idx++] = (a01 * b08 - a00 * b10 - a03 * b06) / det;
			this->vals[idx++] = (a30 * b04 - a31 * b02 + a33 * b00) / det;
			this->vals[idx++] = (a21 * b02 - a20 * b04 - a23 * b00) / det;
			this->vals[idx++] = (a11 * b07 - a10 * b09 - a12 * b06) / det;
			this->vals[idx++] = (a00 * b09 - a01 * b07 + a02 * b06) / det;
			this->vals[idx++] = (a31 * b01 - a30 * b03 - a32 * b00) / det;
			this->vals[idx++] = (a20 * b03 - a21 * b01 + a22 * b00) / det;
		}

		void mul(const Matrix& other) {
			T result[16];
			T* a = this->vals;
			T* b = other.vals;
			result[0] = a[0] * b[0] + a[4] * b[1] + a[8] * b[2] + a[12] * b[3];
			result[1] = a[1] * b[0] + a[5] * b[1] + a[9] * b[2] + a[13] * b[3];
			result[2] = a[2] * b[0] + a[6] * b[1] + a[10] * b[2] + a[14] * b[3];
			result[3] = a[3] * b[0] + a[7] * b[1] + a[11] * b[2] + a[15] * b[3];
			result[4] = a[0] * b[4] + a[4] * b[5] + a[8] * b[6] + a[12] * b[7];
			result[5] = a[1] * b[4] + a[5] * b[5] + a[9] * b[6] + a[13] * b[7];
			result[6] = a[2] * b[4] + a[6] * b[5] + a[10] * b[6] + a[14] * b[7];
			result[7] = a[3] * b[4] + a[7] * b[5] + a[11] * b[6] + a[15] * b[7];
			result[8] = a[0] * b[8] + a[4] * b[9] + a[8] * b[10] + a[12] * b[11];
			result[9] = a[1] * b[8] + a[5] * b[9] + a[9] * b[10] + a[13] * b[11];
			result[10] = a[2] * b[8] + a[6] * b[9] + a[10] * b[10] + a[14] * b[11];
			result[11] = a[3] * b[8] + a[7] * b[9] + a[11] * b[10] + a[15] * b[11];
			result[12] = a[0] * b[12] + a[4] * b[13] + a[8] * b[14] + a[12] * b[15];
			result[13] = a[1] * b[12] + a[5] * b[13] + a[9] * b[14] + a[13] * b[15];
			result[14] = a[2] * b[12] + a[6] * b[13] + a[10] * b[14] + a[14] * b[15];
			result[15] = a[3] * b[12] + a[7] * b[13] + a[11] * b[14] + a[15] * b[15];

			this->vals[0] = result[0];
			this->vals[1] = result[1];
			this->vals[2] = result[2];
			this->vals[3] = result[3];
			this->vals[4] = result[4];
			this->vals[5] = result[5];
			this->vals[6] = result[6];
			this->vals[7] = result[7];
			this->vals[8] = result[8];
			this->vals[9] = result[9];
			this->vals[10] = result[10];
			this->vals[11] = result[11];
			this->vals[12] = result[12];
			this->vals[13] = result[13];
			this->vals[14] = result[14];
			this->vals[15] = result[15];

		}

		// Matrix operator*(const Matrix& other) {
		// 	Matrix res({});
		// 	//
		// 	// for (size_t row_idx_this = 0; row_idx_this < 4; row_idx_this++) {
		// 	// 	for (size_t col_idx_other = 0; col_idx_other < 4; col_idx_other++) {
		// 	// 		for (size_t k = 0; k < 4; k++) { // cols
		// 	// 			result.at(row_idx_this, col_idx_other) += this->at(row_idx_this, k) * other.at(k, col_idx_other);
		// 	// 		}
		// 	// 	}
		// 	// }
		// 	T* result = res.vals;
		// 	T* a = this->vals;
		// 	T* b = other.vals;
		// 	result[0] = a[0] * b[0] + a[4] * b[1] + a[8] * b[2] + a[12] * b[3];
		// 	result[1] = a[1] * b[0] + a[5] * b[1] + a[9] * b[2] + a[13] * b[3];
		// 	result[2] = a[2] * b[0] + a[6] * b[1] + a[10] * b[2] + a[14] * b[3];
		// 	result[3] = a[3] * b[0] + a[7] * b[1] + a[11] * b[2] + a[15] * b[3];
		// 	result[4] = a[0] * b[4] + a[4] * b[5] + a[8] * b[6] + a[12] * b[7];
		// 	result[5] = a[1] * b[4] + a[5] * b[5] + a[9] * b[6] + a[13] * b[7];
		// 	result[6] = a[2] * b[4] + a[6] * b[5] + a[10] * b[6] + a[14] * b[7];
		// 	result[7] = a[3] * b[4] + a[7] * b[5] + a[11] * b[6] + a[15] * b[7];
		// 	result[8] = a[0] * b[8] + a[4] * b[9] + a[8] * b[10] + a[12] * b[11];
		// 	result[9] = a[1] * b[8] + a[5] * b[9] + a[9] * b[10] + a[13] * b[11];
		// 	result[10] = a[2] * b[8] + a[6] * b[9] + a[10] * b[10] + a[14] * b[11];
		// 	result[11] = a[3] * b[8] + a[7] * b[9] + a[11] * b[10] + a[15] * b[11];
		// 	result[12] = a[0] * b[12] + a[4] * b[13] + a[8] * b[14] + a[12] * b[15];
		// 	result[13] = a[1] * b[12] + a[5] * b[13] + a[9] * b[14] + a[13] * b[15];
		// 	result[14] = a[2] * b[12] + a[6] * b[13] + a[10] * b[14] + a[14] * b[15];
		// 	result[15] = a[3] * b[12] + a[7] * b[13] + a[11] * b[14] + a[15] * b[15];
		// 	return res;
		// }

		void get_rotation_matrix(int axis, float radians) {
			this->identity();

			T r1 = cos(radians);
			T r2 = -sin(radians);
			T r3 = -r2;
			// val a = floatArrayOf(1f, 0f, 0f, 0f, 0f, 1f, 0f, 0f, 0f, 0f, 1f, 0f, 0f, 0f, 0f, 1f)

			T* a = this->vals;
			if (axis == 0) {
				a[4 * 1 + 1] *= r1;
				a[4 * 1 + 2] = r2;
				a[4 * 2 + 1] = r3;
				a[4 * 2 + 2] = r1;
			} else if (axis == 1) {
				a[4 * 0 + 0] = r1;
				a[4 * 0 + 2] = -r2;
				a[4 * 2 + 0] = -r3;
				a[4 * 2 + 2] = r1;
			} else if (axis == 2) {
				a[4 * 0 + 0] = r1;
				a[4 * 0 + 1] = r2;
				a[4 * 1 + 0] = r3;
				a[4 * 1 + 1] = r1;
			}
		}

		void get_transformation_matrix(float x, float y, float z) {
			this->at(0, 0) = 1;
			this->at(0, 1) = 0;
			this->at(0, 2) = 0;
			this->at(0, 3) = 0;

			this->at(1, 0) = 0;
			this->at(1, 1) = 1;
			this->at(1, 2) = 0;
			this->at(1, 3) = 0;

			this->at(2, 0) = 0;
			this->at(2, 1) = 0;
			this->at(2, 2) = 1;
			this->at(2, 3) = 0;

			this->at(3, 0) = x;
			this->at(3, 1) = y;
			this->at(3, 2) = z;
			this->at(3, 3) = 1;
		}

		void get_look_at_matrix(glm::vec3 right, glm::vec3 up, glm::vec3 dir, glm::vec3 pos) {
			this->at(0, 0) = right[0];
			this->at(1, 0) = right[1];
			this->at(2, 0) = right[2];
			this->at(3, 0) = 0;

			this->at(0, 1) = -up[0];
			this->at(1, 1) = -up[1];
			this->at(2, 1) = -up[2];
			this->at(3, 1) = 0;

			this->at(0, 2) = -dir[0];
			this->at(1, 2) = -dir[1];
			this->at(2, 2) = -dir[2];
			this->at(3, 2) = 0;

			this->at(0, 3) = 0;
			this->at(1, 3) = 0;
			this->at(2, 3) = 0;
			this->at(3, 3) = 1.0;;

			// this->at(0, 0) = -right[0];
			// this->at(1, 0) = -right[1];
			// this->at(2, 0) = -right[2];
			// this->at(3, 0) = 0;
			//
			// this->at(0, 1) = -up[0];
			// this->at(1, 1) = -up[1];
			// this->at(2, 1) = -up[2];
			// this->at(3, 1) = 0;
			//
			//
			// this->at(0, 2) = -dir[0];
			// this->at(1, 2) = -dir[1];
			// this->at(2, 2) = -dir[2];
			// this->at(3, 2) = 0;
			//
			// this->at(3, 3) = -*pos.x;
			// this->at(3, 3) = -*pos.y;
			// this->at(3, 3) = -*pos.z;
			// this->at(3, 3) = 1.0;;
		}

		void get_perspective_proj_mat(float fov, float _near, float _far, float _xRes, float _yRes) {
			float aspectRatio = 1.0f / (_xRes / _yRes);
			float tanHalfFov = (1.0 / tan(fov));
			float dz = 1.0f / (_far - _near);
			this->at(0, 0) = 1.0f * (tanHalfFov * aspectRatio);
			this->at(0, 1) = 0;
			this->at(0, 2) = 0;
			this->at(0, 3) = 0;

			this->at(1, 0) = 0;
			this->at(1, 1) = tanHalfFov;
			this->at(1, 2) = 0;
			this->at(1, 3) = 0;

			this->at(2, 0) = 0;
			this->at(2, 1) = 0;
			this->at(2, 2) = -(_far + _near) * dz;
			this->at(2, 3) = -1.0;

			this->at(3, 0) = 0;
			this->at(3, 1) = 0;
			this->at(3, 2) = -2.0f * _far * _near * dz;
			this->at(3, 3) = 0;
		}

};
