﻿#include "pch.h"

#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderProgram.h"


class FXPass {
public:
	smallvec<Framebuffer*, 20> fbs;
	
	FXPass() {
		
	}
	
	virtual void resize(uint32_t resx, uint32_t resy) {
		for(auto fb : this->fbs) {
			fb->resize(resx,resy);
		}
	}
};

class FXBloomPass : public FXPass {
public:
	Framebuffer* fb;
	
	ShaderProgram* shader_program;

	float uThreshold = 0.07;
	float uRamp = 0.35;
	float uAmount = 0.;

	int iteration_cnt = 9;
	
	FXBloomPass() {
		this->shader_program = new ShaderProgram({
			WEngine->vert_sh_fullscreen_quad,
			new Shader("engine/boxblur.frag")
		});

		auto format = InternalFormat::RGBA32F;
	
		uint32_t res_x = WEngine->RESX;
		uint32_t res_y = WEngine->RESY;
		for (int i = 0; i < this->iteration_cnt; i++) {
			res_x /= 2;
			res_y /= 2;
			this->fbs.push_back(
				new Framebuffer({{
						WEngine->alloc_textures.push(TextureDesc{res_x, res_y, 1, format}).ptr
					}
				})
			);
		}
	}

	void run (Texture* input_tex, Framebuffer* output_fb) {
		glDisable(GL_DEPTH_TEST);
		shader_program->use();

		shader_program->setUniform("uThreshold", this->uThreshold);
		shader_program->setUniform("uRamp", this->uRamp);
		shader_program->setUniform("uAmount", this->uAmount);

		shader_program->setUniform("uIterationCnt", this->iteration_cnt);
		shader_program->setUniform("uIsDownsampling", 1);
		shader_program->setUniform("uIsUpsampling", 0);
		for(int i = 0; i < this->iteration_cnt; i++) {
			shader_program->setUniform("uCurrIter", i);
			glDisable(GL_BLEND);
			if (i == 0) {
				shader_program->setUniform("s_InputTex", input_tex);
			} else {
					shader_program->setUniform("s_InputTex", (*(this->fbs[i-1])->get_front_textures())[0]);
			}
			this->fbs[i]->bind();
			this->fbs[i]->clear();
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}

		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE,GL_ONE);

		shader_program->setUniform("uIsDownsampling", 0);
		shader_program->setUniform("uIsUpsampling", 1);

		for(int i = 0; i < this->iteration_cnt - 1; i++) {
			shader_program->setUniform("uCurrIter", this->iteration_cnt + i);
			shader_program->setUniform("s_InputTex", (*(this->fbs[this->iteration_cnt - 1 - i])->get_front_textures())[0]);
			this->fbs[this->iteration_cnt - 2 - i]->bind();
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}


		shader_program->setUniform("uCurrIter", this->iteration_cnt*2 - 1);

		glDisable(GL_BLEND);
		
		output_fb->bind();
		output_fb->clear();

		// _inputTexture.setUniform("s_OrigTex")
		// shader_program->setUniform("s_OrigTex", this->iteration_cnt + i);
		shader_program->setUniform("s_OrigTex", input_tex);
		shader_program->setUniform("s_InputTex", (*(this->fbs[0])->get_front_textures())[0]);
		glDrawArrays(GL_TRIANGLES, 0, 6);
		glDisable(GL_BLEND);
	}

	void resize(uint32_t resx, uint32_t resy) override {
		// FXPass::resize(resx, resy);
		for (int i = 0; i < this->iteration_cnt; i++) {
			resx /= 2;
			resy /= 2;
			this->fbs[i]->resize(resx, resy);
			// this->fbs.push_back(
			// 	new Framebuffer({{
			// 			WEngine->alloc_textures.push(TextureDesc{res_x, res_y, 1, format}).ptr
			// 		}
			// 	})
			// );
		}
	}
};

//
// FXPass* bloompass;
// FXPass* potass;
// FXPass* amogpass;
//
// FxComposer comp;
//
// comp.add_pass(bloom_pass);
// comp.add_pass(potass);
// comp.add_pass(amogpass);
//
// comp.run = [&]() {
// 	bloom_pass.run();
// 	
// };


class FxComposer {
public:
	glm::uvec2 internal_res;
	smallvec<FXPass*, 20> passes;
	Framebuffer* internal_fb;
	std::function<Texture*(Texture*)> run;

	FxComposer() {
		this->internal_res = glm::uvec2(WEngine->RESX, WEngine->RESY);
		internal_fb = new Framebuffer( FBDesc {WArray({
			WEngine->alloc_textures.push(
				TextureDesc{this->internal_res.x, this->internal_res.y,1, InternalFormat::RGBA32F}
			).ptr
		}), true, false});
	}

	void add_pass(FXPass* pass){
		this->passes.push_back(pass);
	}

	// void run() {
	// 	
	// }

	void pong() {
		this->internal_fb->pong();
	}

	Texture* get_back_tex() {
		return internal_fb->get_back_textures_at(0);
	}
	
	Texture* get_front_fb() {
		return internal_fb->get_back_textures_at(0);
	}

	void resize(uint32_t resx, uint32_t resy) {
		this->internal_res = glm::uvec2(resx, resy);
		for(auto pass: this->passes) {
			pass->resize(resx, resy);
		}
	}
};