﻿#pragma once
#include "engine/definitions.h"
#include "engine/Engine.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderProgram.h"


class PassBloom {
	public:
		Framebuffer* fb;
		WArray<Framebuffer*> fbs;

		ShaderProgram* shader_program;

		float uThreshold = 0.07;
		float uRamp = 0.35;
		float uAmount = 0.;

		int iteration_cnt = 9;

	public:
		PassBloom(Framebuffer* input_fb) {
			this->shader_program = new ShaderProgram({
				WEngine->vert_sh_fullscreen_quad,
				new Shader("engine/boxblur.frag")
			});

			this->fb = new Framebuffer({WArray({
				// input_fb->textures[0]->clone()
				// new Texture(input_fb->textures[0]->resx, input_fb->textures[0]->resy,1, InternalFormat::RGBA16F)
				WEngine->alloc_textures.push(TextureDesc{input_fb->textures[0]->resx, input_fb->textures[0]->resy,1, InternalFormat::RGBA32F}).ptr
			}), false, false});
			this->fbs = WArray<Framebuffer*>({}, this->iteration_cnt);

			uint32_t res_x = this->fb->textures[0]->resx;
			uint32_t res_y = this->fb->textures[0]->resy;
			for (int i = 0; i < this->iteration_cnt; i++) {
				res_x /= 2;
				res_y /= 2;
				this->fbs.add(
					new Framebuffer({{
							WEngine->alloc_textures.push(TextureDesc{res_x, res_y, 1, this->fb->textures[0]->internal_format}).ptr
						}
					})
				);
			}
		}
		void run(Framebuffer* input_fb) {
			glDisable(GL_DEPTH_TEST);
			shader_program->use();

			shader_program->setUniform("uThreshold", this->uThreshold);
			shader_program->setUniform("uRamp", this->uRamp);
			shader_program->setUniform("uAmount", this->uAmount);

			shader_program->setUniform("uIterationCnt", this->iteration_cnt);
			shader_program->setUniform("uIsDownsampling", 1);
			shader_program->setUniform("uIsUpsampling", 0);
			for(int i = 0; i < this->iteration_cnt; i++) {
				shader_program->setUniform("uCurrIter", i);
				glDisable(GL_BLEND);
				if (i == 0) {
					shader_program->setUniform("s_InputTex", (*input_fb->get_front_textures())[0]);
				} else {
						shader_program->setUniform("s_InputTex", (*(this->fbs[i-1])->get_front_textures())[0]);
				}
				this->fbs[i]->bind();
				this->fbs[i]->clear();
				glDrawArrays(GL_TRIANGLES, 0, 6);
			}

			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE,GL_ONE);

			shader_program->setUniform("uIsDownsampling", 0);
			shader_program->setUniform("uIsUpsampling", 1);

			for(int i = 0; i < this->iteration_cnt - 1; i++) {
				shader_program->setUniform("uCurrIter", this->iteration_cnt + i);
				shader_program->setUniform("s_InputTex", (*(this->fbs[this->iteration_cnt - 1 - i])->get_front_textures())[0]);
				this->fbs[this->iteration_cnt - 2 - i]->bind();
				glDrawArrays(GL_TRIANGLES, 0, 6);
			}


			shader_program->setUniform("uCurrIter", this->iteration_cnt*2 - 1);
			glDisable(GL_BLEND);
			this->fb->bind();
			this->fb->clear();

			// _inputTexture.setUniform("s_OrigTex")
			// shader_program->setUniform("s_OrigTex", this->iteration_cnt + i);
			shader_program->setUniform("s_OrigTex", (*input_fb->get_front_textures())[0]);
			shader_program->setUniform("s_InputTex", (*(this->fbs[0])->get_front_textures())[0]);
			glDrawArrays(GL_TRIANGLES, 0, 6);
			glDisable(GL_BLEND);
		}

};
