﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"

static Lane* seq_amogus;
static Curve curve_posoomba;
static Curve curve_posoomba5;	
static Curve curve_posoomba4;	
static Curve curve_posoomba7;	
static Curve curve_posoomba2;	
static Lane* seq_agoomba;
static Curve curve_agoomba;
static Lane* seq_blorgus;
static Lane* seq_blorba;

static Texture* blue_noise_tex;

// static ShaderProgram* prog_render;
static ShaderProgram* post_shader_prog;
static ShaderProgram* vpt_prog;
static ShaderProgram* sort_prog;
// static ShaderProgram* prog_vpt;
// static ShaderProgram* prog_sim_particles;

static AudioFile* demo_audio;

static Framebuffer* fb_main;

static Framebuffer* vpt_fb;

static Buffer* vpt_buff;

static Texture* source_tex;
static Texture* source_tex_edited;
static Texture* vpt_tex;
static Texture* dith_error_tex;
static Texture* dith_output_tex;
static Texture* cumsum_tex;
static Texture* cumsum_tex_b;
static Texture* cumsum_tex_out;
static Texture* sort_tex;


static ShaderProgram* pre_image_program;
static ShaderProgram* particles_program;
static ShaderProgram* dith_program;
static ShaderProgram* cumsum_program;
static Buffer* cumsum_buff;
static Buffer* sort_buff;



// static Model* scene_model = nullptr;


static WVideo* vid;



glm::vec2 get_circ_pos_from_ang(double a) {
	double c = cos(-a);
	double s = sin(-a);
	return glm::vec2(c, s);
}


// Function to add angle offset to position
glm::vec2 add_ang_to_pos(
	glm::vec2 pos, glm::vec2 ang_offs,
  bool positive, double amt, glm::vec2 aspect_correction
) {
	if (positive) {
		pos[0] += ang_offs.x * amt * aspect_correction[0];
		pos[1] += ang_offs.y * amt * aspect_correction[1];
	} else {
		pos[0] -= ang_offs.x * amt * aspect_correction[0];
		pos[1] -= ang_offs.y * amt * aspect_correction[1];
	}
	return pos;
}


enum class BrushStrokeType {
	Blobs, Long, Tri
};

struct DrawParams {
	glm::vec4 tex_lch_dynamics = glm::vec4(0, 0, 0.2, 0.0);
	glm::vec2 tex_stretch = glm::vec2(0.5 + 1 / 20.0f, 0.5 + 1 / 20.0f);
	glm::vec2 tex_distort = glm::vec2(0, 0);
	
	float tex_dynamics = 0.3;
	float tex_distort_amt = 0;
	float tex_grit = 0;
	float pad;

	// blending_colour_space = BlendingColourSpace.Pigments
	int tex_index = 0;
	int blending_colour_space = 1;
	int buff_idx;
	int entries_cnt;
};

// class BrushStroke {
// public:
// 	BrushStrokeType brush_type = BrushStrokeType::Long;
// 	Texture* brush_texture;
// 	
// 	DrawParams draw_params;
//
// 	std::vector<float> positions;
// 	std::vector<float> rotations;
// 	std::vector<float> sizes;
// 	std::vector<float> opacities;
// 	std::vector<float> colours;
//
// 	int idx = 0;
// 	
// 	BrushStroke(
// 		Texture* brush_texture,
// 		DrawParams draw_params = DrawParams()
// 	)
// 		: brush_texture(brush_texture) {
// 		
// 	}
//
// 	void push_stroke(
// 		glm::vec2 pos,
// 		glm::vec2 rotation,
// 		glm::vec2 size,
// 		float opacity,
// 		glm::vec3 colour
// 	) {
// 		int curr_idx = this->idx;
//
// 		this->positions.push_back(0);
// 		this->positions.push_back(0);
// 		
// 		this->rotations.push_back(0);
// 		this->rotations.push_back(0);
// 		
// 		this->sizes.push_back(0);
// 		this->sizes.push_back(0);
// 		
// 		this->opacities.push_back(0);
// 		
// 		this->colours.push_back(0);
// 		this->colours.push_back(0);
// 		this->colours.push_back(0);
//
// 		this->positions[curr_idx * 2] = pos[0];
// 		this->positions[curr_idx * 2 + 1] = pos[1];
// 		// this->positions.push_back(pos[0]);
// 		// this->positions.push_back(pos[1]);
// 		
// 		this->sizes[curr_idx * 2] = size[0];
// 		this->sizes[curr_idx * 2 + 1] = size[1];
// 		
// 		this->rotations[curr_idx * 2] = rotation[0];
// 		this->rotations[curr_idx * 2 + 1] = rotation[1];
// 		
// 		this->opacities[curr_idx] = opacity;
// 		
// 		this->colours[curr_idx * 3 + 0] = colour[0];
// 		this->colours[curr_idx * 3 + 1] = colour[1];
// 		this->colours[curr_idx * 3 + 2] = colour[2];
//
// 		this->idx++;
// 	}
// };

class CurvePainter {
public:
	struct StrokeEntry {
		float posx;
		float posy;
		float rotx;
		float roty;
		
		float sizex;
		float sizey;
		float opacity;
		float colx;
		
		float coly;
		float colz;
		float pad;
		float padb;
	};
	
	Buffer* buffs[2];
	int buffsz_0 = 0;
	int buffsz_1 = 0;

	int stroke_cnt = 0;
	int buff_idx = 0;
	std::vector<DrawParams> draw_params;
	std::vector<StrokeEntry> stroke_entries;
	
	Buffer* param_buff;

	ShaderProgram* curve_drawer_program;
	int param_buff_sz = 0;
	
	CurvePainter() {
		this->param_buff = new Buffer(BuffDesc{.
			byte_len = 10 * megabyte_in_bytes,
			.immutable_subdata = false,
			// .mappable_write = true,
			// .mappable_read = false,
			// .persistennt = false,
			// .write_from_shader = true,
			// .byte_len = 10000,
			// .immutable_subdata = false,
			// .name = "Buff Drawer2D"
			.name = "Curve Painter Params"
		});
		
		float buffsz_0;
		float buffsz_1;
		
		draw_params.resize(10000);
		stroke_entries.resize(100000);
		// std::vector<float> buffs[2];

		this->curve_drawer_program = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_abstractexpressionism/curve_drawer.vert"),
			new Shader("scene_abstractexpressionism/curve_drawer.frag")
		});

		for(int i = 0; i < 2; i++) {
			std::string name = std::string("Curve Drawer ") ;
			name += i;
			
			buffs[i] = new Buffer(BuffDesc{.
				byte_len = 10 * megabyte_in_bytes,
				.immutable_subdata = false,
				// .immutable_subdata = false,
				// .mappable_write = true,
				// .mappable_read = false,
				// .persistennt = false,
				// .write_from_shader = true,
				.name = name
			});
		}

		// ->cpu_data
	}
	

	void begin_stroke(
		Texture* brush_texture,
		DrawParams _draw_params = DrawParams()
	) {
			_draw_params.tex_index = brush_texture->bindless_idx;
			_draw_params.buff_idx = this->buff_idx;
			// this->draw_params.push_back(_draw_params);
			this->draw_params[stroke_cnt] = _draw_params;
			// this->draw_params[stroke_cnt] = _draw_params;
			// this->draw_params[stroke_cnt];
			stroke_cnt++;
	}

	void push_stroke(
		glm::vec2 pos,
		glm::vec2 rotation,
		glm::vec2 size,
		float opacity,
		glm::vec3 colour
	) {
			StrokeEntry entry;
			entry.posx = pos.x;
			entry.posy = pos.y;
		
			entry.rotx = rotation.x;
			entry.roty = rotation.y;
		
			entry.sizex = size.x;
			entry.sizey = size.y;

			entry.opacity = opacity;
		
			entry.colx = colour.x;
			entry.coly = colour.y;
			entry.colz = colour.z;
		
			this->stroke_entries[this->buff_idx++] = entry;
			this->draw_params[this->stroke_cnt - 1].entries_cnt++;
	}


	void flush_strokes() {
		int upload_idx = 0;
		
		// Determine aspect correction
		glm::vec2 aspect_correction;

		float width = (float)WEngine->RESX;
		float height = (float)WEngine->RESY;
		
		if (width > height) {
				aspect_correction[0] = height / width;
				aspect_correction[1] = 1;
		} else {
				aspect_correction[0] = 1;
				aspect_correction[1] = width / height;
		}

		float* stroke_data_ptr = (float*)this->buffs[0]->cpu_data;
		DrawParams* metadata_data_ptr = (DrawParams*)this->buffs[1]->cpu_data;
		
		
		for(int stroke_i = 0; stroke_i < this->stroke_cnt; stroke_i++) {
			DrawParams stroke_draw_params = this->draw_params[stroke_i];

			int i = 0;
			
			for(
				int buff_i = stroke_draw_params.buff_idx;
				buff_i < stroke_draw_params.buff_idx + stroke_draw_params.entries_cnt - 1;
				buff_i++, i++
			){
				StrokeEntry entry = this->stroke_entries[buff_i];
				StrokeEntry next_entry = this->stroke_entries[buff_i + 1];
				
				double curr_sz = entry.sizey * 0.5;
				double next_sz = next_entry.sizey * 0.5;

				
				glm::vec2 curr_ang = get_circ_pos_from_ang(entry.roty);
				glm::vec2 next_ang = get_circ_pos_from_ang(next_entry.roty);

				glm::vec2 curr_pos = glm::vec2(entry.posx, entry.posy);
				glm::vec2 next_pos = glm::vec2(next_entry.posx, next_entry.posy);


				glm::vec2 curr_pos_left = add_ang_to_pos(curr_pos, curr_ang, true, curr_sz, aspect_correction);
				glm::vec2 curr_pos_right = add_ang_to_pos(curr_pos, curr_ang, false, curr_sz, aspect_correction);

				glm::vec2 next_pos_left = add_ang_to_pos(next_pos, next_ang, true, next_sz, aspect_correction);
				glm::vec2 next_pos_right = add_ang_to_pos(next_pos, next_ang, false, next_sz, aspect_correction);

			
				glm::vec3 curr_col = {entry.colx, entry.coly, entry.colz};
				
				double curr_opacity = entry.opacity;

				glm::vec3 next_col = {next_entry.colx, next_entry.coly, next_entry.colz};
				double next_opacity = next_entry.opacity;

				double curr_v = double(i) / float(stroke_draw_params.entries_cnt);
				double next_v = double(i + 1) / float(stroke_draw_params.entries_cnt);

				*metadata_data_ptr++ = stroke_draw_params;
				*metadata_data_ptr++ = stroke_draw_params;
				*metadata_data_ptr++ = stroke_draw_params;
				
				*metadata_data_ptr++ = stroke_draw_params;
				*metadata_data_ptr++ = stroke_draw_params;
				*metadata_data_ptr++ = stroke_draw_params;
				
				// data_ptr
				stroke_data_ptr[upload_idx++] = curr_pos_left[0];
				stroke_data_ptr[upload_idx++] = curr_pos_left[1];
				stroke_data_ptr[upload_idx++] = curr_col[0];
				stroke_data_ptr[upload_idx++] = curr_col[1];
				stroke_data_ptr[upload_idx++] = curr_col[2];
				stroke_data_ptr[upload_idx++] = 0;
				stroke_data_ptr[upload_idx++] = curr_v;
				stroke_data_ptr[upload_idx++] = curr_opacity;

				stroke_data_ptr[upload_idx++] = curr_pos_right[0];
				stroke_data_ptr[upload_idx++] = curr_pos_right[1];
				stroke_data_ptr[upload_idx++] = curr_col[0];
				stroke_data_ptr[upload_idx++] = curr_col[1];
				stroke_data_ptr[upload_idx++] = curr_col[2];
				stroke_data_ptr[upload_idx++] = 1;
				stroke_data_ptr[upload_idx++] = curr_v;
				stroke_data_ptr[upload_idx++] = curr_opacity;

				stroke_data_ptr[upload_idx++] = next_pos_left[0];
				stroke_data_ptr[upload_idx++] = next_pos_left[1];
				stroke_data_ptr[upload_idx++] = next_col[0];
				stroke_data_ptr[upload_idx++] = next_col[1];
				stroke_data_ptr[upload_idx++] = next_col[2];
				stroke_data_ptr[upload_idx++] = 0;
				stroke_data_ptr[upload_idx++] = next_v;
				stroke_data_ptr[upload_idx++] = next_opacity;

				stroke_data_ptr[upload_idx++] = curr_pos_right[0];
				stroke_data_ptr[upload_idx++] = curr_pos_right[1];
				stroke_data_ptr[upload_idx++] = curr_col[0];
				stroke_data_ptr[upload_idx++] = curr_col[1];
				stroke_data_ptr[upload_idx++] = curr_col[2];
				stroke_data_ptr[upload_idx++] = 1;
				stroke_data_ptr[upload_idx++] = curr_v;
				stroke_data_ptr[upload_idx++] = curr_opacity;

				stroke_data_ptr[upload_idx++] = next_pos_left[0];
				stroke_data_ptr[upload_idx++] = next_pos_left[1];
				stroke_data_ptr[upload_idx++] = next_col[0];
				stroke_data_ptr[upload_idx++] = next_col[1];
				stroke_data_ptr[upload_idx++] = next_col[2];
				stroke_data_ptr[upload_idx++] = 0;
				stroke_data_ptr[upload_idx++] = next_v;
				stroke_data_ptr[upload_idx++] = next_opacity;

				stroke_data_ptr[upload_idx++] = next_pos_right[0];
				stroke_data_ptr[upload_idx++] = next_pos_right[1];
				stroke_data_ptr[upload_idx++] = next_col[0];
				stroke_data_ptr[upload_idx++] = next_col[1];
				stroke_data_ptr[upload_idx++] = next_col[2];
				stroke_data_ptr[upload_idx++] = 1;
				stroke_data_ptr[upload_idx++] = next_v;
				stroke_data_ptr[upload_idx++] = next_opacity;
			}
		}

		this->buffs[0]->upload_sub_data(
			&this->buffs[0]->cpu_data[0],
			upload_idx * 4
		);
		
		this->buffs[1]->upload_sub_data(
			&this->buffs[1]->cpu_data[0],
			((uint8_t*)metadata_data_ptr) - (uint8_t*)(&this->buffs[1]->cpu_data[0])
		);

		this->curve_drawer_program->use();
		this->curve_drawer_program->setUniform("verts", this->buffs[0]);
		this->curve_drawer_program->setUniform("params", this->buffs[1]);
		
			// DrawParams stroke_draw_params = this->draw_params[stroke_i];

		glDrawArrays(
			GL_TRIANGLES, 
			0,
			upload_idx/8
		);

		this->stroke_cnt = 0;
		this->buff_idx = 0;
	}
};



static CurvePainter* curve_painter;

void VarsAbstractExpressionism::init() {
	VarsAbstractExpressionism& v = *this;
	Engine* e = WEngine;
	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
	// source_tex = WEngine->alloc_textures.push(std::string("src/assets/images/bababooey.png")).ptr;
	// source_tex = WEngine->alloc_textures.push(std::string("C:\\Users\\user\\Downloads\\1093231088350019656.png")).ptr;
	source_tex = WEngine->alloc_textures.push(std::string("F:\\Programming\\Design\\_Resources\\dalle\\finished\\DALL·E 2022-07-15 19.19.03 - A portrait of a woman, wearing sunglasses, looking back at the camera. Her body is covered with a deep blue robe. Kodak Ektachrome 35mm. (1).png")).ptr;
	
	source_tex_edited = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = 1024, .resy = 512, .resz = 1,
				.internal_format = InternalFormat::RGBA32F,
				.is_mipmapped = true,
				// .mip_levels = 
				.tex_access = TexAccess::RW
		}
	).ptr;
	
	vpt_tex = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA16UI,
				// .mip_levels = 
				.tex_access = TexAccess::RW
		}
	).ptr;

	glm::uvec2 dith_res {WEngine->RESX/2, WEngine->RESY/2};

	dith_error_tex = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = dith_res.x, .resy = dith_res.y, .resz = 1,
				.internal_format = InternalFormat::R32F,
				// .mip_levels = 
				.tex_access = TexAccess::RW
		}
	).ptr;
	
	
	dith_output_tex = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = dith_res.x, .resy = dith_res.y, .resz = 1,
				.internal_format = InternalFormat::R32F,
				// .mip_levels = 
				.tex_access = TexAccess::RW
		}
	).ptr;

	TextureDesc cumsum_texture_desc = TextureDesc{
		.resx = e->RESX + 2, .resy = e->RESY, .resz = 1,
		.internal_format = InternalFormat::R32F,
		// .mip_levels = 
		.tex_access = TexAccess::RW
	};
	
	cumsum_tex = WEngine->alloc_textures.push( cumsum_texture_desc ).ptr;
	cumsum_tex_b = WEngine->alloc_textures.push( cumsum_texture_desc ).ptr;
	cumsum_tex_out = WEngine->alloc_textures.push( TextureDesc{
		.resx = e->RESX, .resy = e->RESY, .resz = 1,
		.internal_format = InternalFormat::RGBA16F,
		// .mip_levels = 
		.tex_access = TexAccess::RW
	} ).ptr;
	
	sort_tex = WEngine->alloc_textures.push( TextureDesc{
		.resx = e->RESX, .resy = e->RESY, .resz = 1,
		.internal_format = InternalFormat::RGBA16F,
		// .mip_levels = 
		.tex_access = TexAccess::RW
	} ).ptr;

	
	
	
	vpt_buff = new Buffer(BuffDesc{.
		byte_len = 3 * megabyte_in_bytes,
		.immutable_subdata = true,
		.mappable_write = true,
		.mappable_read = false,
		.persistennt = false,
		.write_from_shader = true,
		.name = "VPT buff"
	});
	
	cumsum_buff = new Buffer(BuffDesc{.
		byte_len = 40000 * 8, // could be smaller lolol
		.immutable_subdata = true,
		.mappable_write = true,
		.mappable_read = false,
		.persistennt = false,
		.write_from_shader = true,
		.name = "cumsum buff"
	});

	
	sort_buff = new Buffer(BuffDesc{
		.byte_len = 2000 * 256 * 4 * 2, // could be smaller lolol
		.immutable_subdata = true,
		.mappable_write = true,
		.mappable_read = false,
		.persistennt = false,
		.write_from_shader = true,
		.name = "sort buff"
	});
	
	
	curve_painter = new CurvePainter();

	// BrushStroke stroke(blue_noise_tex);

	
	// curve_painter->fill_buff_for_long_brush(&stroke);
	
	// vid = new WVideo(R"(file:///D:\VJ vids\abstract\pexels-mart-production-7565898.mp4)");
	// vid = new WVideo(R"(file:///D:\VJ vids\nature\Pexels Videos 3593.mp4)");
	
	vid = new WVideo(R"(file:///D:\VJ vids\human\pexels-darina-belonogova-7551560.mp4)");
	// vid = new WVideo(R"(file:///C:\Users\user\Downloads\leviathan-debug_0gyKR9HBkN(2).mp4)");
	// vid = new WVideo(R"(file:///C:\Users\user\Downloads\leviathan-debug_0gyKR9HBkN(2).mp4)");
	// C:\Users\user\Documents\ShareX\Screenshots\2024-06\leviathan-debug_7o8seVwpb3.mp4
	// vid->update();
	
	
	Texture* main_fb_texture = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
		}
	).ptr;
		
	fb_main = new Framebuffer(FBDesc{ .textures = { main_fb_texture }, .name = "main" });

	
	// --- curve stuff
	{
		seq_amogus = WEngine->editor->timeline.AddSequence("Amogus");
		
		curve_posoomba = seq_amogus->AddCurve("asdg");
		curve_posoomba5 = seq_amogus->AddCurve("gdsfgd");	
		curve_posoomba4 = seq_amogus->AddCurve("gasdgsdd");	
		curve_posoomba7 = seq_amogus->AddCurve("fdddd");	
		seq_agoomba = WEngine->editor->timeline.AddSequence("Agoomba");
		seq_agoomba->AddCurve("dg023512");
		seq_agoomba->AddCurve("aaaa");
		seq_blorgus = WEngine->editor->timeline.AddSequence("Blorgus");
		seq_blorgus->AddCurve("tgdsgg");
		seq_blorba = WEngine->editor->timeline.AddSequence("Blorba");
		seq_blorba->AddCurve("339kfawsdthj");

		WEngine->editor->timeline.set_json_path("bruh_timeline.json");
	}

	pre_image_program = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_abstractexpressionism/pre_image.comp")
	});
	
	particles_program = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_abstractexpressionism/particles.comp")
	});
	
	dith_program = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_abstractexpressionism/dith.comp")
	});
	
	cumsum_program = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_abstractexpressionism/cumsum.comp")
	});
	
	

	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_abstractexpressionism/post.frag")
	});
	
	vpt_prog = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_abstractexpressionism/vpt.comp")
	});
	
	
	sort_prog = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_abstractexpressionism/sort.comp")
	});
	
	

}

void VarsAbstractExpressionism::loop() {
	VarsAbstractExpressionism& v = *this;
	Engine* e = WEngine;
	e->start_timer_query("start");
	
	if (e->initializing) {
		demo_audio = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/music/wangler.wav");
		demo_audio->start();
		demo_audio->set_volume(0.0);
		e->initializing = false;
	}

	// ---- gui stuff
	static glm::vec3 some_col = glm::vec3(1);
	static float some_val = 0.1; 
	{
		e->editor->add_gui_control("some_col", &some_col, {.is_colour = true});
		e->editor->add_gui_control("some_val", &some_val);
	}

	static bool reset = true;

	e->editor->add_gui_control("reset", &reset, {.callback = [&]() {
		reset = true;
	}});
	
	e->editor->add_gui_control("play/pause vid", &reset, {.callback = [&]() {
		if(!vid->paused) {
			vid->pause();
		} else {
			vid->play();
		}
	}});

	

	// vpt_buff->bind_as_SSBO(12);
	
	vid->update();
	// vid->play();
	// vid->pause();

	// ---- Image pass
	{
		e->start_timer_query("pre image pass");
		
		// e->default_fb->bind();
		
		// source_tex_edited
		pre_image_program->use();
		// pre_image_program->setUniform("source_tex", source_tex);
		pre_image_program->setUniform("source_tex", vid->tex);
		pre_image_program->setUniform("source_tex_edited", source_tex_edited);

		pre_image_program->dispatch_compute(
			source_tex_edited->resx,
			source_tex_edited->resy,
			1
		);
		// glDrawArrays(GL_TRIANGLES, 0, 6);

		glTextureBarrier();
	}
	
	source_tex_edited->generate_mipmap();


	// int part_cnt = 256*1;
	if(false) {
		
	// ---- Particles pass
	{
		particles_program->use();
		particles_program->setUniform("source_tex", source_tex_edited);
		particles_program->setUniform("vpt_tex", vpt_tex);
		particles_program->setUniform("particles_buff", vpt_buff);

		int part_cnt = 256*1;
		
		particles_program->setUniform("iter", 0);
		particles_program->dispatch_compute(part_cnt,1,1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		
		particles_program->setUniform("iter", 1);
		particles_program->dispatch_compute(part_cnt,1,1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		
		particles_program->setUniform("iter", 2);
		particles_program->dispatch_compute(part_cnt,1,1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
	}
	// ---- VPT pass
	{
		vpt_prog->use();
		vpt_prog->setUniform("out_tex", vpt_tex);
		vpt_prog->setUniform("particles_buff", vpt_buff);
		vpt_buff->bind_as_SSBO(13);
		vpt_prog->dispatch_compute(vpt_tex->resx,vpt_tex->resy,1);

		int iters = 10;
		
		vpt_prog->setUniform("tech", 1);
		vpt_prog->setUniform("iters", iters);

		for(int i = 0; i < iters; i++) {
			vpt_prog->setUniform("iter", i);
			vpt_prog->dispatch_compute(
				vpt_tex->resx,
				vpt_tex->resy,
				1
			);
			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}
		
		
		// glMemoryBarrier(GL_ALL_BARRIER_BITS);
	}
		
	}
	
	static float thresh = 0.25; 
	e->editor->add_gui_control("thresh", &thresh);

	if(true){
		int iters = 11;
		e->start_timer_query("cumsum pass`");
		// int pong = 0;
		cumsum_tex->clear();
		cumsum_program->use();
		cumsum_program->setUniform("input_tex", vid->tex);
		cumsum_program->setUniform("iters", iters);
		cumsum_program->setUniform("thresh", thresh);
		cumsum_program->setUniform("out_tex", cumsum_tex_out);
		
		for(int i = 0; i < iters; i++) {
			cumsum_program->setUniform("iter", i);
			if(i % 2 == 0) {
				cumsum_program->setUniform("cumsum_img", cumsum_tex);
				cumsum_program->setUniform("write_cumsum_img", cumsum_tex_b);
			} else {
				cumsum_program->setUniform("cumsum_img", cumsum_tex_b);
				cumsum_program->setUniform("write_cumsum_img", cumsum_tex);
			}
			cumsum_program->setUniform("cumsum_buff", cumsum_buff);
			cumsum_program->dispatch_compute(cumsum_tex->resx, cumsum_tex->resy, 1);
			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}
	}

	{
		e->start_timer_query("sort pass");
		
		sort_prog->use();
		sort_prog->setUniform("input_tex", vid->tex);
		sort_prog->setUniform("sort_tex", sort_tex);
		sort_prog->setUniform("sort_buff", sort_buff);
		sort_prog->setUniform("cumsum_buff", cumsum_buff);

		sort_buff->clear();
		sort_prog->setUniform("pass", 0);
		sort_prog->dispatch_compute(sort_tex->resx, sort_tex->resy, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		
		sort_prog->setUniform("pass", 1);
		sort_prog->dispatch_compute( sort_tex->resy, 1, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		
		sort_prog->setUniform("pass", 2);
		sort_prog->dispatch_compute( 256, sort_tex->resy, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		
		// sort_prog->setUniform("pass", 1);
		// glMemoryBarrier(GL_ALL_BARRIER_BITS);
		// 
		// 
		
	}

	if(true) {
		e->start_timer_query("dither pass`");
		int dith_rep = 15 + 0;
		// dith_error_tex->clear();
		// dith_output_tex->clear();
		dith_program->use();
		dith_program->setUniform("error_tex", dith_error_tex);
		dith_program->setUniform("output_tex", dith_output_tex);
		// dith_program->setUniform("input_tex", source_tex);
		// dith_program->setUniform("input_tex", vid->tex);
		// dith_program->setUniform("input_tex", vid->tex);
		dith_program->setUniform("input_tex", cumsum_tex_out);
		dith_program->setUniform("dith_rep", dith_rep);
		
		dith_program->setUniform("iter", 0);
		dith_program->dispatch_compute(e->RESX, e->RESY, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		dith_program->setUniform("iter", 1);
		dith_program->dispatch_compute(ceil(e->RESX/dith_rep), ceil(e->RESY/dith_rep), 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		dith_program->setUniform("iter", 2);
		dith_program->dispatch_compute(ceil(e->RESX/dith_rep), ceil(e->RESY/dith_rep), 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
	}
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();
		
		post_shader_prog->use();
		post_shader_prog->setUniform("tex_render", fb_main->textures[0]);
		post_shader_prog->setUniform("source_tex", source_tex_edited);
		post_shader_prog->setUniform("vpt_tex", vpt_tex);
		post_shader_prog->setUniform("output_tex", dith_output_tex);
		post_shader_prog->setUniform("dith_err_tex", dith_error_tex);
		post_shader_prog->setUniform("cumsum_img", cumsum_tex);
		post_shader_prog->setUniform("vid_tex", vid->tex);
		post_shader_prog->setUniform("cumsum_buff", cumsum_buff);
		post_shader_prog->setUniform("cumsum_out_tex", cumsum_tex_out);
		post_shader_prog->setUniform("sort_tex", sort_tex);
		
		
		// post_shader_prog->setUniform("tex_vpt", vpt_fb->get_front_textures_at(0));
		// post_shader_prog->setUniform("source_tex", source_tex);
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	// ---- curve pass
	if(false) {
		curve_painter->begin_stroke(
			blue_noise_tex,
			DrawParams{ }
		);
		
		for(float i = 0; i < 40; i++){
			curve_painter->push_stroke(
				glm::vec2(sin(i),cos(i*1.124)),
				glm::vec2(0.1),
				glm::vec2(0.1),
				1,
				glm::vec3(0)
			);
		}

		curve_painter->flush_strokes();
	}

	e->start_timer_query("end");
	
	reset = false;
}
