﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"

static Lane* seq_amogus;
static Curve curve_posoomba;
static Curve curve_posoomba5;	
static Curve curve_posoomba4;	
static Curve curve_posoomba7;	
static Lane* seq_agoomba;
static Curve curve_agoomba;
static Lane* seq_blorgus;
static Lane* seq_blorba;

static Texture* blue_noise_tex;

static ShaderProgram* prog_render;
static ShaderProgram* post_shader_prog;
static ShaderProgram* prog_thing;

static AudioFile* demo_audio;

static Framebuffer* fb_main;

static Texture* main_fb_texture;

// static Model* scene_model = nullptr;


void VarsCat::init() {
	VarsCat& v = *this;
	Engine* e = WEngine;
	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;

	Texture* main_fb_texture = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
		}
	).ptr;
		
	fb_main = new Framebuffer(FBDesc{ .textures = { main_fb_texture }, .name = "main" });

	// --- curve stuff
	{
		
		seq_amogus = WEngine->editor->timeline.AddSequence("Amogus");
		
		curve_posoomba = seq_amogus->AddCurve("asdg");
		curve_posoomba5 = seq_amogus->AddCurve("gdsfgd");	
		curve_posoomba4 = seq_amogus->AddCurve("gasdgsdd");	
		curve_posoomba7 = seq_amogus->AddCurve("fdddd");	
		seq_agoomba = WEngine->editor->timeline.AddSequence("Agoomba");
		seq_agoomba->AddCurve("dg023512");
		seq_blorgus = WEngine->editor->timeline.AddSequence("Blorgus");
		seq_blorgus->AddCurve("tgdsgg");
		seq_blorba = WEngine->editor->timeline.AddSequence("Blorba");
		seq_blorba->AddCurve("339kfawsdthj");

		WEngine->editor->timeline.set_json_path("bruh_timeline.json");
	}


	prog_render = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_cat/default.frag")
	});
	
	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_cat/post.frag")
	});

	// e->scene_model = e->alloc_scenes.push("assets\\models\\cube3.glb").ptr;
	
	prog_thing = new ShaderProgram({
		new Shader("thing.vert"),
		new Shader("thing.frag")
	}, true);
	
}

void VarsCat::loop() {
	VarsCat& v = *this;
	Engine* e = WEngine;
	e->start_timer_query("start");
	
	if (e->initializing) {
		demo_audio = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/music/wangler.wav");
		demo_audio->start();
		demo_audio->set_volume(0.1);
		e->initializing = false;
	}

	// ---- gui stuff
	static glm::vec3 some_col = glm::vec3(1);
	static float some_val = 0.1; 
	{
		e->editor->add_gui_control("some_col", &some_col, {.is_colour = true});
		e->editor->add_gui_control("some_val", &some_val);
	}
	

	
	// ---- raymarch pass
	{
		e->start_timer_query("raymarch");
		fb_main->bind();
		fb_main->clear();
		
		prog_render->use();
		prog_render->setUniform("u_t", (float)e->time);
		prog_render->setUniform("u_bpm", (float)115.0);
		prog_render->setUniform("some_col", some_col);
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}
	
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();
		
		post_shader_prog->use();
		post_shader_prog->setUniform("tex_render", fb_main->textures[0]);
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	e->start_timer_query("end");
}
