﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"
#include "engine/LUT.h"


void VarsCloudTracer::init() {
		VarsCloudTracer& v = *this;
		Engine* e = WEngine;
		v.lut = new LUT("Canon_EE_709_CL1_WDR.cube");
		WEngine->editor->expose_lut(v.lut);
	
		v.prog_clean_dbg = new ShaderProgram(
			{
				new Shader("engine/clean_dbg.comp")
			}
		);

		v.hdri_tex = WEngine->alloc_textures.push(std::string("src/assets/hdri/aristea_wreck_puresky_4k.hdr")).ptr;
		v.fb_GI = new Framebuffer(FBDesc{
				.textures = {WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
			}).ptr
				}, .name = "GI"
		});
	
		v.fb_Post = new Framebuffer(FBDesc{
			.textures = {WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW, .name = "Post"
			}).ptr}, .name = "Post"
		});

		v.prog_vox = new ShaderProgram(ShaderProgDescr{
			new Shader("cloud_tracer/vox.comp")
		});
		v.prog_bsm = new ShaderProgram(ShaderProgDescr{
			new Shader("cloud_tracer/bsm.comp")
		});
	
		v.prog_render_noise = new ShaderProgram(ShaderProgDescr{
			new Shader("cloud_tracer/render_noise.comp")
		});
	
		v.prog_render = new ShaderProgram(ShaderProgDescr{
			WEngine->vert_sh_fullscreen_quad,
			new Shader("cloud_tracer/render.frag")
		});
		v.prog_post = new ShaderProgram(ShaderProgDescr{
			WEngine->vert_sh_fullscreen_quad,
			new Shader("cloud_tracer/post.frag")
		});
		uint32_t vox_res = 128;
		TextureDesc vox_tex_desc = TextureDesc{
			.resx = vox_res, .resy = vox_res, .resz = vox_res,
			.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
		};
		v.tex_vox = WEngine->alloc_textures.push(vox_tex_desc).ptr;
		v.tex_voxb = WEngine->alloc_textures.push(vox_tex_desc).ptr;
		v.tex_voxc = WEngine->alloc_textures.push(vox_tex_desc).ptr;

		uint32_t bsm_res = 128;
		v.tex_cloud_shadow_map = WEngine->alloc_textures.push(TextureDesc{
			.resx = bsm_res, .resy = bsm_res, .resz = 1,
			.internal_format = InternalFormat::RGBA16F, .tex_access = TexAccess::RW, .name = "Cloud shadow map"
		}).ptr;

		// uint32_t noise_tex_res = 256;
		uint32_t noise_tex_res = 512;
		TextureDesc noise_tex_desc = TextureDesc{
			.resx = noise_tex_res, .resy = noise_tex_res, .resz = noise_tex_res,
			.internal_format = InternalFormat::RGBA16F, 
			.wrap_s = TextureWrapping::Repeat, .wrap_t = TextureWrapping::Repeat,
			.tex_access = TexAccess::RW
		};
		v.noise_tex  = WEngine->alloc_textures.push(vox_tex_desc).ptr;

		v.tex_blue_noise = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
	
}

void VarsCloudTracer::loop() {
	VarsCloudTracer& v = *this;
	Engine* e = WEngine;
	
	ZoneScopedN("LOOP");
	// Sleep(1000);


	if (e->initializing) {
		ZoneScopedN("re-init");
		wlog_info("-- RE-INITIALIZING -- ");




		e->initializing = false;
	}


	e->start_timer_query("clean dbg");
	{
		v.prog_clean_dbg->use();
		v.prog_clean_dbg->dispatch_compute(1, 1, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		v.dbg_tex->clear();
	}
	// ------------------------------- //
	// ---------- GAME LOGIC --------- //
	// ------------------------------- //
	{
		ZoneScopedN("GAME LOGIC");

		e->editor->gui_dbg_print("mouse ndc", e->io->mouse_ndc);
	}

	// ------------------------------- //
	// ----------- RENDER ----------- //
	// ------------------------------- //
	{
		ZoneScopedN("RENDER COMMANDS");
		e->start_timer_query("frame_start");

		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //

		glMemoryBarrier(GL_ALL_BARRIER_BITS);


		static glm::vec2 sun_dir = glm::vec2(-0.2,-0.2);
		static float sun_intensity = 1.0;
		static float cloud_denisty = 1.0;
		static glm::vec3 cloud_colour = glm::vec3(1);
		static glm::vec3 sun_colour = glm::vec3(1);
		static bool enable_bsm = true;
		static bool enable_indirect = true;
		static bool enable_henyey = true;
		
		e->editor->add_gui_control("enable henyey BSDF", &enable_henyey);
		e->editor->add_gui_control("enable indirect", &enable_indirect);
		e->editor->add_gui_control("enable bsm", &enable_bsm);
		e->editor->add_gui_control("sun dir", &sun_dir);
		e->editor->add_gui_control("sun intensity", &sun_intensity);
		e->editor->add_gui_control("cloud density", &cloud_denisty);
		e->editor->add_gui_control("taa jitt", &e->camera->taa_jitt_amt);
		e->editor->add_gui_control("cloud colour", &cloud_colour, {.is_colour = true});
		e->editor->add_gui_control("sun colour", &sun_colour, {.is_colour = true});


		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);

		// ------ Render noise ----- //
		{
			e->start_timer_query("render noise");
			ZoneScopedN("render noise");
			v.prog_render_noise->use();
			v.prog_render_noise->setUniform("noise_tex", v.noise_tex);
			
			glm::ivec3 vox_tex_res = glm::ivec3(
				v.noise_tex->resx, v.noise_tex->resy, v.noise_tex->resz
			);
			v.prog_render_noise->dispatch_compute(vox_tex_res.x,vox_tex_res.y,vox_tex_res.z);
			
			glTextureBarrier();
		}
		
		// ------ Do BSM ----- //
		{
			e->start_timer_query("BSM");
			ZoneScopedN("render BSM");
			if(enable_bsm) {
				v.prog_bsm->use();
				v.prog_bsm->setUniform("vox_tex", v.tex_vox);
				v.prog_bsm->setUniform("vox_texb", v.tex_voxb);
				v.prog_bsm->setUniform("vox_texc", v.tex_voxc);
				v.prog_bsm->setUniform("bsm_tex", v.tex_cloud_shadow_map);
				v.prog_bsm->setUniform("sun_dir", sun_dir);
				v.prog_bsm->setUniform("cloud_density", cloud_denisty);
				v.prog_bsm->setUniform("noise_tex", v.noise_tex);
				v.prog_bsm->setUniform("cloud_colour", cloud_colour);
				v.prog_bsm->setUniform("sun_colour", sun_colour * sun_intensity);
				v.prog_bsm->setUniform("sun_intensity", sun_intensity);
				v.prog_bsm->dispatch_compute(
					v.tex_cloud_shadow_map->resx,v.tex_cloud_shadow_map->resy,1
				);
				glTextureBarrier();
			}
		}
		// ------ Compute ----- //
		{
			e->start_timer_query("compute rad cache");
			ZoneScopedN("compute rad cache");
			if(!enable_bsm) {
				// for (auto shader : this->shader_manager->shader_programs) {
				v.prog_vox->use();
				v.prog_vox->setUniform("vox_tex", v.tex_vox);
				v.prog_vox->setUniform("vox_texb", v.tex_voxb);
				v.prog_vox->setUniform("vox_texc", v.tex_voxc);
				v.prog_vox->setUniform("hdri_tex", v.hdri_tex);
				v.prog_vox->setUniform("sun_dir", sun_dir);
				v.prog_vox->setUniform("sun_intensity", sun_intensity);
				v.prog_vox->setUniform("cloud_density", cloud_denisty);
				v.prog_vox->setUniform("cloud_colour", cloud_colour);
				v.prog_vox->setUniform("sun_colour", sun_colour * sun_intensity);
				v.prog_vox->setUniform("noise_tex", v.noise_tex);

				glm::ivec3 vox_tex_res = glm::ivec3(
					v.tex_vox->resx,
					v.tex_vox->resy,
					v.tex_vox->resz
				);

				v.prog_vox->dispatch_compute(vox_tex_res.x,vox_tex_res.y,vox_tex_res.z);
				e->editor->gui_dbg_print("vox disp sz", glm::vec3(vox_tex_res));
				// glMemoryBarrier(GL_TEXTURE_FETCH_BARRIER_BIT);
				glTextureBarrier();
			}
		}

		// ------ Render ----- //
		{

			// for (auto shader : this->shader_manager->shader_programs) {

			// if(enable_bsm) {
				e->start_timer_query("render");
				ZoneScopedN("Render");
				v.fb_GI->bind();
				v.fb_GI->clear();
				v.prog_render->use();
				v.prog_render->setUniform("vox_tex", v.tex_vox);
				v.prog_render->setUniform("blue_noise_tex", v.tex_blue_noise);
				v.prog_render->setUniform("hdri_tex", v.hdri_tex);
				v.prog_render->setUniform("sun_dir", sun_dir);
				v.prog_render->setUniform("sun_intensity", sun_intensity);
				v.prog_render->setUniform("cloud_density", cloud_denisty);
				v.prog_render->setUniform("cloud_colour", cloud_colour);
				v.prog_render->setUniform("sun_colour", sun_colour * sun_intensity);
				v.prog_render->setUniform("noise_tex", v.noise_tex);
				v.prog_render->setUniform("bsm_tex", v.tex_cloud_shadow_map);
				v.prog_render->setUniform("enable_bsm", enable_bsm);
				v.prog_render->setUniform("enable_indirect", enable_indirect);
				v.prog_render->setUniform("enable_henyey", enable_henyey);
			// }
			
			
			glDrawArrays(GL_TRIANGLES, 0, 6);

			// glMemoryBarrier(GL_ALL_BARRIER_BITS);
			// glTextureBarrier();
		}
		// ------ Post ----- //
		{
			e->start_timer_query("post");
			ZoneScopedN("Post");

			// for (auto shader : this->shader_manager->shader_programs) {

			v.fb_Post->bind();
			v.fb_Post->clear();

			v.prog_post->use();
			v.prog_post->setUniform("render_tex", v.fb_GI->get_front_textures_at(0));
			v.prog_post->setUniform("lut_tex", v.lut->tex);

			glDrawArrays(GL_TRIANGLES, 0, 6);

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
			glTextureBarrier();
		}
		v.fb_Post->blit(e->default_fb);

		// hdri_tex->generate_mipmap();

		// v.prev_frame_gi_tex->clear();
	}

	// ------ End ----- //
	{
		ZoneScopedN("END RENDER");
		e->start_timer_query("frame_end");
		// editor->depth_texture = v.fb_G->get_front_depth();
		// this->endFrame();
	}

	{
		ZoneScopedN("MORE END OF RENDER");
		// audio_mixer->update_player(camera->pos, camera->dir, v.player->get_vel() * 10.0f);

		// drawer_3d->reset();
		// drawer_2d->reset();
		// text_drawer->reset_end_of_frame();

	}

}

