﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"

static Lane* seq_amogus;
static Curve curve_posoomba;
static Curve curve_posoomba5;	
static Curve curve_posoomba4;	
static Curve curve_posoomba7;	
static Curve curve_posoomba2;	
static Lane* seq_agoomba;
static Curve curve_agoomba;
static Lane* seq_blorgus;
static Lane* seq_blorba;

static Texture* blue_noise_tex;

// static ShaderProgram* prog_render;
static ShaderProgram* post_shader_prog;
static ShaderProgram* prog_splat;

static ShaderProgram* scene2_post_shader_prog;
static ShaderProgram* scene2_prog_splat;

static ShaderProgram* scene3_post_shader_prog;
static ShaderProgram* scene3_prog_splat;

static AudioFile* demo_audio;

static Framebuffer* fb_main;
static Framebuffer* fb_splat;
static Texture* splat_tex;
static Texture* splat_tex_b;
static Texture* splat_tex_c;


static int SCENE = 2;



void VarsExegfx::init() {
	VarsExegfx& v = *this;
	Engine* e = WEngine;
	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;

	splat_tex = WEngine->alloc_textures.push(
			TextureDesc{
					.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
					.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW, .name = "splat tex"
			}
		).ptr;
	
	splat_tex_b = WEngine->alloc_textures.push(
			TextureDesc{
					.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
					.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW, .name = "splat tex b"
			}
	).ptr;
	
	splat_tex_c = WEngine->alloc_textures.push(
			TextureDesc{
					.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
					.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW, .name = "splat tex c"
			}
	).ptr;
	
	fb_splat = new Framebuffer(FBDesc{ .textures = {
		WEngine->alloc_textures.push(
    		TextureDesc{
    				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
    				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
    		}
    	).ptr
	}, ._pongable = true, .name = "main" });
	
	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_exegfx/post.frag")
	});
	
	prog_splat = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_exegfx/splat.comp")
	});
	
	
	scene2_post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_exegfx/2post.frag")
	});
	
	scene2_prog_splat = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_exegfx/2splat.comp")
	});
	
	scene3_post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_exegfx/3post.frag")
	});
	
	scene3_prog_splat = new ShaderProgram(ShaderProgDescr{
		new Shader("scene_exegfx/3splat.comp")
	});

}

void VarsExegfx::loop() {
	VarsExegfx& v = *this;
	Engine* e = WEngine;
	e->start_timer_query("start");
	
	if (e->initializing) {
		// demo_audio = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/music/wangler.wav");
		// demo_audio->start();
		// demo_audio->set_volume(0.1);
		// e->initializing = false;
	}

	// ---- gui stuff
	static bool reset = true;
	static int pt_frame = 0;

	
	e->editor->add_gui_control("fov", &e->camera->fov);
	e->editor->add_gui_control("reset", &reset, {.callback = [&]() {
		reset = true;
		pt_frame = 0;
	}});

	if(e->camera->camera_moved_this_frame || e->shader_manager->shader_just_reloaded) {
		reset = true;
		pt_frame = 0;
	}

	
	splat_tex_c->bind_to_image_unit(5);
	splat_tex->bind_to_image_unit(6);
	splat_tex_b->bind_to_image_unit(7);

	{
		e->start_timer_query("splat");
		splat_tex->clear();
		splat_tex_b->clear();
		splat_tex_c->clear();
		if(SCENE == 0) {
			prog_splat->use();
			prog_splat->setUniform("res", glm::vec2(e->RESX, e->RESY));
			prog_splat->dispatch_compute(35,10, 1, false);
		}else if(SCENE == 1) {
			scene2_prog_splat->use();
			scene2_prog_splat->setUniform("res", glm::vec2(e->RESX, e->RESY));
			scene2_prog_splat->dispatch_compute(15,10, 1, false);
		}else if(SCENE == 2) {
			scene3_prog_splat->use();
			scene3_prog_splat->setUniform("res", glm::vec2(e->RESX, e->RESY));
			scene3_prog_splat->dispatch_compute(15,10, 1, false);
		}
	}

	
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();

		
		if(SCENE == 0) {
			post_shader_prog->use();
			post_shader_prog->setUniform("res", glm::vec2(e->RESX, e->RESY));
		} else if(SCENE == 1) {
			scene2_post_shader_prog->use();
			scene2_post_shader_prog->setUniform("res", glm::vec2(e->RESX, e->RESY));
		} else if(SCENE == 2) {
			scene3_post_shader_prog->use();
			scene3_post_shader_prog->setUniform("res", glm::vec2(e->RESX, e->RESY));
		}
		glDrawArrays(GL_TRIANGLES, 0, 6);
			
	}

	e->start_timer_query("end");

	pt_frame++;
	reset = false;
}
