﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"
#include "engine/model/Model.h"
#include "engine/Light.h"
#include "engine/DrawerMDI.h"


class StartMenuB {
	struct UIElemText : UIElem {
		TextDescriptor text;
		static TextDescriptor& get_text(UIElem* elem) { return ((UIElemText*)elem)->text; };
	};

	public:
		UIElemText ui_elem_text_start;
		UIElemText ui_elem_text_pause;
		UIElemText ui_elem_text_settings;
		UIElemText ui_elem_text_quit;
		TextDrawer* text;
		bool toggled = false;
		float opacity = 0.0;
		glm::vec3 button_rgb = glm::vec3(0, 0, 0);
		glm::vec3 button_rgb_selected = glm::vec3(0.1, 0.1, 0.1);
		glm::vec3 button_rgb_enabled = glm::vec3(1.0, 1.0, 1.0);

		float init_text_sz;

		StartMenuB() = default;

		StartMenuB(TextDrawer* _text): text(_text) {
			TextDescriptor text_start = {.text = "Start"};
			this->init_text_sz = text_start.sz;

			this->ui_elem_text_start = UIElemText{
				{
					.rect = RectDescr{
						.col = glm::vec4(button_rgb, 0),
						.pos = glm::vec2(0, 0.75f),
						.sz = glm::vec2(0.4, 0.2)
					},
					.render = [&](UIElem* elem) {
						WEngine->drawer_2d->draw_rect(elem->rect);
						UIElemText::get_text(elem).pos = elem->rect.pos;
						text->draw(UIElemText::get_text(elem));
					},
					.on_mouse_not_over = [&](UIElem* elem) {
						UIElemText::get_text(elem).sz = glm::mix(UIElemText::get_text(elem).sz, init_text_sz,
							WEngine->delta_time * 20.0);
						elem->rect.col.value() = elem->rect.col.value().lerp_to(glm::vec4(button_rgb, this->opacity), WEngine->delta_time * 10.0);
						UIElemText::get_text(elem).col = glm::vec4(1, 1, 1, this->opacity);
					},
					.on_mouse_over = [&](UIElem* elem) {
						UIElemText::get_text(elem).sz = glm::mix(UIElemText::get_text(elem).sz, init_text_sz + 0.02f,
							WEngine->delta_time * 20.0);
						elem->rect.col.value() = glm::mix(elem->rect.col.value(), glm::vec4(button_rgb_selected, this->opacity),
							WEngine->delta_time * 10.0);
						UIElemText::get_text(elem).col = glm::vec4(1, 1, 1, this->opacity);
					},
					.on_mouse_down = [&](UIElem* elem) {
						UIElemText::get_text(elem).sz = glm::mix(UIElemText::get_text(elem).sz, init_text_sz + 0.05f,
							WEngine->delta_time * 20.0);
						elem->rect.col.value() = elem->rect.col.value().lerp_to(glm::vec4(0, 0, 1, this->opacity), WEngine->delta_time * 10.0);
						UIElemText::get_text(elem).col = UIElemText::get_text(elem).col.lerp_to(
							glm::vec4(button_rgb_enabled, this->opacity),
							WEngine->delta_time * 10.0);
					},
				},
				{.text = "Start"},
			};

			ui_elem_text_pause = ui_elem_text_start;
			ui_elem_text_settings = ui_elem_text_start;
			ui_elem_text_quit = ui_elem_text_start;
			float offs = ui_elem_text_start.rect.pos.y;
			const float pad = 0.5f;
			ui_elem_text_pause.text.text = "Pause";
			ui_elem_text_pause.rect.pos.y = offs -= pad;
			ui_elem_text_settings.text.text = "Settings";
			ui_elem_text_settings.rect.pos.y = offs -= pad;
			ui_elem_text_quit.text.text = "Quit";
			ui_elem_text_quit.rect.pos.y = offs -= pad;

			ui_elem_text_quit.on_mouse_unclicked = [&](UIElem* elem) {
				WEngine->editor->finished = true;
			};
		}

		void run() {
			this->opacity = glm::mix(this->opacity, float(this->toggled), WEngine->delta_time * 20.0f);
			WEngine->io->do_ui_elem(&ui_elem_text_start, this->toggled);
			// text->flush();
			// if(this->toggled) {
			WEngine->io->do_ui_elem(&ui_elem_text_pause, this->toggled);
			// }
			// text->flush();
			WEngine->io->do_ui_elem(&ui_elem_text_settings, this->toggled);
			// text->flush();
			WEngine->io->do_ui_elem(&ui_elem_text_quit, this->toggled);
			WEngine->drawer_2d->flush_quads();
			text->flush();
		}

		void toggle() {
			this->toggled = !this->toggled;
		}
};



void VarsGGJ::init() {
	VarsGGJ& v = *this;
	Engine* e = WEngine;

	v.font = new Font(std::string("Vintage_Stylist.otf"));
	e->text_drawer = new TextDrawer(v.font);
	e->scene_model = e->alloc_scenes.push("assets\\models\\cube3.glb").ptr;
	wlog_warn("-- SCENE LOADING FINISHED --");

	// for (int i = 0; i < 10; i++) {
	// 	// shadow_map_textures.push_back(new Framebuffer(nullptr, false, false, n))
	// 	shadow_map_textures.push_back(
	// 		new Framebuffer({
	// 			WArray<Texture*>({}), false, false,
	// 			WEngine->alloc_textures.push(TextureDesc{1024, 1024, 1, InternalFormat::DEPTH_COMPONENT24}).ptr
	// 		})
	// 	);
	// }

	wlog_warn("-- SHADOW MAP TEXTURES ALLOC FINISHED --");

	// ------ AUDIO ------ //
	// ------ AUDIO ------ //
	// ------ AUDIO ------ //
	// ------ AUDIO ------ //
	// ------ AUDIO ------ //
	// ------ AUDIO ------ //

	e->audio_mixer->set_master_volume(0.5f);
	e->audio_mixer->default_sound_group.set_volume(0.25f);
	{
		for (int i = 0; i < 5; i++) {
			std::string name = "src/assets/sounds/sfx/player_step_";
			name = name + std::to_string(i);
			name = name + ".wav";
			v.audio_footsteps[i] = new AudioFile(&e->audio_mixer->engine, name, {.sound_group = &e->audio_mixer->sfx_sound_group});
		}


		v.audio_bababooey = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/ggj_bababooey.mp3",{.sound_group = &e->audio_mixer->sfx_sound_group, .spatialization_enabled = true});
		v.audio_vineboom = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/ggj_vine_boom.wav",{.sound_group = &e->audio_mixer->sfx_sound_group, .spatialization_enabled = true});
		v.audio_huhcat = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/ggj_huh_cat.wav",{.sound_group = &e->audio_mixer->sfx_sound_group, .spatialization_enabled = true});
		v.audio_metalpipe = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/ggj_metal_pipe.wav",{.sound_group = &e->audio_mixer->sfx_sound_group, .spatialization_enabled = true});
		v.audio_metalpipe->set_volume(0.3f);
		
		v.audio_ost = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/1499-ggj.wav");
		v.audio_ost->set_looping_enabled(true);
		v.audio_ost->start();
	}

	wlog_warn("-- AUDIO INIT FINISHED --");

	// ------- SHADERS INIT ------- //
	// ------- SHADERS INIT ------- //
	// ------- SHADERS INIT ------- //
	// ------- SHADERS INIT ------- //
	// ------- SHADERS INIT ------- //

	// ShaderProgram* prog_gi_frag = new ShaderProgram(
	// 	{
	// 		 vert_sh_fullscreen_quad,
	// 	new Shader((char*)"gi.frag", ShaderType::Fragment)
	// 	});

	v.prog_clean_dbg = new ShaderProgram(
		{
			new Shader("engine/clean_dbg.comp")
		}
	);
	v.prog_gi_comp = new ShaderProgram({new Shader("gi.comp")});
	v.prog_di = new ShaderProgram({
			e->vert_sh_fullscreen_quad,
			new Shader("di.frag")
		}
	);
	v.prog_add_gi_and_di = new ShaderProgram({
			e->vert_sh_fullscreen_quad,
			new Shader("add_gi_and_di.frag")
		}
	);
	v.prog_gi_reproj = new ShaderProgram({
		e->vert_sh_fullscreen_quad,
		new Shader("gi_proj.frag")
	});

	v.prog_thing = new ShaderProgram({
		new Shader("thing.vert"),
		new Shader("thing.frag")
	}, true);

	// v.prog_trippy = new ShaderProgram({
	// 	new Shader("trippy.vert"),
	// 	new Shader("thing.frag")
	// }, true);

	// ShaderProgram* prog_march = new ShaderProgram({vert_sh_fullscreen_quad,
	// 	new Shader((char*)"raymarch.frag", ShaderType::Fragment)
	// });
	v.prog_taa = new ShaderProgram({
		e->vert_sh_fullscreen_quad,
		new Shader("taa.frag")
	});
	v.prog_post = new ShaderProgram({
		e->vert_sh_fullscreen_quad,
		new Shader("scene_ggj_2024/post.frag")
	});


	wlog_warn("-- SHADERS INIT FINISHED --");


	v.render_res_x = e->RESX;
	v.render_res_y = e->RESY;

	// ------- FRAMEBUFFERS ------- //
	// ------- FRAMEBUFFERS ------- //
	// ------- FRAMEBUFFERS ------- //
	// ------- FRAMEBUFFERS ------- //
	// ------- FRAMEBUFFERS ------- //
	// ------- FRAMEBUFFERS ------- //

	// Texture* back_depth = new Texture(render_res_x, render_res_y, 1u, depth_format);

	// WEngine->alloc_textures.push(TextureDesc
	{
		v.fb_G = new Framebuffer({
			WArray({
				// new Texture({render_res_x, render_res_y, 1u, InternalFormat::RGBA16F}),
				WEngine->alloc_textures.push(TextureDesc{v.render_res_x, v.render_res_y, 1u, InternalFormat::RGBA8}).ptr,
				WEngine->alloc_textures.push(TextureDesc{v.render_res_x, v.render_res_y, 1u, InternalFormat::RG16UI}).ptr,
				WEngine->alloc_textures.push(TextureDesc{v.render_res_x, v.render_res_y, 1u, InternalFormat::RG16F}).ptr,
			}),
			true, true,
			WEngine->alloc_textures.push(TextureDesc{v.render_res_x, v.render_res_y, 1u, v.depth_format}).ptr,
			false,
			"G Buff"
		});

		// InternalFormat gi_internal_format = InternalFormat::R11F_G11F_B10F;
		// InternalFormat gi_internal_format = InternalFormat::RGBA16F;
		v.fb_GI = new Framebuffer({
			WArray({
				WEngine->alloc_textures.push(TextureDesc{
					.resx = v.render_res_x,
					.resy = v.render_res_y,
					.resz = 1u,
					.internal_format = v.gi_internal_format,
					.tex_access = TexAccess::RW
				}).ptr,
			}),
			false, false,
			nullptr, false,
			"FB GI"
		});

		v.fb_DI = new Framebuffer({
			WArray({
				WEngine->alloc_textures.push(TextureDesc{v.render_res_x, v.render_res_y, 1u, v.gi_internal_format}).ptr,
			}),
			true, true,
			nullptr, false,
			"DI Buff"
		});

		v.fb_GI_plus_DI = new Framebuffer({
			WArray({
				WEngine->alloc_textures.push(TextureDesc{v.render_res_x, v.render_res_y, 1u, v.gi_internal_format}).ptr,
			}),
			false, false,
			nullptr, false,
			"GI+DI Buff"
		});


		v.fb_GI_reproj = new Framebuffer({
			WArray({
				WEngine->alloc_textures.push(TextureDesc{
					.resx = v.render_res_x,
					.resy = v.render_res_y,
					.resz = 1u,
					.internal_format = v.gi_internal_format,
					.tex_access = TexAccess::RW
				}).ptr,
			}),
			false, false, nullptr, false,
			"GI Reproj Buff"
		});

		v.prev_frame_gi_tex = WEngine->alloc_textures.push(TextureDesc{
			.resx = v.render_res_x,
			.resy = v.render_res_y,
			.resz = 1u,
			.internal_format = v.gi_internal_format,
			.tex_access = TexAccess::RW
		}).ptr;
		v.prev_frame_gi_tex->set_name("Prev Frame GI Tex");
		// v.prev_frame_gi_tex->name = "Prev Frame GI Tex";

		v.fb_taa = new Framebuffer({
			WArray({
				// new Texture({this->RESX, this->RESY, 1u, InternalFormat::RGBA32F})
				WEngine->alloc_textures.push(TextureDesc{e->RESX, e->RESY, 1u, v.gi_internal_format}).ptr
			}),
			true, true, nullptr, false, "TAA Buff"
		});

	}

	v.hdri_tex = WEngine->alloc_textures.push(std::string("src/assets/hdri/aristea_wreck_puresky_4k.hdr")).ptr;
	v.eye_open_tex = WEngine->alloc_textures.push(std::string("src/assets/images/eye_open.png")).ptr;
	v.eye_closed_tex = WEngine->alloc_textures.push(std::string("src/assets/images/eye_closed.png")).ptr;

	// PassBloom* pass_bloom = new PassBloom(v.fb_taa);


	// ------- UI INIT -------- //
	// ------- UI INIT -------- //
	// ------- UI INIT -------- //
	// ------- UI INIT -------- //
	// ------- UI INIT -------- //
	v.start_menu = new StartMenuB(e->text_drawer);


	// std::cout << "-- UI INIT FINISHED -- \n";

	// ------- SCENE & MODELS -------- //
	// ------- SCENE & MODELS -------- //
	// ------- SCENE & MODELS -------- //
	// ------- SCENE & MODELS -------- //
	// ------- SCENE & MODELS -------- //


	// allocator.push(GameObject(nullptr, PhysicsCreationInfo().set_physics_enabled(false)));
	// alloc_game_object.push(nullptr, PhysicsCreationInfo().set_physics_enabled(false)).ptr;


	// physics_creation_info.collision_shape = new JPH::SphereShape(0.1f);

	// v.trippy_obj = alloc_game_object.
	// 	push(v.prog_trippy, PhysicsCreationInfo().set_physics_enabled(false)
	// );

	// int vert_cnt = 100;
	// editor->add_gui_control("vert cnt", &vert_cnt, {.range_min = 0, .range_max = 10000});
	// v.trippy_obj.get()->render_function = [&, this](float a, float b, float c, float d, float e, float f) {
	// 	// glEnabl
	// 	vert_feedback_buff_idx = frame % 2;
	// 	Buffer* buff_front = vert_feedback_buff_idx == 0 ? vert_feedback_buff : vert_feedback_buff_other;
	// 	Buffer* buff_back = vert_feedback_buff_idx == 1 ? vert_feedback_buff : vert_feedback_buff_other;
	//
	// 	WEntity* _trippy_obj = v.trippy_obj.get();
	// 	// auto shader_prog = _trippy_obj->shader_prog;
	//
	// 	ShaderProgram* shader_p = _trippy_obj->shader_prog;
	// 	if(f > 0.0) {
	// 		if(shader_p->shadow_map_variation) {
	// 			shader_p = shader_p->shadow_map_variation;
	// 		}
	// 	}
	//
	// 	shader_p->use();
	//
	// 	shader_p->setUniform("buff_front", buff_front);
	// 	shader_p->setUniform("buff_back", buff_back);
	// 	shader_p->setUniform("mat", _trippy_obj->matrix);
	// 	shader_p->setUniform("prev_mat", _trippy_obj->prev_matrix);
	// 	shader_p->setUniform("rot_mat", _trippy_obj->rot_matrix);
	// 	// int elem_cnt = indices_buff->elem_cnt;
	// 	// elem_cnt = 16;
	// 	glDrawArrays(GL_TRIANGLES, 0, vert_cnt);
	// 	// glDrawArrays(GL_TRIANGLES, 0, 0);
	// };


	int vert_feedback_buff_byte_len = 2000;
	int vert_feedback_buff_idx = 0;

	// Buffer* vert_feedback_buff = new Buffer(new char8_t[vert_feedback_buff_byte_len], vert_feedback_buff_byte_len);
	// Buffer* vert_feedback_buff_other = new Buffer(new char8_t[vert_feedback_buff_byte_len], vert_feedback_buff_byte_len);


	// WEngine->editor->gui_dbg_notif_msg(str.c_str());
	// audio_obj_fall->start(-1, 1.0f - 1.0f/(1.0f + len*2.5f), me->get_pos());


	// this->lights.push_back(light_sun);

	// Light* pt_light = alloc_lights.push(LightDesc{2048, 2048, LightType::POINT});
	// pt_light->pos = player_start_pos;
	// pt_light->pos.y = 3.260f;
	// pt_light->near_plane = 0.001f;

}
void VarsGGJ::loop() {
	VarsGGJ& v = *this;
	Engine* e = WEngine;
	
	// auto parse_level

	ZoneScopedN("LOOP");
	// Sleep(1000);
	// this->startFrame();
	
	auto get_cell = [&](glm::ivec2 pos) -> Cell& {
		return v.level.get_cell(pos);
		// ent->set_pos(glm::vec3(pos.x, 0, pos.y)*2.0f);
	};
	
	auto move_obj_to_cell = [&](WEntity* ent, glm::ivec2 pos) {
		ent->props.ppprops.pos = pos;
		ent->set_pos(glm::vec3(pos.x, 0, pos.y)*2.0f);
	};
	
	auto is_cell_walkable = [&](Cell& c) {
		if(
			c.cell_type == CellType::WALL
		) {
			return false;
		}
		return true;
	};
	auto try_move_obj_to_cell = [&](WEntity* ent, glm::ivec2 dir) -> bool {
		glm::ivec2 pos = ent->props.ppprops.pos;
		glm::ivec2 desired_pos = pos + dir;
		
		Cell& c = v.level.get_cell(desired_pos);

		bool walkable = is_cell_walkable(c);

		if(walkable) {
			move_obj_to_cell(ent, desired_pos);
			v.audio_footsteps[0]->start();
		} else {
			// if()
			
		}
		return walkable;
	};
	
	

	if (e->initializing) {
		ZoneScopedN("re-init");
		std::cout << "-- RE-INITIALIZING -- \n";

		// ------ CREATE GAME OBJECTS ------ //
		// ------ CREATE GAME OBJECTS ------ //
		// ------ CREATE GAME OBJECTS ------ //
		// ------ CREATE GAME OBJECTS ------ //
		// ------ CREATE GAME OBJECTS ------ //
		// ------ CREATE GAME OBJECTS ------ //
		
		v.prefab_player = e->scene_model->parent_node->children_map["_Player"];

		// v.player_start_pos = this->scene_model->parent_node->children_map["_PlayerStart"]->position;
		v.player_start_pos = glm::vec3(v.level.player_start_pos.x,0,v.level.player_start_pos.y)*2.0f;
		v.player_pos = v.level.player_start_pos;
		
		v.player = e->alloc_game_object.push(v.prefab_player, nullptr).ptr;
		{
			// physics.body_interface->DestroyBodies()
			for (auto& child : e->scene_model->parent_node->children_map) {
				const std::string name = child.first;
				ModelNode* model = child.second;
				bool valid = true;
				bool is_prefab = name.at(0) == '_';
				if (is_prefab) {
					valid = false;
				}
				if (!is_prefab) {
					if (model->children.size() <= 0 && model->meshes.size() <= 0) {
						valid = false;
					}

				}
				if (valid) {
					auto game_object = e->alloc_game_object.push(
						model,
						v.prog_thing
						// PhysicsCreationInfo()
					).ptr;

					// game_object->render_function = [game_object](
					// 	float a, float b, float c, float d, float e, float f) {
					// 		ShaderProgram* shader_p = game_object->shader_prog;
					// 		if (f > 0.0) {
					// 			if (game_object->shader_prog->shadow_map_variation) {
					// 				shader_p = game_object->shader_prog->shadow_map_variation;
					// 			}
					// 		}
					//
					// 		shader_p->use();
					// 		shader_p->setUniform("rot_mat", game_object->rot_matrix);
					//
					// 		for (auto mesh : game_object->model_node->meshes) {
					// 			auto indices_buff = mesh->buff_vert_indices;
					// 			auto verts_buff = mesh->bufF_verts;
					// 			auto texcoords_buff = mesh->buff_tex_coords;
					// 			auto normals_buff = mesh->buff_normals;
					//
					// 			shader_p->setUniform("indices_buff", indices_buff);
					// 			shader_p->setUniform("verts_buff", verts_buff);
					// 			shader_p->setUniform("texcoords_buff", texcoords_buff);
					// 			shader_p->setUniform("normals_buff", normals_buff);
					// 			if (mesh->textures.size() > 0) {
					// 				shader_p->setUniform("tex_albedo", mesh->textures[0]);
					// 				shader_p->setUniform("has_albedo_tex", true);
					// 			} else {
					// 				shader_p->setUniform("has_albedo_tex", false);
					// 			}
					// 			shader_p->setUniform("mat", game_object->matrix);
					// 			shader_p->setUniform("prev_mat", game_object->prev_matrix);
					// 			//
					// 			shader_p->setUniform("has_texcoords", texcoords_buff != nullptr);
					// 			//
					// 			// int elem_cnt = indices_buff->elem_cnt / 3 ;
					// 			int elem_cnt = indices_buff->elem_cnt;
					// 			// elem_cnt = 16;
					// 			glDrawArrays(GL_TRIANGLES, 0, elem_cnt);
					// 		}
					// 	};
				}
			}

			std::cout << "-- INITIALIZED OBJECTS -- \n";
			// v.light_sun = alloc_lights.push(LightDesc{32, 32, LightType::DIRECTIONAL}).ptr;
			// v.light_sun->dir = glm::normalize(glm::vec3(1));

			if(false) {
			for (aiLight ai_light : e->scene_model->lights) {
				LightType light_type;
				switch (ai_light.mType) {
					case aiLightSource_DIRECTIONAL:
						light_type = LightType::DIRECTIONAL;
						break;
					case aiLightSource_POINT:
						light_type = LightType::POINT;
						break;
					case aiLightSource_SPOT:
						break;
					case aiLightSource_AMBIENT:
						break;
					case aiLightSource_AREA:
						break;
					case _aiLightSource_Force32Bit:
						break;
					default: ;
				}
				e->alloc_lights.push(LightDesc{
					.type = light_type,
					.intensity = ai_light.mAttenuationQuadratic,
					.colour = glm::vec3(ai_light.mColorDiffuse.r, ai_light.mColorDiffuse.g, ai_light.mColorDiffuse.b),
					.pos = glm::vec3(ai_light.mPosition.x, ai_light.mPosition.y, ai_light.mPosition.z),
					.target = glm::vec3(ai_light.mPosition.x, ai_light.mPosition.y, ai_light.mPosition.z),
					.dir = glm::vec3(ai_light.mDirection.x, ai_light.mDirection.y, ai_light.mDirection.z),
				}).ptr;
			}
				
			}
			
			std::cout << "-- INITIALIZED LIGHTS -- \n";
		}

		v.block_prefab = e->scene_model->parent_node->children_map["_Wall"];

		v.level.parse(v.level.curr_lev);

		// ------- parse and init map ------- //
		for(int h = 0; h < v.level.height; h++) {
			for(int w = 0; w < v.level.width; w++) {
				Cell& c = v.level.level[h][w];

				glm::ivec2 ipos = glm::ivec2(w,h);
				glm::vec3 fpos = glm::vec3(w, 0, h)*2.0f;
				
				WEntity* game_object = nullptr;
				if(c.cell_type == CellType::WALL) {
						game_object = e->alloc_game_object.push(
							v.block_prefab,
							// model,
							v.prog_thing
							// PhysicsCreationInfo()
						).ptr;
					game_object->obj_type = ObjType::Wall;
				} else if(c.cell_type == CellType::ENEMY) {
					
					ModelNode* prefab;

					switch (c.enemy_type) {
						case EnemyType::Bababooey:
							prefab = e->scene_model->parent_node->children_map["_EnemyBababooey"];
							break;
						case EnemyType::Pipefalling:
							prefab = e->scene_model->parent_node->children_map["_EnemyPipefalling"];
							break;
						case EnemyType::Vineboom:
							prefab = e->scene_model->parent_node->children_map["_EnemyVineboom"];
							break;
						case EnemyType::Huhcat:
							prefab = e->scene_model->parent_node->children_map["_EnemyHuhcat"];
							break;
					}
					
// enum class EnemyType: char {
// 	Bababooey = 'B',
// 	Pipefalling = 'P',
// 	Vineboom = 'V',
// 	Huhcat = 'H',
// };
					game_object = e->alloc_game_object.push(
						// v.prefab_enemy,
						// this->scene_model->parent_node->children_map["_Enemy"],
						prefab,
						// model,
						v.prog_thing
						// PhysicsCreationInfo()
					).ptr;
					game_object->obj_type = ObjType::Enemy;
					game_object->props.ppprops.enemy_type = c.enemy_type;
					game_object->props.ppprops.enemy_state = EnemyState::Patrolling;
					game_object->props.ppprops.patrol_dir = glm::ivec2(1,0);
					game_object->props.ppprops.spawn_pos = ipos;
					
				} else if (c.cell_type == CellType::FLOOR) {
					
				} else if (c.cell_type == CellType::DOOR) {
					ModelNode* door_prefab = e->scene_model->parent_node->children_map["_Door"];
					auto door_obj = e->alloc_game_object.push(
						// v.prefab_enemy,
						// this->scene_model->parent_node->children_map["_Enemy"],
						door_prefab,
						// model,
						v.prog_thing
						// PhysicsCreationInfo()
					).ptr;
					door_obj->set_pos(fpos);
					door_obj->props.ppprops.pos = ipos;
					
				}
				// if(c.cell_type ==)
				// if(game_object.)
				if(game_object != nullptr) {
					// game_object
					glm::vec3 pos = glm::vec3(w,0,h)*2.0f;
					game_object->set_pos(pos);
					game_object->props.ppprops.pos = ipos;
				}
			}
		}

		v.fb_taa->clear();
		v.fb_G->clear();
		v.fb_DI->clear();
		v.fb_GI->clear();
		v.fb_GI_reproj->clear();
		v.fb_GI_plus_DI->clear();

		v.prev_frame_gi_tex->clear();
		
		for(glm::ivec2 light_p : v.level.torches) {
				glm::vec3 lpos = glm::vec3(light_p.x, 0, light_p.y)*2.0f;
				e->alloc_lights.push(LightDesc{
					.shadowmap_x = 256,
					.shadowmap_y = 256,
					.type = LightType::POINT,
					.intensity = 5,
					.attenuation = 3.0f,
					.colour = glm::vec3(1,0.7,0.6) ,
					.pos = lpos,
					.target = lpos,
					.dir = glm::vec3(0,0,0),
				}).ptr;

				WEntity* torch = e->alloc_game_object.push(
					e->scene_model->parent_node->children_map["_Torch"],
					v.prog_thing
				).ptr;

				torch->set_pos(lpos);
			}




		// PhysicsCreationInfo physics_creation_info = PhysicsCreationInfo();
		// physics_creation_info = physics_creation_info.set_gravity_factor(0.8);






		v.player->props.ppprops.pos = v.level.player_start_pos;
		v.player->set_pos(glm::vec3(v.level.player_start_pos.x, 0, v.level.player_start_pos.y)*2.0f);
		
		e->camera->pos = v.player->get_pos();

		e->camera->mode = CameraMode::LOOKAT;





		
		e->editor->add_gui_control("master vol", &e->audio_mixer->master_vol, {
			.range_min = 0., .range_max = 1.0, .precision = 0.01,
			.callback = [&]() { e->audio_mixer->set_master_volume(e->audio_mixer->master_vol); }
		});

		e->editor->add_gui_control("sfx vol", &e->audio_mixer->sfx_sound_group.vol, {
			.range_min = 0., .range_max = 1.0, .precision = 0.01,
			.callback = [&]() { e->audio_mixer->sfx_sound_group.set_volume(e->audio_mixer->sfx_sound_group.vol); }
		});

		e->editor->add_gui_control("music vol", &e->audio_mixer->default_sound_group.vol, {
			.range_min = 0., .range_max = 1.0, .precision = 0.01,
			.callback = [&]() { e->audio_mixer->default_sound_group.set_volume(e->audio_mixer->default_sound_group.vol); }
		});


		e->editor->add_gui_control("compute_subdiv", &v.compute_subdiv);
		e->editor->add_gui_control("add_prev_di_and_gi", &v.add_prev_di_and_gi);

		// editor->add_gui_control("tex anchor", &v.tex_anchor, {
		// 	.callback = [&]() {
		// 		text_test.anchor = TextAnchor(v.tex_anchor);
		// 	}
		// });


		// editor->add_gui_control("camrot speed", &v.camrot_speed);

		// editor->add_gui_control("taa jitt", &camera->taa_jitt_amt);

		// editor->add_gui_control("light_sun", &v.light_sun->intensity);
		e->editor->add_gui_control("ambient_light", &e->ambient_light_intensity);

		WEngine->physics.physics_system->OptimizeBroadPhase();

		e->initializing = false;
	}
	
	bool player_visible = false;
	
	// ------------------------------- //
	// ---------- GAME LOGIC --------- //
	// ------------------------------- //
	{

		ZoneScopedN("GAME LOGIC");

		auto body_stats = e->physics.physics_system->GetBodyStats();
		e->editor->gui_dbg_print("Physics bodies cnt: ", body_stats.mNumBodies);
		// physics.temp_allocator
		// editor->gui_dbg_print("Physics bodies cnt: ", );

		glm::vec3 player_pos = v.player->get_pos();

		// this->drawer_3d->draw_line(
		// 	{player_pos}, {player_pos + glm::vec3(v.player_facing_dir.x, 0, v.player_facing_dir.y)});



		// v.light_sun->intensity = 0.0f;
		// v.light_sun->target = player_pos;

		e->camera->viewspace_offs.x = sin(e->time * v.camrot_speed) * 0.005;
		e->camera->viewspace_offs.x = sin(e->time * v.camrot_speed) * 0.05;

		e->camera->viewspace_offs.x =
			sin(e->time * v.camrot_speed) * 0.0;
		e->camera->viewspace_offs.y = cos(e->time * v.camrot_speed) * 0.05;
		e->camera->viewspace_offs = glm::vec2(0, 0);

		// int amog = 1;
		// editor->gui_dbg_print_msg(std::to_string(amog));
		// std::cout << "amog";
		// static text_buff
		#ifdef EDITOR
		// static char char_buff[512];
		// if (ImGui::InputTextMultiline("text", char_buff, 1000)) {
		// 	// text->text = std::string(char_buff);
		// 	text_test.text = char_buff;
		// }
		#endif
		// editor->gui_dbg_print("player pos", player_pos);
		// editor->gui_dbg_print("player vel", v.player->get_vel());
		// editor->gui_dbg_print("mouse ndc", this->io->mouse_ndc);



		v.potato = 5;
		// editor->gui_dbg_print("Potato: ", v.potato);
		
		e->editor->add_gui_control("Level", &v.level.curr_lev);

		glm::vec3 vel = v.player->get_vel();

		// editor->gui_dbg_print("Velocity: ", vel);

		{

					// static RayCastResClosest ray_cast_res_forward;
					//
					// auto filter = JPH::IgnoreSingleBodyFilter(v.player->jph_body_id);
					//
					// ray_cast_res_forward = physics.RayCastClosest(
					// 	camera->pos,
					// 	glm::vec3(camera->dir * 10.0f),
					// 	filter, {}, {}
					// );
					//
					//
					// glm::vec4 lcol = glm::vec4(1, 0, 0, 1);
					//
					// if (ray_cast_res_forward.hit) {
































			// }









		}

		v.t_since_player_jump += e->delta_time;

		static RayCastResClosest ray_cast_res_player_floor;
		auto filter = JPH::IgnoreSingleBodyFilter(v.player->jph_body_id);

		ray_cast_res_player_floor = e->physics.RayCastClosest(
			e->camera->pos,
			glm::vec3(0, -10, 0),
			filter, {}, {}
		);

		bool is_grounded = false;
		if (ray_cast_res_player_floor.hit && ray_cast_res_player_floor.dist < 1.2) {
			is_grounded = true;
		}
		if (is_grounded) {
			glm::vec3 player_vel = v.player->get_vel();
			player_vel.x *= 0.6 * e->delta_time;
			player_vel.z *= 0.6 * e->delta_time;
			v.player->set_vel(player_vel);
		}

		// if(io->get_key(Key::E).just_pressed) {
		// 	glm::vec3 vel = ico_object->get_vel();
		// 	vel += glm::vec3(0,1,0);
		// 	ico_object->set_vel(vel);
		// }

		
		auto restart = [&]() {
			v.dead = false;
			for(int i = 0; i <= e->alloc_game_object.alloc_cnt; i++) {
				if(e->alloc_game_object.slot_active[i]) {
					WEntity* entt = &e->alloc_game_object.arr[i];
					if(entt->obj_type == ObjType::Enemy) {
						PlayerProps* pp = &entt->props.ppprops;
						pp->enemy_state = EnemyState::Patrolling;
						pp->pos = pp->spawn_pos;
					}
				}
















			}
			v.player->props.ppprops.pos = v.level.player_start_pos;
			v.player->set_pos(glm::vec3(
				v.player->props.ppprops.pos.x, 0, v.player->props.ppprops.pos.y
			)*2.0f);
			// for()
		};

		auto die = [&]() {
			v.dead = true;
		};
		
		auto finish_level = [&]() {
			v.finished_level_screen = true;
			v.level.curr_lev += 1;
		};













		if(v.level.finished_game) {
			if(e->io->get_key(Key::Space).just_pressed) {
				e->editor->finished = true;
			}
			TextDescriptor text_tesc {
					.sz = 0.35f,
					.centered = true,
					.text = "Finished game.\n"
							"Space to quit."
			};
			e->text_drawer->draw(text_tesc);
		} else if(v.finished_level_screen) {
			if(e->io->get_key(Key::Space).just_pressed) {
				e->to_reset_whole_engine = true;
				v.finished_level_screen = false;
			}
			TextDescriptor text_tesc {
					.sz = 0.35f,
					.centered = true,
					.text = "Finished level.\n"
							"\n"
							"Space to continue."
			};
			e->text_drawer->draw(text_tesc);
		} else if(v.intro_screen) {
			
		} else if(v.dead) {
			if(e->io->get_key(Key::R).just_pressed) {
				restart();
			}
		} else {
			bool player_walked = false;
			bool player_skipped_turn = false;
			bool turn_happened = false;
		
			if (WEngine->io->get_key(Key::Space).just_pressed) {
				player_skipped_turn = true;
			}

			if (!e->camera->noclip && !e->io->imgui_wants_mouse) {
				v.player->set_rot(
					glm::quatLookAt(glm::normalize(glm::vec3(e->camera->dir.x, 0., e->camera->dir.z)), glm::vec3(0, 1, 0)));
			
				glm::vec2 keyInput = glm::vec2(0, 0);
				if (WEngine->io->get_key(Key::A).just_pressed) {
					keyInput[0] -= 1;
				}
				if (WEngine->io->get_key(Key::D).just_pressed) {
					keyInput[0] += 1;
				}
				if (WEngine->io->get_key(Key::W).just_pressed) {
					keyInput[1] += 1;
				}
				if (WEngine->io->get_key(Key::S).just_pressed) {
					keyInput[1] -= 1;
				}

				e->editor->gui_dbg_print("player dir", v.player_facing_dir);
			
				if (abs(keyInput.x) > 0.01) {
					float walk_speed = WEngine->delta_time * v.player_speed; // deltatime

					glm::vec3 new_player_dir_f = glm::vec3(v.player_facing_dir.x, 0, v.player_facing_dir.y);
					glm::mat3 rot_mat = get_rot_mat(1, keyInput.x*pi/2.0);
					new_player_dir_f = rot_mat * new_player_dir_f;
					new_player_dir_f *= 1.4;
					v.player_facing_dir = glm::ivec2(new_player_dir_f.x, new_player_dir_f.z);

					// camera->pitch_yaw_roll.y += 0.1*delta_time;
					e->camera->pitch_yaw_roll.y = glm::atan(new_player_dir_f.x, new_player_dir_f.z);

					glm::vec3 offs_vec = e->camera->dir * keyInput[1] + e->camera->right * keyInput[0];
					offs_vec = glm::normalize(offs_vec);

					// float offs_vec_xz_len = glm::length(glm::vec2(offs_vec.x, offs_vec.z));

					glm::vec3 walk_vec = walk_speed * offs_vec * glm::vec3(1, 0, 1);
					// player_pos += walk_vec;
					if (is_grounded) {
						const float dist_btw_footsteps = 0.95;
						v.dist_since_footstep += glm::length(walk_vec);
						if (v.dist_since_footstep > dist_btw_footsteps) {
							v.audio_footsteps[v.footstep_idx++]->start();
							v.footstep_idx = v.footstep_idx % 5;
							v.dist_since_footstep = 0.0;
						}
					}
					e->camera->viewspace_offs.y = sin(v.dist_since_footstep * pi) * 0.025;
				}
				if(e->io->get_key(Key::W).just_pressed) {
					try_move_obj_to_cell(v.player, v.player_facing_dir);
					player_walked = true;
				} else if (e->io->get_key(Key::S).just_pressed) {
					try_move_obj_to_cell(v.player, -v.player_facing_dir);
					player_walked = true;
				}

				// die if player walks onto enemy.
				for(int i = 0; i <= e->alloc_game_object.alloc_cnt; i++) {
					if(e->alloc_game_object.slot_active[i]) {
						WEntity* entt = &e->alloc_game_object.arr[i];
						if(entt->obj_type == ObjType::Enemy) {
							if(entt->props.ppprops.pos == v.player->props.ppprops.pos) {
								die();
							}
						}
					}
				}
				if(v.player->props.ppprops.pos == v.level.exit_door_pos) {
					finish_level();
				}
			}

			if(player_walked || player_skipped_turn) {
				turn_happened = true;
			}
		
			for(int i = 0; i < e->alloc_lights.alloc_cnt; i++) {
				if(e->alloc_lights.slot_active[i]) {
					Light& light = e->alloc_lights.arr[i];
					if(glm::length(light.pos - v.player->get_pos()) < 3.0f) {
						player_visible = true;
					}
				}

			}
		

			// ---- ENEMY LOGIC ---- //
			for(int i = 0; i <= e->alloc_game_object.alloc_cnt; i++) {
				if(e->alloc_game_object.slot_active[i]) {
					WEntity* entt = &e->alloc_game_object.arr[i];
					if(entt->obj_type == ObjType::Enemy) {
						EnemyState& enemy_state = entt->props.ppprops.enemy_state;
						glm::ivec2& enemy_pos = entt->props.ppprops.pos;
					
						glm::ivec2 prev_enemy_pos_i = enemy_pos;
						glm::vec3 prev_enemy_pos_f = entt->get_pos();
						glm::vec3 desired_enemy_pos_f = glm::vec3(
							enemy_pos.x,
							0,
							enemy_pos.y
						)*2.0f;
						// glm::ivec2 pos_before entt->props.ppprops.pos;

						if(turn_happened) {
							// start alert
							if(player_visible) {
								if(glm::length(v.player->get_pos() - entt->get_pos()) < 8.0f) {
									entt->props.ppprops.attention_tar = v.player->props.ppprops.pos;
									entt->props.ppprops.enemy_state = EnemyState::Alerted;
								}
							}
							if (entt->props.ppprops.enemy_state == EnemyState::Alerted) {
								if(glm::length(glm::vec2(entt->props.ppprops.pos - v.player->props.ppprops.pos)) <= 1.03f) {
									die();
								}
								// end alert
								if(entt->props.ppprops.pos == entt->props.ppprops.attention_tar) {
									entt->props.ppprops.enemy_state = EnemyState::BackToPatrol;
								} else {
									// do alert
									bool did_find_path = v.level.bfs.run(
										entt->props.ppprops.pos,
										entt->props.ppprops.attention_tar,
										is_cell_walkable,
										e->drawer_3d
									);
									if(did_find_path) {
										entt->props.ppprops.pos += v.level.bfs.found_walk_dir;
									}
								}
							}
							if(enemy_state == EnemyState::BackToPatrol) {
								if(entt->props.ppprops.spawn_pos == entt->props.ppprops.pos) {
									entt->props.ppprops.enemy_state = EnemyState::Patrolling;
								} else {
									bool did_find_path = v.level.bfs.run(
										entt->props.ppprops.pos,
										entt->props.ppprops.spawn_pos,
										is_cell_walkable,
										e->drawer_3d
									);
									if(did_find_path) {
										entt->props.ppprops.pos += v.level.bfs.found_walk_dir;
									}
								}
							}
						
							// patrol
							if(enemy_state == EnemyState::Patrolling) {
								// try find visible player
								glm::ivec2& enemy_patrol_dir = entt->props.ppprops.patrol_dir;
								glm::ivec2 enemy_desired_pos = enemy_pos + enemy_patrol_dir;
								bool next_cell_walkable = is_cell_walkable(get_cell(enemy_desired_pos));
								if(!next_cell_walkable) {
									enemy_patrol_dir = -enemy_patrol_dir;
								}
								enemy_pos += enemy_patrol_dir;
							
								desired_enemy_pos_f = glm::vec3(
									enemy_pos.x,
									0,
									enemy_pos.y
								)*2.0f;
							}
						}

						desired_enemy_pos_f = glm::vec3(
							enemy_pos.x,
							0,
							enemy_pos.y
						)*2.0f;

						if(enemy_pos != prev_enemy_pos_i){
							if(entt->props.ppprops.enemy_type == EnemyType::Bababooey) {
								v.audio_bababooey->start({.pos = desired_enemy_pos_f});
							} else if(entt->props.ppprops.enemy_type == EnemyType::Huhcat) {
								v.audio_huhcat->start({.pos = desired_enemy_pos_f});
							} else if(entt->props.ppprops.enemy_type == EnemyType::Vineboom) {
								v.audio_vineboom->start({.pos = desired_enemy_pos_f});
							} else if(entt->props.ppprops.enemy_type == EnemyType::Pipefalling) {
								v.audio_metalpipe->start({.pos = desired_enemy_pos_f});
							}
						}
					
					

						entt->set_rot(
							glm::quatLookAt(glm::normalize((player_pos - entt->get_pos()) * glm::vec3(1,0,1)), glm::vec3(0, 1, 0))
						);

						entt->set_pos(
							glm::mix(
								prev_enemy_pos_f,
								desired_enemy_pos_f,
								float(e->delta_time*10.0f)
							)
						);
					}
				}
			
			}
		
			if (!e->camera->noclip) {
				// camera->pos = v.player->get_pos() + glm::vec3(0.0f, 0.45f, 0.0f);
				e->camera->pos = glm::mix(
					e->camera->pos,
					v.player->get_pos() + glm::vec3(0.0f, 0.45f, 0.0f),
					(float)e->delta_time*10.0f
				);
				// camera->look_at = v.player->get_pos() + glm::vec3(v.player_facing_dir.x, 0.45f, v.player_facing_dir.y);
				e->camera->look_at = glm::mix(
					e->camera->look_at,
					v.player->get_pos() + glm::vec3(v.player_facing_dir.x, 0.45f, v.player_facing_dir.y),
					(float)e->delta_time*10.0f
				);
			}
		}

		// v.player->set_pos(player_pos);

	}

	e->editor->add_gui_control("kill everything", &e->to_reset_whole_engine);

	e->start_timer_query("update_matrices");
	{
		ZoneScopedN("UPDATE MATRICES");

		for (std::pair<const long, WEntity*> game_object : e->entitties) {
			// long trippy_obj_id = trippy_obj.get()->game_obj_id;
			// if(game_object.second->obj_type == ObjType::IcosaTest) {
			// if(game_object.first == trippy_obj_id) {
			// 	std::cout << "amogus";
			// }
			game_object.second->recalc_matrices();
		}
	}

	// ------------------------------- //
	// ----------- TICKING ----------- //
	// ------------------------------- //
	{
		ZoneScopedN("PHYSICS TICK");
		for (auto& game_object : e->entitties) {
			if (game_object.second->tick_function) {
				game_object.second->tick_function();
			}
		}
		e->physics.tick();
	}

	// ------------------------------- //
	// ----------- RENDER ----------- //
	// ------------------------------- //
	{
		ZoneScopedN("RENDER COMMANDS");
		e->start_timer_query("frame_start");

		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //

		glMemoryBarrier(GL_ALL_BARRIER_BITS);

		e->start_timer_query("sun");

		static std::vector<WEntity*> entities_without_mdi;
		entities_without_mdi.resize(0);

		int k = 0;
		for (auto id_and_game_object : e->entitties) {
			WEntity* game_object = id_and_game_object.second;
			if(game_object->shader_prog) {
				if(game_object->shader_prog != v.prog_thing) {
					entities_without_mdi.push_back(game_object);
				} else if(game_object->model_node) {
					// if(k > 0) { break;
					// }
					e->drawer_mdi->push(game_object,false);
					k++;
				}
			}
			// game_object->draw();
		}
		e->drawer_mdi->upload_verts();


		// ------ SUN ----- //

		if(true) {
			ZoneScopedN("SUN");
			// light_sun->calc_matrices();
			// float* buff_ptr = (float*)this->shared_ssbo->cpu_data;
			// for (int i = 0; i < 16; i++) {
			// 	(buff_ptr)[256 + i] = (&light_sun->view_mat[0][0])[i];
			// 	(buff_ptr)[256 + 16 + i] = (&light_sun->proj_mat[0][0])[i];
			// }
			// glNamedBufferSubData(this->shared_ssbo->pid, 0, (256 + 32) * 4, (void*)buff_ptr);


			// this->editor->gui_dbg_print("cam pitch", camera->pitch_yaw_roll.x);
			// this->editor->gui_dbg_print("cam yaw", camera->pitch_yaw_roll.y);
			// this->editor->gui_dbg_print("cam row", camera->pitch_yaw_roll.z);

			// glEnable(GL_CULL_FACE); // increases perf
			glDisable(GL_CULL_FACE);
			glCullFace(GL_BACK);
			glEnable(GL_DEPTH_TEST);


			// for (auto shader : this->shader_manager->shader_programs) {
			v.prog_thing->use();
			v.prog_thing->setUniform("is_shadow", true);
			v.prog_thing->setUniform("mdi_uniforms_buff", e->drawer_mdi->uniforms_buff);

			// shader->setUniform("is_shadow", true);
			// }

			// if(false)
			for (int i = 0; i <= e->alloc_lights.max_slot; i++) {
				if (e->alloc_lights.slot_active[i]) {
					v.prog_thing->setUniform("curr_rendering_light", i);
					Light& light = e->alloc_lights.arr[i];
					if(light.type == LightType::DIRECTIONAL) {
						light.fb->bind();
						light.fb->clear();
						v.prog_thing->use();

						// glProgramUniform1i(v.prog_thing->pid, glGetUniformLocation(v.prog_thing->pid, "is_shadow"), 1);
						// editor->gui_dbg_print_msg("mogus");
						// this->drawer_mdi->flush_draw(false);
						e->drawer_mdi->draw();
					} else if(light.type == LightType::POINT) {
						v.prog_thing->setUniform("curr_rendering_light", i);
						// light.fb->clear();
						// for(int face = 0; face < 6; face++) {
						// 		light.fb->bind_layer(face);
						// }

						for(int face = 0; face < 6; face++) {
								light.fb->bind_layer(face);
								light.fb->clear_layer(face);
								v.prog_thing->setUniform("spot_sublight_idx", face);
								e->drawer_mdi->draw();
						}


					}
					// for (auto id_and_game_object : entitties) {
					// 	WEntity* game_object = id_and_game_object.second;
					// 	if (game_object->shader_prog && game_object->shader_prog->shadow_map_variation) {
					// 		game_object->shader_prog->shadow_map_variation->use();
					// 		game_object->shader_prog->shadow_map_variation->setUniform("curr_rendering_light", i);
					// 	}
					// 	// game_object->draw(0, 0, 0, 0, 0, 1);
					// }
				}
			}

			v.prog_thing->setUniform("is_shadow", false);

			// for (auto shader : this->shader_manager->shader_programs) {
				// shader->setUniform("is_shadow", false);
			// }

			// for (int i = 0; i < 16; i++) {
			// 	(buff_ptr)[256 + i] = this->camera->V.vals[i];
			// 	(buff_ptr)[256 + 16 + i] = this->camera->P.vals[i];
			// }
			// // glNamedBufferSubData(this->shared_ssbo->pid, 4 * 256, 32 * 4, (void*)buff_ptr);
			// glNamedBufferSubData(this->shared_ssbo->pid, 0, (256 + 32) * 4, (void*)buff_ptr);

		}

		// hdri_tex->generate_mipmap();

		// v.prev_frame_gi_tex->clear();


		e->start_timer_query("clean dbg");
		{
			v.prog_clean_dbg->use();
			v.prog_clean_dbg->dispatch_compute(1, 1, 1);
			glMemoryBarrier(GL_ALL_BARRIER_BITS);
			v.dbg_tex->clear();
		}

		// ------ GBUFF ----- //
		e->start_timer_query("draw");
		e->default_fb->clear();
		if(true){
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);
			glEnable(GL_DEPTH_TEST);
			v.fb_G->bind();
			v.fb_G->clear(0, 0, 0, 0);


			
			v.prog_thing->use();
			v.prog_thing->setUniform("mdi_uniforms_buff", e->drawer_mdi->uniforms_buff);
			v.prog_thing->setUniform("is_shadow", false);


			// this->drawer_mdi->flush_draw();
			e->drawer_mdi->draw();

			// prog_march->use();
			// glDrawArrays(GL_TRIANGLES, 0, 6);

			glDisable(GL_DEPTH_TEST);
			glDisable(GL_CULL_FACE);
			// glDepthMask(GL_FALSE);
		}

		e->drawer_mdi->reset();

		// ------ DI ----- //
		e->start_timer_query("DI");
		{
			v.prog_di->use();

			// v.prog_di->setUniform("light_depth_tex", v.light_sun->fb->depth_texture);
			// v.prog_di->setUniform("light_view_mat", v.light_sun->view_mat);
			// v.prog_di->setUniform("light_proj_mat", v.light_sun->proj_mat);
			// v.prog_di->setUniform("light_dir", v.light_sun->dir);


			v.prog_di->setUniform("hdri_tex", v.hdri_tex);
			v.prog_di->setUniform("gi_tex", v.fb_GI->get_front_textures_at(0));
			v.prog_di->setUniform("prev_gi_tex", v.prev_frame_gi_tex);
			v.prog_di->setUniform("col_tex", v.fb_G->get_front_textures_at(0));
			v.prog_di->setUniform("norm_mat_tex", v.fb_G->get_front_textures_at(1));
			v.prog_di->setUniform("depth_tex", v.fb_G->get_front_depth());

			v.fb_DI->bind();
			v.fb_DI->clear();
			glDrawArrays(GL_TRIANGLES, 0, 6);

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
			glTextureBarrier();
		}

		e->editor->gui_dbg_print("lights cnt", e->alloc_lights.alloc_cnt);
		e->start_timer_query("Add GI and DI");
		if (v.add_prev_di_and_gi) {
			v.fb_GI_plus_DI->bind();
			v.fb_GI_plus_DI->clear();
			v.prog_add_gi_and_di->use();
			v.prog_add_gi_and_di->setUniform("prev_gi_tex", v.prev_frame_gi_tex);
			v.prog_add_gi_and_di->setUniform("prev_di_tex", v.fb_DI->get_back_textures_at(0));
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}


		// glMemoryBarrier(GL_ALL_BARRIER_BITS);
		// ------ GI ----- //
		e->start_timer_query("GI");
		{
			v.prog_gi_comp->use();

			// v.prog_gi_comp->setUniform("compute_subdiv", v.compute_subdiv);
			// v.prog_gi_comp->setUniform("light_depth_tex", v.light_sun->fb->depth_texture);
			// v.prog_gi_comp->setUniform("light_view_mat", v.light_sun->view_mat);
			// v.prog_gi_comp->setUniform("light_proj_mat", v.light_sun->proj_mat);
			// v.prog_gi_comp->setUniform("light_dir", v.light_sun->dir);


			// v.prog_gi_comp->setUniform("add_prev_di_and_gi", v.add_prev_di_and_gi);
			if (v.add_prev_di_and_gi) {
				v.prog_gi_comp->setUniform("prev_gi_tex", v.fb_GI_plus_DI->get_front_textures_at(0));
			} else {
				v.prog_gi_comp->setUniform("di_tex", v.fb_DI->get_back_textures_at(0));
				v.prog_gi_comp->setUniform("prev_gi_tex", v.prev_frame_gi_tex);
			}


			// editor->gui_dbg_print_msg(std::string("") + std::to_string(v.hdri_tex->bindless_idx) );
			v.prog_gi_comp->setUniform("hdri_tex", v.hdri_tex);
			v.prog_gi_comp->setUniform("gi_tex", v.fb_GI->get_front_textures_at(0));
			v.prog_gi_comp->setUniform("col_tex", v.fb_G->get_front_textures_at(0));
			v.prog_gi_comp->setUniform("norm_mat_tex", v.fb_G->get_front_textures_at(1));
			v.prog_gi_comp->setUniform("depth_tex", v.fb_G->get_front_depth());

			if (!v.compute_subdiv) {
				for (int iter = 0; iter < 4; iter++) {
					v.prog_gi_comp->setUniform("iter", iter);
					v.prog_gi_comp->dispatch_compute(WEngine->RESX / 2, WEngine->RESY / 2, 1);
					glMemoryBarrier(GL_ALL_BARRIER_BITS);
				}
			} else {
				for (int iter = 0; iter < 4 * 4; iter++) {
					v.prog_gi_comp->setUniform("iter", iter);
					v.prog_gi_comp->dispatch_compute(WEngine->RESX / 4, WEngine->RESY / 4, 1);
					glMemoryBarrier(GL_ALL_BARRIER_BITS);
				}
			}

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
			glTextureBarrier();
		}

		// ------ GI reproj ----- //
		e->start_timer_query("GI reproj");
		{
			v.fb_GI_reproj->bind();
			v.fb_GI_reproj->clear();

			v.prog_gi_reproj->use();

			v.prog_gi_reproj->setUniform("hdri_tex", v.hdri_tex);
			v.prog_gi_reproj->setUniform("movec_tex", v.fb_G->get_front_textures_at(2));
			v.prog_gi_reproj->setUniform("sun_pos", v.fb_G->get_front_textures_at(1));
			v.prog_gi_reproj->setUniform("norm_mat_tex", v.fb_G->get_front_textures_at(1));
			v.prog_gi_reproj->setUniform("depth_tex", v.fb_G->get_front_depth());
			v.prog_gi_reproj->setUniform("depth_tex_prev", v.fb_G->get_back_depth());
			v.prog_gi_reproj->setUniform("norm_mat_tex_prev", v.fb_G->get_back_textures_at(1));
			v.prog_gi_reproj->setUniform("prev_gi_tex", v.prev_frame_gi_tex);
			// prog_gi_reproj->setUniform("gi_tex", fb_GI->textures[0]);
			v.prog_gi_reproj->setUniform("gi_tex", v.fb_GI->get_front_textures_at(0));
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}
		glMemoryBarrier(GL_ALL_BARRIER_BITS);

		// ------ TAA ----- //
		e->start_timer_query("taa");
		{
			v.fb_taa->bind();
			v.fb_taa->clear();

			v.prog_taa->use();
			v.prog_taa->setUniform("hdri_tex", v.hdri_tex);
			// prog_taa->setUniform("taa_hist_tex", (*fb_taa->get_back_textures())[0]);
			v.prog_taa->setUniform("movec_tex", v.fb_G->get_front_textures_at(2));
			v.prog_taa->setUniform("taa_hist_tex", v.fb_taa->get_back_textures_at(0));
			v.prog_taa->setUniform("col_tex", v.fb_G->get_front_textures_at(0));
			v.prog_taa->setUniform("dbg_gi_tex", v.fb_GI->get_front_textures_at(0));
			// inttasdg

			v.prog_taa->setUniform("gi_tex", v.fb_GI_reproj->get_front_textures_at(0));
			v.prog_taa->setUniform("prev_gi_tex", v.prev_frame_gi_tex);
			v.prog_taa->setUniform("di_tex", v.fb_DI->get_front_textures_at(0));
			v.prog_taa->setUniform("norm_tex", v.fb_G->get_front_textures_at(1));
			v.prog_taa->setUniform("depth_tex", v.fb_G->get_front_depth());
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}
		glMemoryBarrier(GL_ALL_BARRIER_BITS);

		// ------ Bloom ----- //
		// start_timer_query("bloom");
		// {
		// 	pass_bloom->run(fb_taa);
		// }
		
		// ------ Eye ----- //
		// ------ Text ----- //
		e->start_timer_query("eye");
		{
			e->text_drawer->flush();
			if(v.intro_screen) {
				
				TextDescriptor text_tesc;
				if(v.intro_segment == 0) {
					text_tesc = {
						.sz = 0.35f,
						.centered = true,
						.text = "Mankind is dead.\n"
								"Laughter is fuel.\n"
								"Hell is empty.\n"
								"\n"
								"Space to continue."
					};
				} else if (v.intro_segment == 1) {
					text_tesc = {
						.sz = 0.35f,
						.centered = true,
						.text = "WASD to move.\n"
								"Space to wait.\n"
								"R to restart.\n"
								"\n"
								"Space to continue."
					};
				}
				if(e->io->get_key(Key::Space).just_pressed) {
					v.intro_segment++;
					if(v.intro_segment == 2) {
						v.intro_screen = false;
					}
				}
				e->text_drawer->draw(text_tesc);
				e->text_drawer->flush();
			}
			// Texture* tex;
			
			if(!v.dead && !v.intro_screen && !v.level.finished_game && !v.finished_level_screen) {
				if(player_visible) {
					e->drawer_2d->draw_rect(
						RectDescr{
							.col = glm::vec4(1,1,1, 1),
							.pos = glm::vec2(0, 0.5),
							.sz = glm::vec2(0.2, 0.2*float(e->RESX)/float(e->RESY))*1.5f,
							.tex = v.eye_open_tex
						}
					);
				} else {
					e->drawer_2d->draw_rect(
						RectDescr{
							.col = glm::vec4(1,1,1, 0.35),
							.pos = glm::vec2(0, 0.5),
							.sz = glm::vec2(0.2, 0.2*float(e->RESX)/float(e->RESY))*1.5f,
							.tex = v.eye_closed_tex
						}
					);
				}
				e->drawer_2d->flush_quads();
			}

			if(v.dead) {
				TextDescriptor text_tesc{
					.sz = 0.35f,
					.centered = true,
					.text = "You are dead\n"
							"Press R to try again."
				};
				e->text_drawer->draw(text_tesc);
				e->text_drawer->flush();
			}
			// v.eye_open_tex.
		
		}


		// ------ Post ----- //
		e->start_timer_query("post");
		{

			e->default_fb->bind();
			e->default_fb->clear(0, 0, 0, 0);
			
			v.prog_post->use();
			v.prog_post->setUniform("taa_tex", v.fb_taa->get_front_textures_at(0));

			v.prog_post->setUniform("hdri_tex", v.hdri_tex);
			// prog_post->setUniform("gi_tex", (*fb_GI_reproj->get_front_textures())[0]);
			v.prog_post->setUniform("gi_tex", v.fb_GI_reproj->get_front_textures_at(0));
			v.prog_post->setUniform("prev_gi_tex", v.prev_frame_gi_tex);
			v.prog_post->setUniform("depth_tex", v.fb_G->get_front_depth());
			// prog_post->setUniform("pt_light", pt_light->fb->get_front_depth());

			glDrawArrays(GL_TRIANGLES, 0, 6);
		}
		glMemoryBarrier(GL_ALL_BARRIER_BITS);

	}

	// ------ Text ----- //
	{
		ZoneScopedN("UI");

		// auto get_text = [](UIElem* elem) -> TextDescriptor& {return ((UIElemText*)elem)->text;};
		// if (io->get_key(Key::Esc).just_pressed) {
		// 	v.start_menu->toggle();
		// }
		// v.start_menu->run();
		e->drawer_3d->flush();
		e->drawer_2d->flush_quads();
		{
			e->text_drawer->flush();
		}

	}
	// ------ End ----- //
	{
		ZoneScopedN("END RENDER");
		e->start_timer_query("frame_end");
		e->editor->depth_texture = v.fb_G->get_front_depth();
	}

	{
		ZoneScopedN("MORE END OF RENDER");
		e->audio_mixer->update_player(e->camera->pos, e->camera->dir, v.player->get_vel() * 10.0f);

		e->drawer_3d->reset();
		e->text_drawer->reset_end_of_frame();

	}


}

