﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"




static ShaderProgram* prog_arrow;
static ShaderProgram* prog_rotato;
static ShaderProgram* prog_fontxor;
static ShaderProgram* prog_asym;
static ShaderProgram* prog_glass;
static ShaderProgram* prog_bubble;
static ShaderProgram* name_writer;
static ShaderProgram* post_shader_prog;
static ShaderProgram* prog_thing;

static AudioFile* demo_audio;

static Framebuffer* fb_names;
static Framebuffer* fb_main;

static Texture* main_fb_texture;
static Texture* blue_noise_tex;
// static Model* scene_model = nullptr;


void VarsNordlicht24::init() {
	VarsNordlicht24& v = *this;
	Engine* e = WEngine;


	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;

	Texture* main_fb_texture = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
		}
	).ptr;
		
	fb_main = new Framebuffer(FBDesc{ .textures = { main_fb_texture }, .name = "main" });

	Texture* name_fb_texture = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = 18, .resy = 10, .resz = 1,
				.internal_format = InternalFormat::R32I, .tex_access = TexAccess::RW
		}
	).ptr;
		
	fb_names = new Framebuffer(FBDesc{ .textures = { name_fb_texture }, .name = "names" });


	prog_arrow = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/arrows.frag")
	});
	
	prog_rotato = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/rotate-thicc.frag")
	});
	
	prog_fontxor = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/fontxor.frag")
	});
	
	prog_asym = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/asym.frag")
	});

	name_writer = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/name_writer.frag")
	});
	
	
	prog_glass = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/glass.frag")
	});
	
	prog_bubble = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		
		new Shader("scene_nordlicht24/bubble.frag")
	});
	
	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_nordlicht24/post.frag")
	});

	// e->scene_model = e->alloc_scenes.push("assets\\models\\cube3.glb").ptr;
	
	prog_thing = new ShaderProgram({
		new Shader("thing.vert"),
		new Shader("thing.frag")
	}, true);
	
}

void VarsNordlicht24::loop() {
	VarsNordlicht24& v = *this;
	Engine* e = WEngine;
	double BPM = 175.0;
	e->start_timer_query("start");
	
	if (e->initializing) {
		fb_names->bind();
        fb_names->clear();
        name_writer->use();
		glDrawArrays(GL_TRIANGLES, 0, 6);

		demo_audio = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/music/dnb_peak-hour.mp3");
		demo_audio->start();
		demo_audio->set_volume(1.0);
		e->initializing = false;
	}

	// ---- gui stuff
	static glm::vec3 bg_col = glm::vec3(1., .996,.930);
	static glm::vec3 main_col = glm::vec3(.005,.007,.151);
	static glm::vec3 hl_col = glm::vec3(.936,.098,.420);
	static glm::vec3 silent_col = glm::vec3(.627,.884,.731);
	{
		e->editor->add_gui_control("bg_col", &bg_col, {.is_colour = true});
		e->editor->add_gui_control("main_col", &main_col, {.is_colour = true});
		e->editor->add_gui_control("highlight_col", &hl_col, {.is_colour = true});
		e->editor->add_gui_control("silent_col", &silent_col, {.is_colour = true});
	}
	

	
	// ---- raymarch pass
	{
		e->start_timer_query("main");
		fb_main->bind();
		fb_main->clear();

		double beat = (e->time / (60.0/BPM));
		if(beat < 4.0*16.0 )
		{
		prog_arrow->use();
		prog_arrow->setUniform("u_t", (float)e->time);
		prog_arrow->setUniform("u_bpm", (float)BPM);
		prog_arrow->setUniform("bg_col", bg_col);
		prog_arrow->setUniform("main_col", main_col);
		prog_arrow->setUniform("silent_col", silent_col);
		prog_arrow->setUniform("hl_col", hl_col);
		}
		else if(beat < 4.0*24.0)
		{
		prog_rotato->use();
		prog_rotato->setUniform("u_t", (float)e->time);
		prog_rotato->setUniform("u_bpm", (float)BPM);
		prog_rotato->setUniform("bg_col", bg_col);
		prog_rotato->setUniform("main_col", main_col);
		prog_rotato->setUniform("silent_col", silent_col);
		prog_rotato->setUniform("hl_col", hl_col);
		}
		else if(beat < 4.0*32.0)
		{
		prog_fontxor->use();
		prog_fontxor->setUniform("u_t", (float)e->time);
		prog_fontxor->setUniform("u_bpm", (float)BPM);
		prog_fontxor->setUniform("bg_col", bg_col);
		prog_fontxor->setUniform("main_col", main_col);
		prog_fontxor->setUniform("silent_col", silent_col);
		prog_fontxor->setUniform("hl_col", hl_col);
		}
		else if(beat < 4.0*40.0)
		{
		prog_asym->use();
		prog_asym->setUniform("u_t", (float)e->time);
		prog_asym->setUniform("u_bpm", (float)BPM);
		prog_asym->setUniform("bg_col", bg_col);
		prog_asym->setUniform("main_col", main_col);
		prog_asym->setUniform("silent_col", silent_col);
		prog_asym->setUniform("hl_col", hl_col);
		}
		else if(beat < 4.0*56.0)
		{
		prog_glass->use();
		prog_glass->setUniform("u_t", (float)e->time);
		prog_glass->setUniform("u_bpm", (float)BPM);
		prog_glass->setUniform("bg_col", bg_col);
		prog_glass->setUniform("main_col", main_col);
		prog_glass->setUniform("silent_col", silent_col);
		prog_glass->setUniform("hl_col", hl_col);
		prog_glass->setUniform("tex", fb_names->get_front_textures_at(0));
		}
		else
		{
		prog_bubble->use();
		prog_bubble->setUniform("u_t", (float)e->time);
		prog_bubble->setUniform("u_bpm", (float)BPM);
		prog_bubble->setUniform("bg_col", bg_col);
		prog_bubble->setUniform("main_col", main_col);
		prog_bubble->setUniform("silent_col", silent_col);
		prog_bubble->setUniform("hl_col", hl_col);
		}
		
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}
	
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();
		
		post_shader_prog->use();
		post_shader_prog->setUniform("tex_render", fb_main->textures[0]);
		
		post_shader_prog->setUniform("tex", fb_names->get_front_textures_at(0));
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}
	{
		if (WEngine->io->get_key(Key::Esc).just_pressed) {
			WEngine->editor->finished=true;
        }
	}
	e->start_timer_query("end");
}
