﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"

static Lane* seq_amogus;
static Curve curve_posoomba;
static Curve curve_posoomba5;	
static Curve curve_posoomba4;	
static Curve curve_posoomba7;	
static Curve curve_posoomba2;	
static Lane* seq_agoomba;
static Curve curve_agoomba;
static Lane* seq_blorgus;
static Lane* seq_blorba;

static Texture* blue_noise_tex;

// static ShaderProgram* prog_render;
static ShaderProgram* post_shader_prog;
static ShaderProgram* prog_opflow;
static ShaderProgram* prog_feedback;

static Framebuffer* fb_opflow;
static Framebuffer* fb_opflow_luma;
static Framebuffer* fb_feedback_effect;

static WVideo* vid;



void VarsOpflow::init() {
	VarsOpflow& v = *this;
	Engine* e = WEngine;
	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
	
	// vid = new WVideo(R"(file:///D:\VJ vids\abstract\pexels-mart-production-7565898.mp4)");
	// vid = new WVideo(R"(file:///D:\VJ vids\nature\Pexels Videos 3593.mp4)");
	
	vid = new WVideo(R"(file:///D:\VJ vids\human\pexels-darina-belonogova-7551560.mp4)");
	// vid = new WVideo(R"(file:///C:\Users\user\Downloads\leviathan-debug_0gyKR9HBkN(2).mp4)");
	// vid = new WVideo(R"(file:///C:\Users\user\Downloads\leviathan-debug_0gyKR9HBkN(2).mp4)");
	// C:\Users\user\Documents\ShareX\Screenshots\2024-06\leviathan-debug_7o8seVwpb3.mp4
	// vid->update();

	
	fb_opflow = new Framebuffer(FBDesc{ .textures = {
		WEngine->alloc_textures.push(
    		TextureDesc{
    				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
    				.internal_format = InternalFormat::RGBA16F, .tex_access = TexAccess::RW
    		}
    	).ptr
	}, ._pongable = true, .name = "main" });
	
	fb_opflow_luma = new Framebuffer(FBDesc{ .textures = {
		WEngine->alloc_textures.push(
    		TextureDesc{
    				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
    				.internal_format = InternalFormat::R16F, .tex_access = TexAccess::RW
    		}
    	).ptr
	}, ._pongable = true, .name = "main" });

	fb_feedback_effect = new Framebuffer(FBDesc{ .textures = {
		WEngine->alloc_textures.push(
    		TextureDesc{
    				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
    				.internal_format = InternalFormat::RGBA16F, .tex_access = TexAccess::RW
    		}
    	).ptr
	}, ._pongable = true, .name = "main" });
	
	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_opflow/post.frag")
	});
	
	prog_opflow = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_opflow/opflow.frag")
	});
	
	prog_feedback = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_opflow/feedback.frag")
	});
}

void VarsOpflow::loop() {
	VarsOpflow& v = *this;
	Engine* e = WEngine;
	e->start_timer_query("start");
	
	// ---- gui stuff
	static bool reset = true;
	static int pt_frame = 0;

	
	e->editor->add_gui_control("fov", &e->camera->fov);
	// e->editor->add_gui_control("reset", &reset, {.callback = [&]() {
	// 	reset = true;
	// 	pt_frame = 0;
	// }});
	
	e->editor->add_gui_control("play/pause vid", &reset, {.callback = [&]() {
		if(!vid->paused) {
			vid->pause();
		} else {
			vid->play();
		}
	}});
	
	vid->update();

	// if(e->camera->camera_moved_this_frame || e->shader_manager->shader_just_reloaded) {
	// 	reset = true;
	// 	pt_frame = 0;
	// }

	
	{
		e->start_timer_query("opflow");
		glDisable(GL_BLEND);
		fb_opflow->bind();
		fb_opflow->clear();
		
		prog_opflow->use();
		prog_opflow->setUniform("input_tex", vid->tex);
		prog_opflow->setUniform("prev_grad_tex", fb_opflow->get_back_textures_at(0));
		prog_opflow->setUniform("prev_lum_tex", fb_opflow_luma->get_back_textures_at(0));
		prog_opflow->setUniform("curr_lum_tex", fb_opflow_luma->get_front_textures_at(0));

		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	{
		e->start_timer_query("feedback");
		prog_feedback->use();
		fb_feedback_effect->bind();
		fb_feedback_effect->clear();
		prog_feedback->setUniform("prev_frame", fb_feedback_effect->get_back_textures_at(0));
		prog_feedback->setUniform("input_tex", vid->tex);
		prog_feedback->setUniform("grad_tex", fb_opflow->get_front_textures_at(0));

		glDrawArrays(GL_TRIANGLES, 0, 6);
	}
	
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();
		
		post_shader_prog->use();
		post_shader_prog->setUniform("res", glm::vec2(e->RESX, e->RESY));
		post_shader_prog->setUniform("movec_tex", fb_opflow->get_front_textures_at(0));
		post_shader_prog->setUniform("feedback_tex", fb_feedback_effect->get_front_textures_at(0));
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	e->start_timer_query("end");

	pt_frame++;
	reset = false;
}
