﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"

static Lane* seq_amogus;
static Curve curve_posoomba;
static Curve curve_posoomba5;	
static Curve curve_posoomba4;	
static Curve curve_posoomba7;	
static Curve curve_posoomba2;	
static Lane* seq_agoomba;
static Curve curve_agoomba;
static Lane* seq_blorgus;
static Lane* seq_blorba;

static Texture* blue_noise_tex;

// static ShaderProgram* prog_render;
static ShaderProgram* post_shader_prog;
static ShaderProgram* prog_pathtrace;

static AudioFile* demo_audio;

static Framebuffer* fb_main;
static Framebuffer* fb_opflow;



void VarsPathtrace::init() {
	VarsPathtrace& v = *this;
	Engine* e = WEngine;
	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
	
	fb_opflow = new Framebuffer(FBDesc{ .textures = {
		WEngine->alloc_textures.push(
    		TextureDesc{
    				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
    				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
    		}
    	).ptr
	}, ._pongable = true, .name = "main" });
	

	// --- curve stuff
	{
		seq_amogus = WEngine->editor->timeline.AddSequence("Amogus");
		
		curve_posoomba = seq_amogus->AddCurve("asdg");
		curve_posoomba5 = seq_amogus->AddCurve("gdsfgd");	
		curve_posoomba4 = seq_amogus->AddCurve("gasdgsdd");	
		curve_posoomba7 = seq_amogus->AddCurve("fdddd");	
		seq_agoomba = WEngine->editor->timeline.AddSequence("Agoomba");
		seq_agoomba->AddCurve("dg023512");
		seq_agoomba->AddCurve("aaaa");
		seq_blorgus = WEngine->editor->timeline.AddSequence("Blorgus");
		seq_blorgus->AddCurve("tgdsgg");
		seq_blorba = WEngine->editor->timeline.AddSequence("Blorba");
		seq_blorba->AddCurve("339kfawsdthj");

		WEngine->editor->timeline.set_json_path("bruh_timeline.json");
	}

	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_pathtrace/post.frag")
	});
	
	prog_pathtrace = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_pathtrace/pt.frag")
	});

}

void VarsPathtrace::loop() {
	VarsPathtrace& v = *this;
	Engine* e = WEngine;
	e->start_timer_query("start");
	
	if (e->initializing) {
		// demo_audio = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/music/wangler.wav");
		// demo_audio->start();
		// demo_audio->set_volume(0.1);
		// e->initializing = false;
	}

	// ---- gui stuff
	static bool reset = true;
	static int pt_frame = 0;

	
	e->editor->add_gui_control("fov", &e->camera->fov);
	e->editor->add_gui_control("reset", &reset, {.callback = [&]() {
		reset = true;
		pt_frame = 0;
	}});

	if(e->camera->camera_moved_this_frame || e->shader_manager->shader_just_reloaded) {
		reset = true;
		pt_frame = 0;
	}

	
	


	// ---- pathtrace pass
	{
		e->start_timer_query("pathtrace");
		fb_opflow->bind();
		fb_opflow->clear();
		prog_pathtrace->use();
		prog_pathtrace->setUniform("prev_frame_tex", fb_opflow->get_back_textures_at(0));
		prog_pathtrace->setUniform("pt_frame", pt_frame);
		// glDrawArrays(o)
		glDisable(GL_BLEND);
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	// vpt_buff->bind_as_SSBO(12);



	
	// ---- raymarch pass
	// {
	// 	e->start_timer_query("raymarch");
	// 	fb_main->bind();
	// 	fb_main->clear();
	// 	
	// 	prog_render->use();
	// 	prog_render->setUniform("u_t", (float)e->time);
	// 	prog_render->setUniform("u_bpm", (float)115.0);
	// 	prog_render->setUniform("some_col", some_col);
	// 	
	// 	glDrawArrays(GL_TRIANGLES, 0, 6);
	// }
	
	// prog_sim_particles->setUniform("reset", reset);
	// prog_sim_particles->setUniform("tex_vpt", vpt_fb->get_back_textures_at(0));
	// prog_sim_particles->setUniform("source_tex", source_tex);
	
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();
		
		post_shader_prog->use();
		post_shader_prog->setUniform("tex_render", fb_opflow->textures[0]);
		// post_shader_prog->setUniform("tex_vpt", vpt_fb->get_front_textures_at(0));
		// post_shader_prog->setUniform("source_tex", source_tex);
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	e->start_timer_query("end");

	pt_frame++;
	reset = false;
}
