﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"
// #include "engine/Engine.h"

void VarsRevision2024::init() {
		VarsRevision2024& v = *this;
		WEngine->camera->noclip = true;
	
		v.frame = 0;
		v.prog_clean_dbg = new ShaderProgram(
			{
				new Shader("engine/clean_dbg.comp")
			}
		);

		v.hdri_tex = WEngine->alloc_textures.push(std::string("src/assets/hdri/table_mountain_split.hdr")).ptr;
		v.font = new Font(std::string("Vintage_Stylist.otf"));

		v.fb_PT = new Framebuffer(FBDesc{
			.textures = {
				WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = v.gi_internal_format, .tex_access = TexAccess::RW
			}).ptr,
				}, ._pongable = true, .autopong = true, .name = "PT"
		});

		v.img_r = WEngine->alloc_textures.push(TextureDesc{
			.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
			.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW
		}).ptr;
		v.img_g = WEngine->alloc_textures.push(TextureDesc{
			.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
			.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW
		}).ptr;
		v.img_b = WEngine->alloc_textures.push(TextureDesc{
			.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
			.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW
		}).ptr;
	
		v.fb_Post = new Framebuffer(FBDesc{
			.textures = {WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
			}).ptr}, .name = "Post"
		});
	

		v.prog_render = new ShaderProgram(ShaderProgDescr{
			// WEngine->vert_sh_fullscreen_quad,
			new Shader("scene_revision_2024_4k/render.comp")
		});
		v.prog_post = new ShaderProgram(ShaderProgDescr{
			WEngine->vert_sh_fullscreen_quad,
			new Shader("scene_revision_2024_4k/post.frag")
		});

		v.buff = new Buffer(BuffDesc{
			.byte_len = 1000, .immutable_subdata = false,
			.read_from_shader = true, .write_from_shader = true
		});
		v.buff->bind_as_SSBO(6);
	
		// v.tex_vox = WEngine->alloc_textures.push(vox_tex_desc).ptr;
		// v.tex_voxb = WEngine->alloc_textures.push(vox_tex_desc).ptr;
		// v.tex_voxc = WEngine->alloc_textures.push(vox_tex_desc).ptr;

		v.tex_blue_noise = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
}

void VarsRevision2024::loop() {
	VarsRevision2024& v = *this;
	ZoneScopedN("LOOP");
	Engine* e = WEngine;
	// Sleep(1000);
	// this->startFrame();

	glDisable(GL_BLEND);

	e->camera->speed = 0.3;
	e->camera->_far = 200.0;

	static bool temporal_restir = true;
	static bool reset_cam_on_move = false;
	
	if (e->initializing) {
		ZoneScopedN("re-init");
		wlog_info("-- RE-INITIALIZING -- ");
		e->camera->pos = glm::vec3(0.062, 0.0, 1.181);

		e->initializing = false;
	}

	v.frame++;
	

	e->start_timer_query("clean dbg");
	{
		v.prog_clean_dbg->use();
		v.prog_clean_dbg->dispatch_compute(1, 1, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		v.dbg_tex->clear();
	}
	// ------------------------------- //
	// ---------- GAME LOGIC --------- //
	// ------------------------------- //
	{
		ZoneScopedN("GAME LOGIC");

		e->editor->gui_dbg_print("mouse ndc", e->io->mouse_ndc);
	}


	// ------------------------------- //
	// ----------- RENDER ----------- //
	// ------------------------------- //
	{
		ZoneScopedN("RENDER COMMANDS");
		e->start_timer_query("frame_start");

		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //

		glMemoryBarrier(GL_ALL_BARRIER_BITS);

		
		e->editor->add_gui_control("reset cam on move", &reset_cam_on_move);
		e->editor->add_gui_control("temporal_restir", &temporal_restir);
		e->editor->add_gui_control("taa jitt", &e->camera->taa_jitt_amt);

		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);

		// bool to_reset = false;
		//
		// if(
		// 	(this->camera->camera_moved_this_frame && reset_cam_on_move) || this->shader_manager->shader_just_reloaded) {
		// 	to_reset = true;
		// }
		
		// if(to_reset) {
		// 	v.fb_PT->clear_back();
		// 	v.frame = 1;
		// }

		// this->camera->fov = 
		e->editor->add_gui_control("can_fov", &e->camera->fov);

		glDisable(GL_DEPTH_TEST);

		// ------ Render ----- //
		{
			e->start_timer_query("render");
			ZoneScopedN("Render");

			// for (auto shader : this->shader_manager->shader_programs) {

			// v.fb_PT->bind();
			// v.fb_PT->clear();
			v.img_r->clear();
			v.img_g->clear();
			v.img_b->clear();
			
			v.prog_render->use();
			// v.prog_render->setUniform("vox_tex", v.tex_vox);
			v.prog_render->setUniform("blue_noise_tex", v.tex_blue_noise);
			v.prog_render->setUniform("hdri_tex", v.hdri_tex);
			v.prog_render->setUniform("img_r", v.img_r);
			v.prog_render->setUniform("img_g", v.img_g);
			v.prog_render->setUniform("img_b", v.img_b);

			v.prog_render->dispatch_compute(1200,1200,1);
			// glDrawArrays(GL_TRIANGLES, 0, 6);

			// glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}

		// ------ Post ----- //
		{
			e->start_timer_query("post");
			ZoneScopedN("Post");

			// for (auto shader : this->shader_manager->shader_programs) {

			v.fb_Post->bind();
			v.fb_Post->clear();

			v.prog_post->use();
			v.prog_post->setUniform("frame", v.frame);
			v.prog_post->setUniform("render_tex", v.fb_PT->get_front_textures_at(0));
			v.prog_post->setUniform("img_r", v.img_r);
			v.prog_post->setUniform("img_g", v.img_g);
			v.prog_post->setUniform("img_b", v.img_b);

			glDrawArrays(GL_TRIANGLES, 0, 6);

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
			glTextureBarrier();
		}
		v.fb_Post->blit(e->default_fb);

		// hdri_tex->generate_mipmap();

		// v.prev_frame_gi_tex->clear();
	}

	// ------ End ----- //
	{
		ZoneScopedN("END RENDER");
		e->start_timer_query("frame_end");
		// editor->depth_texture = v.fb_G->get_front_depth();
		// this->endFrame();
	}

	{
		ZoneScopedN("MORE END OF RENDER");
		// audio_mixer->update_player(camera->pos, camera->dir, v.player->get_vel() * 10.0f);

		e->drawer_3d->reset();
		e->drawer_2d->reset();
		// text_drawer->reset_end_of_frame();

	}

}

