﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"



void VarsRevision2024B::init() {
		VarsRevision2024B& v = *this;
	
		v.prog_clean_dbg = new ShaderProgram(
			{
				new Shader("engine/clean_dbg.comp")
			}
		);


		v.hdri_tex = WEngine->alloc_textures.push(std::string("src/assets/hdri/aristea_wreck_puresky_4k.hdr")).ptr;
		v.fb_GI = new Framebuffer(FBDesc{
				{WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::R11F_G11F_B10F, .tex_access = TexAccess::RW
			}).ptr
				}
		});
	// InternalFormat::R11F_G11F_B10F
		v.fb_Post = new Framebuffer(FBDesc{
			{WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::R11F_G11F_B10F, .tex_access = TexAccess::RW
			}).ptr}
		});

		v.prog_vox = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_revision_2024_b/vox.comp")
		});
	
		v.prog_render_noise = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_revision_2024_b/render_noise.comp")
		});
	
		v.prog_render = new ShaderProgram(ShaderProgDescr{
			WEngine->vert_sh_fullscreen_quad,
			new Shader("scene_revision_2024_b/render.frag")
		});
		v.prog_post = new ShaderProgram(ShaderProgDescr{
			WEngine->vert_sh_fullscreen_quad,
			new Shader("scene_revision_2024_b/post.frag")
		});
		// uint32_t vox_res = 128;
		// uint32_t vox_res = 256;
		uint32_t vox_res = 256;
		// uint32_t vox_res = 512;
		// uint32_t vox_res = 64;
		TextureDesc vox_tex_desc = TextureDesc{
			.resx = vox_res, .resy = vox_res, .resz = vox_res,
			// .internal_format = InternalFormat::RGBA16F, .tex_access = TexAccess::RW
			.internal_format = InternalFormat::R11F_G11F_B10F, .tex_access = TexAccess::RW
			// .internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
		};
		v.tex_vox = WEngine->alloc_textures.push(vox_tex_desc).ptr;
		v.tex_voxb = WEngine->alloc_textures.push(vox_tex_desc).ptr;
		v.tex_voxc = WEngine->alloc_textures.push(vox_tex_desc).ptr;

		// uint32_t noise_tex_res = 256;
		uint32_t noise_tex_res = 1024;
		TextureDesc noise_tex_desc = TextureDesc{
			.resx = noise_tex_res, .resy = noise_tex_res, .resz = noise_tex_res,
			.internal_format = InternalFormat::R11F_G11F_B10F, 
			.wrap_s = TextureWrapping::Repeat, .wrap_t = TextureWrapping::Repeat,
			.tex_access = TexAccess::RW
		};
		v.noise_tex  = WEngine->alloc_textures.push(vox_tex_desc).ptr;

		v.tex_blue_noise = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
	
}

void VarsRevision2024B::loop() {
	VarsRevision2024B& v = *this;
	Engine* e = WEngine;
	
	ZoneScopedN("LOOP");
	// Sleep(1000);


	if (e->initializing) {
		ZoneScopedN("re-init");
		wlog_info("-- RE-INITIALIZING -- ");


		// v.demo_audio = new AudioFile(&audio_mixer->engine, "D:\\Programming\\Ableton\\revision24 track.wav");
		v.demo_audio = new AudioFile(&e->audio_mixer->engine, "D:\\Programming\\Demoscene\\Levwrightian-exemusic\\Oidos\\easy_exe\\music.wav");
		
		v.demo_audio->set_looping_enabled(true);
		v.demo_audio->start();
		// v.demo_audio->set_volume()
	


		e->initializing = false;
	}


	e->start_timer_query("clean dbg");
	{
		v.prog_clean_dbg->use();
		v.prog_clean_dbg->dispatch_compute(1, 1, 1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		v.dbg_tex->clear();
	}
	// ------------------------------- //
	// ---------- GAME LOGIC --------- //
	// ------------------------------- //
	{
		ZoneScopedN("GAME LOGIC");

		e->editor->gui_dbg_print("mouse ndc", e->io->mouse_ndc);
	}

	e->editor->gui_dbg_print("audio pos", v.demo_audio->get_current_time());
	e->editor->gui_dbg_print("titty", (float)e->time);
	if(e->editor->timePaused) {
		v.demo_audio->set_volume(0);
	} else {
		v.demo_audio->set_volume(1);
	}
	if(abs(v.demo_audio->get_current_time() - e->time) > 0.5) {
		v.demo_audio->set_current_time(float(e->time));
	}
	// float sound_cursor;
	// ma_sound_get_cursor_in_seconds(&v.demo_audio->sounds[0], &sound_cursor);
	// editor->gui_dbg_print("audio pos", sound_cursor);
					
	// ------------------------------- //
	// ----------- RENDER ----------- //
	// ------------------------------- //
	{
		ZoneScopedN("RENDER COMMANDS");
		e->start_timer_query("frame_start");

		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //
		// ------------- RENDER ------------ //

		glMemoryBarrier(GL_ALL_BARRIER_BITS);


		static glm::vec2 sun_dir = glm::vec2(0,0);
		static float sun_intensity = 1.0;
		static float cloud_denisty = 1.0;
		static glm::vec3 cloud_colour = glm::vec3(1);
		static glm::vec3 sun_colour = glm::vec3(1);
		
		static float hyper_sc = 0.0;
		
		e->editor->add_gui_control("sun dir", &sun_dir);
		e->editor->add_gui_control("sun intensity", &sun_intensity);
		e->editor->add_gui_control("cloud density", &cloud_denisty);
		e->editor->add_gui_control("taa jitt", &e->camera->taa_jitt_amt);
		e->editor->add_gui_control("cloud colour", &cloud_colour, {.is_colour = true});
		e->editor->add_gui_control("sun colour", &sun_colour, {.is_colour = true});
		e->editor->add_gui_control("fov", &e->camera->fov);
		e->editor->add_gui_control("hyper", &hyper_sc);

		v.tex_vox->bind_to_texture_unit(1);
		v.tex_vox->bind_to_image_unit(1);
		v.tex_voxb->bind_to_image_unit(2);
		v.tex_voxb->bind_to_texture_unit(2);
		v.tex_voxc->bind_to_image_unit(3);
		v.tex_voxc->bind_to_texture_unit(3);
		v.fb_GI->textures[0]->bind_to_image_unit(4);
		v.fb_GI->textures[0]->bind_to_texture_unit(4);
		v.fb_Post->textures[0]->bind_to_image_unit(5);
		v.fb_Post->textures[0]->bind_to_texture_unit(5);

		static int b = 0;
		e->editor->add_gui_control("reset", &b, GUIControlArgs{});
		if(b != 0) {
			b = 0;
			v.tex_vox->clear();
			v.tex_voxb->clear();
			v.tex_voxc->clear();
		}

		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);

		// ------ Render noise ----- //
		if(false) {
			e->start_timer_query("render noise");
			ZoneScopedN("render noise");
			v.prog_render_noise->use();
			v.prog_render_noise->setUniform("noise_tex", v.noise_tex);
			
			glm::ivec3 vox_tex_res = glm::ivec3(
				v.noise_tex->resx, v.noise_tex->resy, v.noise_tex->resz
			);
			v.prog_render_noise->dispatch_compute(vox_tex_res.x,vox_tex_res.y,vox_tex_res.z);
			
			glTextureBarrier();
		}
		// ------ Compute ----- //
		{
			ZoneScopedN("Compute");

			// for (auto shader : this->shader_manager->shader_programs) {
			v.prog_vox->use();
			// v.prog_vox->setUniform("hyper", hyper_sc);
			v.prog_vox->setUniform("vox_tex", v.tex_vox);
			v.prog_vox->setUniform("vox_texb", v.tex_voxb);
			v.prog_vox->setUniform("vox_texc", v.tex_voxc);
			// v.prog_vox->setUniform("hdri_tex", v.hdri_tex);
			v.prog_vox->setUniform("sun_dir", sun_dir);
			v.prog_vox->setUniform("sun_intensity", sun_intensity);
			// v.prog_vox->setUniform("cloud_density", cloud_denisty);
			// v.prog_vox->setUniform("cloud_colour", cloud_colour);
			// v.prog_vox->setUniform("sun_colour", sun_colour * sun_intensity);
			// v.prog_vox->setUniform("noise_tex", v.noise_tex);
			v.prog_vox->setUniform("render_tex", v.fb_GI->textures[0]);
			v.prog_vox->setUniform("out_tex", v.fb_Post->textures[0]);
			v.prog_vox->setUniform("ti", (float)e->time);
			// v.prog_vox->setUniform("ti", (float)846.5);
			// v.prog_vox->setUniform("f", (int)(frame % 10000));
			// v.prog_vox->setUniform("fi", (int)(frame % 10000));
			// v.prog_vox->setUniform("blue_noise_tex", v.tex_blue_noise);

			e->start_timer_query("compute 3");
			// v.prog_vox->setUniform("iter", 3);
			v.prog_vox->setUniform("fi", glm::ivec2(e->frame % 10000, 3));
			v.prog_vox->dispatch_compute(v.tex_vox->resx,v.tex_vox->resy,v.tex_vox->resz);
			glTextureBarrier();
			e->start_timer_query("compute 0");
			// v.prog_vox->setUniform("iter", 0);
			v.prog_vox->setUniform("fi", glm::ivec2(e->frame % 10000, 0));
			v.prog_vox->dispatch_compute(v.tex_vox->resx,v.tex_vox->resy,v.tex_vox->resz);
			glTextureBarrier();
			e->start_timer_query("compute 1");
			// v.prog_vox->setUniform("iter", 1);
			v.prog_vox->setUniform("fi", glm::ivec2(e->frame % 10000, 1));
			v.prog_vox->dispatch_compute(v.fb_GI->textures[0]->resx,v.fb_GI->textures[0]->resy);
			glTextureBarrier();
			e->start_timer_query("compute 2");
			// v.prog_vox->setUniform("iter", 2);
			v.prog_vox->setUniform("fi", glm::ivec2(e->frame % 10000, 2));
			v.prog_vox->dispatch_compute(v.fb_Post->textures[0]->resx,v.fb_Post->textures[0]->resy);
			glTextureBarrier();
			// v.prog_vox->setUniform("iter", 2);
			// v.prog_vox->dispatch_compute(v.fb_GI->textures[0]->resx,v.fb_GI->textures[0]->resy);

			// editor->gui_dbg_print("vox disp sz", glm::vec3(vox_tex_res));

			// glMemoryBarrier(GL_TEXTURE_FETCH_BARRIER_BIT);
		}

		// ------ Render ----- //
		{
			e->start_timer_query("render");
			ZoneScopedN("Render");
		//
		// 	// for (auto shader : this->shader_manager->shader_programs) {
		//
		// 	v.fb_GI->bind();
		// 	v.fb_GI->clear();
		// 	v.prog_render->use();
		// 	v.prog_render->setUniform("vox_tex", v.tex_vox);
		// 	v.prog_render->setUniform("vox_texb", v.tex_voxb);
		// 	v.prog_render->setUniform("vox_texc", v.tex_voxc);
		// 	v.prog_render->setUniform("blue_noise_tex", v.tex_blue_noise);
		// 	v.prog_render->setUniform("hdri_tex", v.hdri_tex);
		// 	v.prog_render->setUniform("sun_dir", sun_dir);
		// 	v.prog_render->setUniform("sun_intensity", sun_intensity);
		// 	v.prog_render->setUniform("cloud_density", cloud_denisty);
		// 	v.prog_render->setUniform("cloud_colour", cloud_colour);
		// 	v.prog_render->setUniform("sun_colour", sun_colour * sun_intensity);
		// 	v.prog_render->setUniform("noise_tex", v.noise_tex);
		// 	
		// 	glDrawArrays(GL_TRIANGLES, 0, 6);
		//
		// 	// glMemoryBarrier(GL_ALL_BARRIER_BITS);
		// 	// glTextureBarrier();
		}
		// ------ Post ----- //
		{
			e->start_timer_query("post");
			ZoneScopedN("Post");

			// for (auto shader : this->shader_manager->shader_programs) {

			// v.fb_Post->bind();
			// v.fb_Post->clear();
			//
			// v.prog_post->use();
			// v.prog_post->setUniform("render_tex", v.fb_GI->get_front_textures_at(0));
			//
			// glDrawArrays(GL_TRIANGLES, 0, 6);
			//
			// glMemoryBarrier(GL_ALL_BARRIER_BITS);
			// glTextureBarrier();
		}
		v.fb_Post->blit(e->default_fb);

		// hdri_tex->generate_mipmap();

		// v.prev_frame_gi_tex->clear();
	}

	// ------ End ----- //
	{
		ZoneScopedN("END RENDER");
		e->start_timer_query("frame_end");
		// editor->depth_texture = v.fb_G->get_front_depth();
		// this->endFrame();
	}

	{
		ZoneScopedN("MORE END OF RENDER");
		// audio_mixer->update_player(camera->pos, camera->dir, v.player->get_vel() * 10.0f);

		// drawer_3d->reset();
		// drawer_2d->reset();
		// text_drawer->reset_end_of_frame();

	}

}

