﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/Gif.h"
#include "engine/Video.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"
#include "passes/FXComposer.h"

static Lane* seq_amogus;
static Curve curve_posoomba;
static Curve curve_posoomba5;	
static Curve curve_posoomba4;	
static Curve curve_posoomba7;	
static Curve curve_posoomba2;	
static Lane* seq_agoomba;
static Curve curve_agoomba;
static Lane* seq_blorgus;
static Lane* seq_blorba;

static Texture* blue_noise_tex;

static ShaderProgram* prog_render;
static ShaderProgram* post_shader_prog;
static ShaderProgram* prog_vpt;
static ShaderProgram* prog_sim_particles;

static AudioFile* demo_audio;

static Framebuffer* fb_main;

static Texture* main_fb_texture;
static Framebuffer* vpt_fb;

static Buffer* vpt_buff;

static Texture* source_tex;

// static Model* scene_model = nullptr;

static FxComposer* composer;
static FXBloomPass* bloom_pass;

static WVideo* vid;

void VarsStippling::init() {
	VarsStippling& v = *this;
	Engine* e = WEngine;
	blue_noise_tex = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
	// source_tex = WEngine->alloc_textures.push(std::string("src/assets/images/bababooey.png")).ptr;
	// source_tex = WEngine->alloc_textures.push(std::string("C:\\Users\\user\\Downloads\\1093231088350019656.png")).ptr;
	source_tex = WEngine->alloc_textures.push(std::string("F:\\Programming\\Design\\_Resources\\dalle\\finished\\DALL·E 2022-07-15 19.19.03 - A portrait of a woman, wearing sunglasses, looking back at the camera. Her body is covered with a deep blue robe. Kodak Ektachrome 35mm. (1).png")).ptr;

	
	// vid = new WVideo(R"(file:///D:\VJ vids\abstract\pexels-mart-production-7565898.mp4)");
	vid = new WVideo(R"(file:///D:\VJ vids\human\pexels-darina-belonogova-7551560.mp4)");
	// vid = new WVideo(R"(file:///C:\Users\user\Downloads\leviathan-debug_0gyKR9HBkN(2).mp4)");
	// vid = new WVideo(R"(file:///C:\Users\user\Downloads\leviathan-debug_0gyKR9HBkN(2).mp4)");
	// C:\Users\user\Documents\ShareX\Screenshots\2024-06\leviathan-debug_7o8seVwpb3.mp4
	vid->update();
	
	
	

	Texture* main_fb_texture = WEngine->alloc_textures.push(
		TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
		}
	).ptr;
		
	fb_main = new Framebuffer(FBDesc{ .textures = { main_fb_texture }, .name = "main" });
	
	vpt_fb = new Framebuffer(FBDesc{ .textures = {
		WEngine->alloc_textures.push(
			TextureDesc{
					.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
					.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW
			}
		).ptr
	}, ._pongable = true,.name = "vpt fb" });
	
	vpt_buff = new Buffer(BuffDesc{.
		byte_len = 256 * megabyte_in_bytes,
		.immutable_subdata = true,
		.mappable_write = true,
		.mappable_read = false,
		.persistennt = false,
		.write_from_shader = true,
		.name = "Particles buff B"
	});

	// --- curve stuff
	{
		seq_amogus = WEngine->editor->timeline.AddSequence("Amogus");
		
		curve_posoomba = seq_amogus->AddCurve("asdg");
		curve_posoomba5 = seq_amogus->AddCurve("gdsfgd");	
		curve_posoomba4 = seq_amogus->AddCurve("gasdgsdd");	
		curve_posoomba7 = seq_amogus->AddCurve("fdddd");	
		seq_agoomba = WEngine->editor->timeline.AddSequence("Agoomba");
		seq_agoomba->AddCurve("dg023512");
		seq_agoomba->AddCurve("aaaa");
		seq_blorgus = WEngine->editor->timeline.AddSequence("Blorgus");
		seq_blorgus->AddCurve("tgdsgg");
		seq_blorba = WEngine->editor->timeline.AddSequence("Blorba");
		seq_blorba->AddCurve("339kfawsdthj");

		WEngine->editor->timeline.set_json_path("bruh_timeline.json");
	}

	bloom_pass = new FXBloomPass();
	composer = new FxComposer();
	composer->add_pass(bloom_pass);
	

	prog_render = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_stippling/default.frag")
	});
	
	post_shader_prog = new ShaderProgram(ShaderProgDescr{
		WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_stippling/post.frag")
	});

	// e->scene_model = e->alloc_scenes.push("assets\\models\\cube3.glb").ptr;
	
	prog_sim_particles = new ShaderProgram({ new Shader("scene_stippling/sim_particles.comp") });
	
	prog_vpt = new ShaderProgram(ShaderProgDescr{
		// WEngine->vert_sh_fullscreen_quad,
		new Shader("scene_stippling/vpt.comp")
	});
}

void VarsStippling::loop() {
	VarsStippling& v = *this;
	Engine* e = WEngine;
	e->start_timer_query("start");
	
	if (e->initializing) {
		demo_audio = new AudioFile(&e->audio_mixer->engine, "src/assets/sounds/music/wangler.wav");
		demo_audio->start();
		demo_audio->set_volume(0.1);
		e->initializing = false;
	}

	// ---- gui stuff
	static glm::vec3 some_col = glm::vec3(1);
	static float some_val = 0.1; 
	{
		e->editor->add_gui_control("some_col", &some_col, {.is_colour = true});
		e->editor->add_gui_control("some_val", &some_val);
	}

	static bool reset = true;

	e->editor->add_gui_control("reset", &reset, {.callback = [&]() {
		reset = true;
	}});

	if(reset) {
		vpt_fb->resize(
			vpt_fb->textures[0]->resx,
			vpt_fb->textures[0]->resy
		);
	}

	vpt_buff->bind_as_SSBO(12);



	
	// ---- raymarch pass
	// {
	// 	e->start_timer_query("raymarch");
	// 	fb_main->bind();
	// 	fb_main->clear();
	// 	
	// 	prog_render->use();
	// 	prog_render->setUniform("u_t", (float)e->time);
	// 	prog_render->setUniform("u_bpm", (float)115.0);
	// 	prog_render->setUniform("some_col", some_col);
	// 	
	// 	glDrawArrays(GL_TRIANGLES, 0, 6);
	// }
	
	vid->update();
	
	prog_sim_particles->setUniform("reset", reset);
	prog_sim_particles->setUniform("tex_vpt", vpt_fb->get_back_textures_at(0));
	prog_sim_particles->setUniform("source_tex", source_tex);

	int vpt_cnt = 100000;
	// ---- init particles pass
	{
		e->start_timer_query("sim particles");
		prog_sim_particles->use();
		prog_sim_particles->setUniform("pass", 0);
		prog_sim_particles->dispatch_compute(vpt_cnt,1,1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
	}
	
	{
		e->start_timer_query("spawn particles");
		prog_sim_particles->use();
		prog_sim_particles->setUniform("pass", 1);
		prog_sim_particles->dispatch_compute(vpt_cnt,1,1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
	}
	
	{
		e->start_timer_query("kill particles");
		prog_sim_particles->use();
		prog_sim_particles->setUniform("pass", 2);
		prog_sim_particles->dispatch_compute(vpt_cnt,1,1);
		glMemoryBarrier(GL_ALL_BARRIER_BITS);
	}

	{
		e->start_timer_query("vpt track");
		prog_vpt->use();
		prog_vpt->setUniform("tex_vpt_prev", vpt_fb->get_back_textures_at(0));
		// prog_vpt->setUniform("source_tex", source_tex);
		prog_vpt->setUniform("source_tex", vid->tex);
		prog_vpt->setUniform("out_tex", vpt_fb->get_front_textures_at(0));

		int iters = 9;
		
		if(false) {
			prog_vpt->setUniform("tech", 0);
			prog_vpt->dispatch_compute(
				vpt_fb->get_front_textures_at(0)->resx,
				vpt_fb->get_front_textures_at(0)->resy,
				1
			);
			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		} else {
			prog_vpt->setUniform("tech", 1);
			prog_vpt->setUniform("iters", iters);

			for(int i = 0; i < iters; i++) {
				prog_vpt->setUniform("iter", i);
				prog_vpt->dispatch_compute(
					vpt_fb->get_front_textures_at(0)->resx,
					vpt_fb->get_front_textures_at(0)->resy,
					1
				);
				glMemoryBarrier(GL_ALL_BARRIER_BITS);
			}
		}
		// vpt_fb->bind();
		// vpt_fb->clear();

		// glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	// composer->run();
	composer->run = [&](Texture* input_tex) -> Texture* {
		if(composer->internal_res.x != input_tex->resx || composer->internal_res.y != input_tex->resy) {
			composer->resize(input_tex->resx, input_tex->resy);
		}
			
		FXBloomPass* bloom_pass = (FXBloomPass*)composer->passes[0];

		bloom_pass->run(input_tex, composer->internal_fb );
		composer->internal_fb->pong();

		return composer->get_back_tex();
	};

	Texture* composer_out_tex = composer->run(fb_main->textures[0]);
	
	// ---- Post pass
	{
		e->start_timer_query("post");
		e->default_fb->bind();
		e->default_fb->clear();
		
		post_shader_prog->use();
		// post_shader_prog->setUniform("tex_render", fb_main->textures[0]);
		post_shader_prog->setUniform("tex_vpt", vpt_fb->get_front_textures_at(0));
		post_shader_prog->setUniform("source_tex", source_tex);
		
		glDrawArrays(GL_TRIANGLES, 0, 6);
	}

	e->start_timer_query("end");
	
	reset = false;
}
