﻿#include "pch.h"
// #include "Light.h"
// #include "WAlloc.h"
#include "engine/Camera.h"
#include "engine/Engine.h"
#include "engine/audio/AudioMixer.h"
#include "gl/Framebuffer.h"
#include "gl/Shader.h"
#include "gl/ShaderManager.h"
#include "gl/WGL.h"
#include "engine/LUT.h"


// (4 + 4 + 4 + 4) * 4
int particle_byte_sz = 4 * (4 + 4 + 4 + 4 + 4 + 4);
// int particles_cnt = 20000;
int particles_cnt = ((20000)/256)*256;
int tree_res = 15;
int tree_particles_in_cell = 300;
struct DispatchIndirectCommand{
	uint32_t sz_x;
	uint32_t sz_y;
	uint32_t sz_z;
};

Buffer* particles_buff_b;
Buffer* particles_to_spawn_buffer;
Buffer* dispatch_indirect_buffer;

void VarsTreeParticles::init() {
		VarsTreeParticles& v = *this;
		Engine* e = WEngine;
	
		WEngine->camera->noclip = true;
	
		v.splat_tex = WEngine->alloc_textures.push(TextureDesc{
			.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
			.internal_format = InternalFormat::R32UI, .tex_access = TexAccess::RW, .name = "Splat"
		}).ptr;
	
		v.fb_Post = new Framebuffer(FBDesc{
			.textures = {WEngine->alloc_textures.push(TextureDesc{
				.resx = WEngine->RESX, .resy = WEngine->RESY, .resz = 1,
				.internal_format = InternalFormat::RGBA32F, .tex_access = TexAccess::RW
			}).ptr}, .name = "Post"
		});
	

		v.tree_buff = new Buffer(BuffDesc{.
			byte_len = tree_res * tree_res * (tree_particles_in_cell + 10) * 4,
			.immutable_subdata = true,
			.mappable_write = true,
			.mappable_read = false,
			.persistennt = false,
			.write_from_shader = true,
			.name = "Particles tree buff"
		});
		
		particles_to_spawn_buffer = new Buffer(BuffDesc{.
			byte_len = (50000) * particle_byte_sz,
			.immutable_subdata = true,
			.mappable_write = true,
			.mappable_read = false,
			.persistennt = false,
			.write_from_shader = true,
			.name = "Particles to spawn buff"
		});
	
		particles_buff_b = new Buffer(BuffDesc{.
			byte_len = (particles_cnt + 100) * particle_byte_sz,
			.immutable_subdata = true,
			.mappable_write = true,
			.mappable_read = false,
			.persistennt = false,
			.write_from_shader = true,
			.name = "Particles buff B"
		});
	
		v.particles_buff = new Buffer(BuffDesc{.
			byte_len = (particles_cnt + 100) * particle_byte_sz,
			.immutable_subdata = true,
			.mappable_write = true,
			.mappable_read = false,
			.persistennt = false,
			.write_from_shader = true,
			.name = "Particles buff"
		});
		v.particles_free_slots_buff = new Buffer(BuffDesc{.
			byte_len = (particles_cnt + 10) * 4,
			.immutable_subdata = false,
			// .mappable_write = true,
			// .mappable_read = false,
			// .persistennt = false,
			.write_from_shader = true,
			.name = "Particles free slots buff"
		});
		
		dispatch_indirect_buffer = new Buffer(BuffDesc{.
			byte_len = 100,
			.immutable_subdata = false,
			.write_from_shader = true,
			.name = "Dispatch indirect"
		});
		dispatch_indirect_buffer->bind_as_indirect_compute_buff();
	
	
	  std::vector<int> particle_free_slots;
		// particle_free_slots.reserve(particles_cnt + 4);
		particle_free_slots.push_back(particles_cnt-1);
		particle_free_slots.push_back(0);
		particle_free_slots.push_back(0);
		particle_free_slots.push_back(0);
	  for(int i = 0; i < particles_cnt; i++) {
			particle_free_slots.push_back(i);
	  }
		v.particles_free_slots_buff->upload_sub_data(&particle_free_slots[0], v.particles_free_slots_buff->byte_len, 0);
	

		v.prog_fill_tree = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_tree_particles/fill_tree.comp")
		});
		v.prog_clean_tree = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_tree_particles/clean_tree.comp")
		});
		v.prog_spawn_particles = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_tree_particles/spawn_particles.comp")
		});
		v.prog_simulate = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_tree_particles/simulate.comp")
		});
		v.prog_render = new ShaderProgram(ShaderProgDescr{
			new Shader("scene_tree_particles/render.comp")
		});
		v.prog_post = new ShaderProgram(ShaderProgDescr{
			WEngine->vert_sh_fullscreen_quad,
			new Shader("scene_tree_particles/post.frag")
		});

		v.tex_blue_noise = WEngine->alloc_textures.push(std::string("src/assets/images/blue_noise_64_64_RGBA.png")).ptr;
}

void VarsTreeParticles::loop() {
	VarsTreeParticles& v = *this;
	Engine* e = WEngine;
	
	ZoneScopedN("LOOP");
	// Sleep(1000);
	// this->startFrame();

	glDisable(GL_BLEND);

	static int pong = 0;
	static int substeps_count = 1;
	 //  std::vector<int> particle_free_slots;
		// // particle_free_slots.reserve(particles_cnt + 4);
		// particle_free_slots.push_back(particles_cnt);
		// particle_free_slots.push_back(0);
		// particle_free_slots.push_back(0);
		// particle_free_slots.push_back(0);
	 //  for(int i = 0; i < particles_cnt; i++) {
		// 	particle_free_slots.push_back(i);
	 //  }
		// v.particles_free_slots_buff->upload_sub_data(&particle_free_slots[0], v.particles_free_slots_buff->byte_len, 0);
	e->camera->speed = 0.3;
	e->camera->_far = 200.0;

	static bool indirect_dispatch = false;
	
	if (e->initializing) {
		ZoneScopedN("re-init");
		wlog_info("-- RE-INITIALIZING -- ");
		e->camera->pos = glm::vec3(0.062, 0.0, 1.181);

		e->initializing = false;
	}
	
	{
		e->editor->gui_dbg_print("mouse ndc", e->io->mouse_ndc);
	}

	static bool to_reset = false;
	static bool spawn_particle = false;
	static bool reset_prev_frame = false;
	
	if(reset_prev_frame || to_reset) {
	  std::vector<int> particle_free_slots;
		// particle_free_slots.reserve(particles_cnt + 4);
		particle_free_slots.push_back(particles_cnt-1);
		particle_free_slots.push_back(0);
		particle_free_slots.push_back(0);
		particle_free_slots.push_back(0);
	  for(int i = 0; i < particles_cnt; i++) {
			particle_free_slots.push_back(i);
	  }
		v.particles_free_slots_buff->upload_sub_data(&particle_free_slots[0], v.particles_free_slots_buff->byte_len, 0);
	}
	{
		ZoneScopedN("RENDER COMMANDS");
		e->start_timer_query("frame_start");

		glMemoryBarrier(GL_ALL_BARRIER_BITS);
		
		e->editor->add_gui_control("cam pos", &e->camera->pos);

		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);

		// e->editor->add_gui_control("reset")
		e->editor->add_gui_control("reset", &to_reset, {.callback = [&]() {
			to_reset = true;
		}});
		e->editor->add_gui_control("indirect_dispatch", &indirect_dispatch);
		
		e->editor->add_gui_control("substeps_count", &substeps_count);
		e->editor->add_gui_control("spawn_particle", &spawn_particle, {.callback = [&]() {
			spawn_particle = true;
		}});
		
		e->editor->add_gui_control("can_fov", &e->camera->fov);
		
		glDisable(GL_DEPTH_TEST);
		tree_buff->bind_as_SSBO(5);

		// if(pong == 0){
		// 	particles_buff->bind_as_SSBO(6);
		// 	particles_buff_b->bind_as_SSBO(7);
		// } else {
		// 	particles_buff->bind_as_SSBO(7);
		// 	particles_buff_b->bind_as_SSBO(6);
		// }
		
		particles_free_slots_buff->bind_as_SSBO(8);
		particles_to_spawn_buffer->bind_as_SSBO(9);
		v.splat_tex->clear();

		// ------ Render ----- //
		
		// Buffer* tree_buff;
		// Buffer* particles_buff;
		// Buffer* particles_free_slots_buff;
		// int tree_res = 10;
		// int tree_particles_in_cell = 1000;

		int effective_substeps_count = 1 + (substeps_count - 1)*2;
		{
			e->start_timer_query("simulate");
			

			for(int i = 0; i < effective_substeps_count; i++) {
				uint32_t dispatch_indirect_size[] = {
					uint32_t(particles_cnt/32),
					1u,
					1u
				};
				if(!indirect_dispatch) {
					dispatch_indirect_buffer->upload_sub_data(dispatch_indirect_size,4 * 3);
				}
				if(i != 0) {
					pong = 1-pong;
				}
				
				if(pong == 0){
					particles_buff->bind_as_SSBO(6);
					particles_buff_b->bind_as_SSBO(7);
				} else {
					particles_buff->bind_as_SSBO(7);
					particles_buff_b->bind_as_SSBO(6);
				}
				
				v.prog_simulate->use();

				// effective_substeps_count
				v.prog_simulate->setUniform("indirect_dispatch", indirect_dispatch);
				v.prog_simulate->setUniform("substeps_count", effective_substeps_count);
				v.prog_simulate->setUniform("substep", i);
				v.prog_simulate->setUniform("tree_res", tree_res);
				v.prog_simulate->setUniform("tree_particles_in_cell", tree_particles_in_cell);
				v.prog_simulate->setUniform("tree_buff", tree_buff);
				// v.prog_simulate->setUniform("particles_buff", particles_buff);
				v.prog_simulate->setUniform("particles_free_slots_buff", particles_free_slots_buff);
				v.prog_simulate->setUniform("to_reset", reset_prev_frame || to_reset);
				v.prog_simulate->setUniform("spawn_particle", spawn_particle);
				v.prog_simulate->setUniform("splat_tex", v.splat_tex);
				v.prog_simulate->setUniform("dispatch_indirect_buffer", dispatch_indirect_buffer);
				

				
			// uint32_t dispatch_indirect_size[] = {
			// 	uint32_t((particles_cnt + v.prog_simulate->wkg_sz.x - 1) / (v.prog_simulate->wkg_sz.x)),
			// 	1u,
			// 	1u
			// };
				// dispatch_indirect_buffer->upload_sub_data(dispatch_indirect_size,4 * 3);
				glDispatchComputeIndirect(0);
				// v.prog_simulate->dispatch_compute(particles_cnt,1,1);

				glMemoryBarrier(GL_ALL_BARRIER_BITS);
				
			}
		}
		
		{
			e->start_timer_query("spawn particles");
			
			v.prog_spawn_particles->use();
			
			v.prog_spawn_particles->setUniform("tree_res", tree_res);
			v.prog_spawn_particles->setUniform("tree_particles_in_cell", tree_particles_in_cell);
			v.prog_spawn_particles->setUniform("tree_buff", tree_buff);
			// v.prog_simulate->setUniform("particles_buff", particles_buff);
			v.prog_spawn_particles->setUniform("particles_free_slots_buff", particles_free_slots_buff);
			v.prog_spawn_particles->setUniform("to_reset", reset_prev_frame || to_reset);
			v.prog_spawn_particles->setUniform("spawn_particle", spawn_particle);
			v.prog_spawn_particles->setUniform("splat_tex", v.splat_tex);
			
			v.prog_spawn_particles->dispatch_compute(particles_cnt,1,1);

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}
		
		{
			e->start_timer_query("clean tree");
			prog_clean_tree->use();
			
			v.prog_clean_tree->setUniform("tree_res", tree_res);
			v.prog_clean_tree->setUniform("tree_particles_in_cell", tree_particles_in_cell);
			v.prog_clean_tree->setUniform("tree_buff", tree_buff);
			v.prog_clean_tree->setUniform("particles_buff", particles_buff);
			v.prog_clean_tree->setUniform("particles_free_slots_buff", particles_free_slots_buff);
			
			v.prog_clean_tree->dispatch_compute(1,1,1);
			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}
		
		
		{
			e->start_timer_query("fill tree");
			prog_fill_tree->use();
			
			v.prog_fill_tree->setUniform("tree_res", tree_res);
			v.prog_fill_tree->setUniform("tree_particles_in_cell", tree_particles_in_cell);
			v.prog_fill_tree->setUniform("tree_buff", tree_buff);
			// v.prog_fill_tree->setUniform("particles_buff", particles_buff);
			v.prog_fill_tree->setUniform("particles_free_slots_buff", particles_free_slots_buff);
			
			v.prog_fill_tree->dispatch_compute(particles_cnt,1,1);
			
			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}
		
		{
			e->start_timer_query("render");
			ZoneScopedN("Render");

			
			v.prog_render->use();
			v.prog_render->setUniform("blue_noise_tex", v.tex_blue_noise);
			v.prog_render->setUniform("splat_tex", v.splat_tex);
			v.prog_render->setUniform("dispatch_indirect_buffer", dispatch_indirect_buffer);
			v.prog_render->setUniform("indirect_dispatch", indirect_dispatch);
			v.prog_render->setUniform("to_reset", reset_prev_frame || to_reset);
			
			v.prog_render->setUniform("tree_res", tree_res);
			v.prog_render->setUniform("tree_particles_in_cell", tree_particles_in_cell);
			v.prog_render->setUniform("tree_buff", tree_buff);
			// v.prog_render->setUniform("particles_buff", particles_buff);
			v.prog_render->setUniform("particles_free_slots_buff", particles_free_slots_buff);
			
			v.prog_render->dispatch_compute(particles_cnt,1,1);
			

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
		}

		// ------ Post ----- //
		{
			e->start_timer_query("post");
			ZoneScopedN("Post");

			fb_Post->bind();
			fb_Post->clear();

			prog_post->use();
			prog_post->setUniform("splat_tex", v.splat_tex);
			
			prog_post->setUniform("tree_res", tree_res);
			prog_post->setUniform("tree_particles_in_cell", tree_particles_in_cell);
			prog_post->setUniform("tree_buff", tree_buff);
			prog_post->setUniform("particles_buff", particles_buff);
			prog_post->setUniform("particles_free_slots_buff", particles_free_slots_buff);

			glDrawArrays(GL_TRIANGLES, 0, 6);

			glMemoryBarrier(GL_ALL_BARRIER_BITS);
			glTextureBarrier();
		}
		v.fb_Post->blit(e->default_fb);
	}

	// ------ End ----- //
	{
		ZoneScopedN("END RENDER");
		e->start_timer_query("frame_end");
	}

	{
		ZoneScopedN("MORE END OF RENDER");
		e->drawer_3d->reset();
		e->drawer_2d->reset();
	}
	reset_prev_frame = to_reset;
	to_reset = false;
	spawn_particle = false;
	pong = 1-pong;
}

